/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <exception>
#include <iterator>
#include <iostream>

#include "../sorted_stream.h"
#include "../elements.h"
#include "../stopwatch.h"
#include "../../SortGen/random.h"
#include "../../Sort/Sort.h"
#include "../../Sort/funnel/stack_allocator.h"
#include "../../Sort/funnel/merge_tree.h"
#include "../../Sort/funnel/merge_tree_types.h"

#include "../../Sort/PointerBMT.h"
#include "../../Sort/ImplicitBMT.h"
#include "../../Sort/LazyMerger.h"
#include "../../Sort/HeapMerger.h"

template<funnel::order_t order>
class my_splitter
{
public:
	static inline funnel::order_t prefered_order_output(_uint64_t N)
		{ return (funnel::order_t)floor(pow((double)N, 0.5)/16+.5); }
	static inline funnel::order_t prefered_order_input(_uint64_t N)
		{ return (funnel::order_t)floor(pow((double)N, 0.5)/16+.5); }
	static inline uint64_t switch_to_cache_aware()
		{ return out_buffer_size(order+1); }
	static inline funnel::height_t split(funnel::height_t h)
			{ return h/2; }
	static inline size_t out_buffer_size(funnel::order_t k)
		{ return 16*static_cast<size_t>(k)*k; }
};

template<class Allocator>
class allocator_factory
{
public:
	inline Allocator operator()(typename Allocator::size_type size)
		{ return Allocator(); }
};
template<class T>
class allocator_factory<stack_allocator<T> >
{
public:
	inline stack_allocator<T> operator()(typename stack_allocator<T>::size_type size)
		{ return stack_allocator<T>(size); }
};

template<class Merger>
void run_merger(funnel::basic_order_t order, const char *v)
{
	std::cout << "\t<run buildtime=\"" << sys_names::buildtime() << "\" compiler=\"" << sys_names::compiler() << "\" execstatus=\"0\">" << std::endl;
	std::cout << "\t\t<size>" << order << "</size>" << std::endl;
	std::cout << "\t\t<algorithm>" << v << "</algorithm>" << std::endl;
	const funnel::order_t k = 120;
	typedef typename Merger::value_type T;
	typedef typename Merger::stream stream;
	try
	{
		stopwatch w;
		miss_counter mc;
		allocator_factory<typename Merger::allocator> af;
		typename Merger::allocator alloc = af(Merger::size_of(k));
		Merger merger(k,alloc);
		T *input = new T[k*k*k];

		random_source rs;
		T chksum;
		rs >> chksum;
		*input = chksum;
		for( T *p=input+1; p!=input+k*k*k; ++p )
			rs >> *p, chksum += *p;
		for( T *p=input; p+k*k<=input+k*k*k; p+=k*k)
			std::sort(p, p+k*k);
		mc.start();
		w.start();
		for( int i=0; i<8; ++i )
		{
			sorted_stream<T> output(k*k*k);
			for( T *p=input; p+k*k<=input+k*k*k; p+=k*k)
				merger.add_stream(stream(p,p+k*k));
			if( merger(output.begin(), output.end()) != output.end() )
				std::cout << "\t\t<error>Incomplete output!</error>" << std::endl;
			if( output.checksum() != chksum )
				std::cout << "\t\t<error>Checksum error: " << output.checksum() << " != " << chksum << "</error>" << std::endl;
			merger.reset();
		}
		w.stop();
		mc.stop();
		delete[] input;
		std::cout << w.xmltag(2) << mc.xmltag() << std::endl;
	}
	catch( std::exception& ex )
	{
		std::cout << "\t\t<error>Caught std::exception: " << ex.what() << "</error>" << std::endl;
	}
	std::cout << "\t</run>" << std::endl;
}

template<funnel::basic_order_t order>
void run_common()
{
	using namespace funnel;
	run_merger<typename merge_tree_types<element,order,buffer<element*>,simple_for_merger<order,buffer<element*>,element>,my_splitter<order> >::rec_heap_mveb>(order,"simple_for_merger");
	run_merger<typename merge_tree_types<element,order,buffer<element*>,for_merger<order,buffer<element*>,element>,my_splitter<order> >::rec_heap_mveb>(order,"for_merger");
	run_merger<typename merge_tree_types<element,order,buffer<element*>,simple_for_val_merger<order,buffer<element*>,element>,my_splitter<order> >::rec_heap_mveb>(order,"simple_for_val_merger");
	run_merger<typename merge_tree_types<element,order,buffer<element*>,for_val_merger<order,buffer<element*>,element>,my_splitter<order> >::rec_heap_mveb>(order,"for_val_merger");
	run_merger<typename merge_tree_types<element,order,buffer<element*>,simple_comp_merger<order,buffer<element*>,element>,my_splitter<order> >::rec_heap_mveb>(order,"simple_comp_merger");
	run_merger<typename merge_tree_types<element,order,buffer<element*>,comp_merger<order,buffer<element*>,element>,my_splitter<order> >::rec_heap_mveb>(order,"comp_merger");
	run_merger<typename merge_tree_types<element,order,buffer<element*>,loser_merger<order,buffer<element*>,element>,my_splitter<order> >::rec_heap_mveb>(order,"loser_merger");
}

template<funnel::basic_order_t order>
void run()
{
	run_common<order>();
}
template<>
void run<2>()
{
	using namespace funnel;
	run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(2,"simple_merger");
	run_merger<merge_tree_types<element,2,buffer<element*>,two_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(2,"two_merger");
	run_merger<merge_tree_types<element,2,buffer<element*>,hl_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(2,"hl_merger");
	run_merger<merge_tree_types<element,2,buffer<element*>,hyb0_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(2,"hyb0_merger");
	run_merger<merge_tree_types<element,2,buffer<element*>,hyb_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(2,"hyb_merger");
	run_merger<merge_tree_types<element,2,buffer<element*>,hyb3_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(2,"hyb3_merger");
	run_common<2>();
}
template<>
void run<4>()
{
	using namespace funnel;
	run_merger<merge_tree_types<element,4,buffer<element*>,four_merger<4,buffer<element*>,element>,my_splitter<4> >::rec_heap_mveb>(4,"four_merger");
	run_common<4>();
}

int main(int argc, char *argv[])
{
	local_time t;
	std::cout << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << std::endl;
	std::cout << "<benchmark date=\"" << t << "\" os=\"" << sys_names::os() << "\" version=\"" << sys_names::release() << "\" host=\"" << sys_names::host() << "\">" << std::endl;
	run<2>();
	run<3>();
	run<4>();
	run<5>();
	run<6>();
	run<7>();
	run<8>();
	run<9>();
	std::cout << "</benchmark>" << std::endl;
}
