/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <exception>
#include <iostream>

#include "../sorted_stream.h"
#include "../elements.h"
#include "../stopwatch.h"
#include "../../SortGen/random.h"
#include "../../Sort/Sort.h"
#include "../../Sort/funnel/stack_allocator.h"
#include "../../Sort/funnel/merge_tree.h"
#include "../../Sort/funnel/merge_tree_types.h"

#include "../../Sort/PointerBMT.h"
#include "../../Sort/ImplicitBMT.h"
#include "../../Sort/LazyMerger.h"
#include "../../Sort/HeapMerger.h"

template<funnel::order_t order>
class my_splitter
{
public:
	static inline funnel::order_t prefered_order_output(_uint64_t N)
		{ return (funnel::order_t)floor(pow((double)N, 1.0/3)/1+.5); }
	static inline funnel::order_t prefered_order_input(_uint64_t N)
		{ return (funnel::order_t)floor(pow((double)N, 1.0/3)/1+.5); }
	static inline uint64_t switch_to_cache_aware()
		{ return out_buffer_size(order+1); }
	static inline funnel::height_t split(funnel::height_t h)
			{ return h/2; }
	static inline size_t out_buffer_size(funnel::order_t k)
		{ return 1*static_cast<size_t>(k)*k*k; }
};

template<class Allocator>
class allocator_factory
{
public:
	inline Allocator operator()(typename Allocator::size_type size)
		{ return Allocator(); }
};
template<class T>
class allocator_factory<stack_allocator<T> >
{
public:
	inline stack_allocator<T> operator()(typename stack_allocator<T>::size_type size)
		{ return stack_allocator<T>(size); }
};

template<class Merger>
void run_merger(funnel::order_t k, const char* v, int balanced = 0)
{
	std::cout << "\t<run buildtime=\"" << sys_names::buildtime() << "\" compiler=\"" << sys_names::compiler() << "\" execstatus=\"0\">" << std::endl;
	std::cout << "\t\t<size>" << k << "</size>" << std::endl;
	std::cout << "\t\t<algorithm>" << v << "</algorithm>" << std::endl;
	typedef typename Merger::value_type T;
	typedef typename Merger::stream stream;
	try
	{
		funnel::order_t bk = k;
		if( balanced == 2 )
			bk = funnel::pow_of_order<2>(funnel::logc<2>(k));
		if( balanced == 4 )
			bk = funnel::pow_of_order<4>(funnel::logc<4>(k));
		stopwatch w;
		miss_counter mc;
		allocator_factory<typename Merger::allocator> af;
		typename Merger::allocator alloc = af(Merger::size_of(bk));
		Merger merger(bk,alloc);
		T *input = new T[k*k*k];
		sorted_stream<T> output(k*k*k);

		random_source rs;
		T chksum;
		rs >> chksum;
		*input = chksum;
		for( T *p=input+1; p!=input+k*k*k; ++p )
			rs >> *p, chksum += *p;
		for( T *p=input; p+k*k<=input+k*k*k; p+=k*k)
			std::sort(p, p+k*k);
		mc.start();
		w.start();
		for( int i=0; i<20000000/(k*k*k); ++i )
		{
			for( T *p=input; p+k*k<=input+k*k*k; p+=k*k)
				merger.add_stream(stream(p,p+k*k));
			merger(output.begin(), output.end());
			merger.reset();
		}
		w.stop();
		mc.stop();
		delete[] input;
		std::cout << w.xmltag(2) << mc.xmltag() << std::endl;
	}
	catch( std::exception& ex )
	{
		std::cerr << "Caught std::exception: " << ex.what() << std::endl;
	}
	std::cout << "\t</run>" << std::endl;
}

int main(int argc, char *argv[])
{
	local_time t;
	std::cout << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << std::endl;
	std::cout << "<benchmark date=\"" << t << "\" os=\"" << sys_names::os() << "\" version=\"" << sys_names::release() << "\" host=\"" << sys_names::host() << "\">" << std::endl;
	using namespace funnel;
	for( funnel::order_t k=15; k<=271; k+=5 )
	{
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mveb>(k,"rec_heap_mveb");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_stack_mveb>(k,"rec_stack_mveb");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_veb>(k,"rec_heap_veb");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_stack_veb>(k,"rec_stack_veb");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mdf>(k,"rec_heap_mdf");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_stack_mdf>(k,"rec_stack_mdf");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_df>(k,"rec_heap_df");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_stack_df>(k,"rec_stack_df");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_mbf>(k,"rec_heap_mbf");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_stack_mbf>(k,"rec_stack_mbf");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_heap_bf>(k,"rec_heap_bf");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::rec_stack_bf>(k,"rec_stack_bf");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::pb_heap_mveb>(k,"pb_heap_mveb");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::pb_stack_mveb>(k,"pb_stack_mveb");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::impl_mveb>(k,"impl_mveb",2);
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::pb_heap_veb>(k,"pb_heap_veb");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::pb_stack_veb>(k,"pb_stack_veb");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::impl_veb>(k,"impl_veb");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::pb_heap_mdf>(k,"pb_heap_mdf");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::pb_stack_mdf>(k,"pb_stack_mdf");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::pb_heap_df>(k,"pb_heap_df");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::pb_stack_df>(k,"pb_stack_df");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::impl_df>(k,"impl_df");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::pb_heap_mbf>(k,"pb_heap_mbf");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::pb_stack_mbf>(k,"pb_stack_mbf");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::pb_heap_bf>(k,"pb_heap_bf");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::pb_stack_bf>(k,"pb_stack_bf");
		run_merger<merge_tree_types<element,2,buffer<element*>,simple_merger<2,buffer<element*>,element>,my_splitter<2> >::impl_bf>(k,"impl_bf");

		run_merger<LazyMerger<PBBinMergeTree<element>, InputStream<element*> > >(k,"pb_lazy",2);
		run_merger<LazyMerger<ImplBinMergeTree<element>, InputStream<element*> > >(k,"impl_lazy",2);
		run_merger<HeapMerger<element, 20000000, 3500, InputStream<element*> > > (k,"bin_heap");
	}
	std::cout << "</benchmark>" << std::endl;
}
