/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __FARRAY_H_INCLUDED_
#define __FARRAY_H_INCLUDED_

#include "aarray.aux.h"

#include <limits.h>
#include <cassert>
#include <stdexcept>
#include <iterator>
#include <iostream>

#ifdef _MSC_VER
#include <xmmintrin.h>
#define inline __forceinline
#endif // _MSC_VER

#ifdef WIN32
DWORD WINAPI __PrefetchThreadFunc(LPVOID lpParameter);
#else // WIN32
void* __PrefetchThreadFunc(void* lpParameter);
#endif // WIN32

extern const void* fetch_rq[1024];
extern const void* fetch_co[1024];
extern int rqc, coc;

class __PrefetchThread
{
#ifdef WIN32
	friend DWORD WINAPI __PrefetchThreadFunc(LPVOID lpParameter);
#else // WIN32
	friend void* __PrefetchThreadFunc(void* lpParameter);
#endif // WIN32
public:
	__PrefetchThread();
	~__PrefetchThread();
	void Fetch(const void* p);
private:
#ifdef WIN32
	HANDLE m_hThread;
#else // WIN32
	pthread_t m_Thread;
#endif // WIN32
	struct __PrefetchThreadData
	{
#ifdef WIN32
		HANDLE m_hsEmpty;
		HANDLE m_hsFull;
#else // WIN32
		sem_t m_sEmpty;
		sem_t m_sFull;
#endif // WIN32
		const void* m_pFetchAddr[16];
	} m_Shared;
	unsigned char m_Head;
};

namespace auto_array
{

inline void CachePrefetch(const void* addr)
{
#ifdef _MSC_VER
	_mm_prefetch((char*)addr, _MM_HINT_T2);
#else // _MSC_VER
#ifdef __INTEL_COMPILER
#else // __INTEL_COMPILER
#if __GNUC__ >= 3
	__builtin_prefetch(addr, 0, 2);
#else // __GNUC__
	/* unknown compiler */
#endif // __GNUC__
#endif // __INTEL_COMPILER
#endif // _MSC_VER
}

inline void MemPrefetch(const void* addr)
{
	static __PrefetchThread pt;
	if( addr )
		pt.Fetch(addr);
}

inline void LockMem(const void* start, const void* end);
inline void UnlockMem(const void* start, const void* end);



template<class T, size_t BufferSize = 4096>
class binfile_ostream
{
public:
	class iterator : public std::iterator<std::output_iterator_tag,T>
	{
		friend class binfile_ostream<T>;
	public:
		inline T& operator*()
		{
			return *stream->here;
		}
		inline iterator& operator++()
		{
			++stream->here;
			if( stream->here == stream->last )
				stream->flush();
			return *this;
		}
		inline iterator operator++(int)
		{
			iterator ret = *this;
			++stream->here;
			if( stream->here == stream->last )
				stream->flush();
			return ret;
		}
	private:
		inline explicit iterator(binfile_ostream *stream) : stream(stream) { }
		binfile_ostream *stream;
	};

	explicit binfile_ostream(const char *filename) : here(buffer), last(buffer+BufferSize)
	{
#ifdef WIN32
		m_Handle = CreateFile(filename, GENERIC_READ | GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
		if( m_Handle == INVALID_HANDLE_VALUE )
#else //WIN32
		fd = open(filename, O_CREAT | O_TRUNC | O_RDWR | O_LARGEFILE, 0666);
		if( fd == -1 )
#endif // WIN32
			throw std::runtime_error("Error opening file");
	}
	virtual ~binfile_ostream()
	{
		flush();
#ifdef WIN32
		CloseHandle(m_Handle);
#else //WIN32
		close(fd);
#endif // WIN32
	}

	inline iterator begin()
	{
		return iterator(this);
	}

private:
	inline void flush()
	{
		char *buf = reinterpret_cast<char*>(buffer);
		char *h = reinterpret_cast<char*>(here);
		size_t missing = h-buf;
		while( missing )
		{
#ifdef WIN32
			DWORD written;
			if( WriteFile(m_Handle,h-missing,missing,&written,NULL) == 0 )
#else //WIN32
			int written = write(fd,h-missing,missing);
			if( written == -1 )
#endif // WIN32
			{
//				throw std::runtime_error("Error flushing");
				std::cerr << "Error flushing" << std::endl;
				abort();
			}
			missing -= written;
		}
		here = buffer;
	}

	T buffer[BufferSize];
	T *here, *last;
#ifdef WIN32
	HANDLE m_Handle;
#else //WIN32
	int fd;
#endif // WIN32
};



class HeapMemMan
{
public:
	template<class T>
	static T* Alloc()
		{ return new T(); }
	template<class T, class A>
	static T* Alloc(A a)
		{ return new T(a); }
	template<class T, class A, class B>
	static T* Alloc(A a, B b)
		{ return new T(a, b); }
	template<class T>
	static T* Alloca(size_t n)
		{ return new T[n]; }
	template<class T>
	static T* Allocap(size_t n)
		{ return (T*)malloc(n*sizeof(T)); }
	template<class T>
	static void DAlloc(T* t)
		{ delete t; }
	template<class T>
	static void DAlloca(T* t)
		{ delete[] t; }
	template<class T>
	static void DAllocap(T* t)
		{ free(t); }
};

template<class T>
class FArray
{
public:
	typedef size_t size_type;
	typedef T* iterator;

	FArray() { }
	FArray(T* p) : array(p) { }
	virtual ~FArray() { }

	inline T& operator[](size_t i)
		{ return array[i]; }
	inline const T& operator[](size_t i) const
		{ return array[i]; }
	inline FArray operator+(size_t offset) const
		{ return FArray(array+offset); }
	inline bool operator==(void* p)
		{ return p==NULL && array==NULL; }
	inline T* begin()
		{ return array; }
	inline T* end(size_t size)
		{ return array+size; }

protected:
	T* array;
};

template<class T, class MemMan = HeapMemMan>
class dynarray : public FArray<T>
{
public:
	inline dynarray() : FArray<T>(NULL) { }
	inline explicit dynarray(size_t size) : FArray<T>(MemMan::template Alloca<T>((size_t)size)), m_pRefCount(MemMan::template Alloc<size_t>(1))
		{ INIT_MUTEX(mutex); }
	inline dynarray(const dynarray<T, MemMan>& src);
	virtual ~dynarray();
	inline dynarray<T, MemMan>& operator=(dynarray<T, MemMan>& rhs);

	inline dynarray<T, MemMan> operator+(size_t offset) const
#ifdef _MT
		{ return dynarray<T, MemMan>(array+offset, m_pRefCount, mutex); }
#else // _MT
		{ return dynarray<T, MemMan>(array+offset, m_pRefCount); }
#endif // _MT

protected:
	using FArray<T>::array;
#ifdef _MT
	dynarray(T* array, size_t* pRC, MUTEX mutex) : FArray<T>(array), m_pRefCount(pRC), mutex(mutex)
#else // _MT
	dynarray(T* array, size_t* pRC) : FArray<T>(array), m_pRefCount(pRC)
#endif // _MT
		{ IncRefCount(); }

	inline bool DecRefCount();
	inline void IncRefCount();

	size_t* m_pRefCount;
	DECLARE_MUTEX(mutex)
};


template<class T, class MemMan = HeapMemMan>
class SafeArray : public dynarray<T, MemMan>
{
	using dynarray<T,MemMan>::array;
	using dynarray<T,MemMan>::m_pRefCount;
public:
	SafeArray() : length(0) { }
	explicit SafeArray(size_t size) : dynarray<T, MemMan>(size), length(size) { }
	SafeArray(const SafeArray<T, MemMan>& src) : dynarray<T, MemMan>(src), length(src.length) { }
	SafeArray<T, MemMan>& operator=(SafeArray<T, MemMan>& rhs)
	{
		dynarray<T, MemMan>::operator=(rhs);
		if( this != &rhs && rhs.array )
			length = rhs.length;
		return *this;
	}

	T& operator[](size_t i)
		{ assert( i < length ); return array[i]; }
	const T& operator[](size_t i) const
		{ assert( i < length ); return array[i]; }
	SafeArray<T> operator+(size_t offset) const
#ifdef _MT
		{ assert( offset < length ); return SafeArray<T>(array+offset, m_pRefCount, length-offset, mutex); }
#else // _MT
		{ assert( offset < length ); return SafeArray<T>(array+offset, m_pRefCount, length-offset); }
#endif // _MT

	class iterator : public std::iterator<std::random_access_iterator_tag, T>
	{
		friend class SafeArray<T, MemMan>;
	public:
		iterator() : i(NULL), begin(NULL), end(NULL) { }
		iterator(const iterator& src) : i(src.i), begin(src.begin), end(src.end) { }

		iterator& operator=(const iterator& rhs)
			{ i = rhs.i; begin = rhs.begin; end = rhs.end; return *this; }
		iterator operator+(ptrdiff_t k) const
			{ return iterator(i+k, begin, end); }
		ptrdiff_t operator-(const iterator& it) const
			{ return i-it.i; }
		iterator operator-(ptrdiff_t k) const
			{ return iterator(i-k, begin, end); }
		iterator& operator+=(ptrdiff_t k)
			{ i += k; return *this; }
		iterator& operator-=(ptrdiff_t k)
			{ i -= k; return *this; }
		iterator& operator++()
			{ ++i; return *this; }
		iterator operator++(int)
			{ iterator ret=*this; i++; return ret; }
		iterator& operator--()
			{ --i; return *this; }
		iterator operator--(int)
			{ iterator ret=*this; i--; return ret; }
		bool operator<(const iterator& it) const
			{ return i < it.i; }
		bool operator>(const iterator& it) const
			{ return i > it.i; }
		bool operator<=(const iterator& it) const
			{ return i <= it.i; }
		bool operator>=(const iterator& it) const
			{ return i >= it.i; }
		bool operator==(const iterator& it) const
			{ return i == it.i; }
		bool operator!=(const iterator& it) const
			{ return i != it.i; }

		T& operator*()
			{ assert( i ); assert( begin <= i && i < end ); return *i; }
		const T& operator*() const
			{ assert( i ); assert( begin <= i && i < end ); return *i; }
		T* operator->()
			{ assert( begin <= i && i < end ); return i; }
		const T* operator->() const
			{ assert( i ); assert( begin <= i && i < end ); return i; }
	private:
		iterator(T* elt, T* begin, T* end) : i(elt), begin(begin), end(end) { }
		T* i;
		T* begin;
		T* end;
	};
	iterator begin()
		{ return iterator(array, array, array+length); }
	iterator end(size_t size)
		{ return iterator(array+size, array, array+length); }
//	typedef T* iterator;
protected:
#ifdef _MT
	SafeArray(T* a, size_t* pRC, size_t len, MUTEX mutex) : dynarray<T>(a, pRC, mutex)
#else // _MT
	SafeArray(T* a, size_t* pRC, size_t len) : dynarray<T>(a, pRC)
#endif // _MT
		{ length = len; }
	size_t length;
};


//	<summary>
//		Don't expect to be able to map to a file, which you don't have readaccess to
//		No read access and streaming implies streaming store type access
//		Read access and streaming implies streaming store type access
//		temporary provides no guaranties of changes 
//		temporary and create deletes the created file, when refcount becomes 0
//		temporary and existing may have wierd effects
// </summary>
enum map_construct_flags
{
	MAP_CREATE = 1,
	MAP_EXISTING = 3,
	MAP_READ = 4,
	MAP_WRITE = 8,
	MAP_TEMPORARY = 16,
	MAP_STREAMING = 32
};

template<class T>
class map_array : public dynarray<T>
{
	using dynarray<T>::array;
	using dynarray<T>::DecRefCount;
	using dynarray<T>::IncRefCount;
	using dynarray<T>::m_pRefCount;
public:
	inline map_array() { }
	inline map_array(const char* szImageName, int type=MAP_EXISTING | MAP_READ | MAP_WRITE, size_t size=0, uint64_t offset=0) : m_Size(size), m_Type(type)
	{
		offset *= sizeof(T);
		if( size )
			size *= sizeof(T);
		else
		{
			assert( (type&3) == MAP_EXISTING );
#ifdef WIN32
			HANDLE hFile = CreateFile(szImageName, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
			size = GetFileSize(hFile, NULL)-(size_t)offset;
			m_Size = size/sizeof(T);
			CloseHandle(hFile);
#else // WIN32
			struct stat s;
			stat(szImageName, &s);
			size = s.st_size-offset;
			m_Size = size/sizeof(T);
#endif // WIN32
		}
		INIT_MUTEX(mutex);
		DoMap(szImageName, (size_t)size, (size_t)offset, type);
	}

	inline map_array<T>& operator=(const map_array<T>& rhs)
	{
		if( this != &rhs )
		{
			if( array )
			{
				if( DecRefCount() )
				{
					DoUnMap();
					CLEANUP_MUTEX(mutex);
				}
			}
			if( rhs.array )
			{
				array = rhs.array;
				unmap_this = rhs.unmap_this;
				m_pRefCount = rhs.m_pRefCount;
				m_szImageName = rhs.m_szImageName;
#ifdef WIN32
				m_hMapping = rhs.m_hMapping;
#else // WIN32
				umsize = rhs.umsize;
#endif // WIN32
				m_Size = rhs.m_Size;
				COPY_MUTEX(mutex, rhs.mutex);
				IncRefCount();
			}
			else
				array = NULL;
		}
		return *this;
	}

	virtual ~map_array()
	{
		if( array )
			if( DecRefCount() )
			{
				DoUnMap();
				CLEANUP_MUTEX(mutex);
			}
		array = NULL; // To avoid the DynArray destuctor
	}

	inline bool operator==(const map_array<T>& rhs)
		{ return array == rhs.array; }

	inline map_array<T> operator+(size_t offset) const
#ifdef WIN32
#ifdef _MT
		{ return map_array<T>(array+offset, m_pRefCount, m_hMapping, unmap_this, m_szImageName, mutex); }
#else // _MT
		{ return map_array<T>(array+offset, m_pRefCount, m_hMapping, unmap_this, m_szImageName); }
#endif // _MT
#else // WIN32
#ifdef _MT
		{ return map_array<T>(array+offset, m_pRefCount, m_Size, unmap_this, m_szImageName, mutex); }
#else // _MT
		{ return map_array<T>(array+offset, m_pRefCount, m_Size, unmap_this, m_szImageName); }
#endif // _MT
#endif // WIN32

	class stream_iterator : public std::iterator<std::random_access_iterator_tag, T>
	{
		friend class map_array<T>;
	public:
		stream_iterator() : i(NULL) { }
		stream_iterator(const stream_iterator& src) : i(src.i) { }

		stream_iterator& operator=(const stream_iterator& rhs)
			{ i = rhs.i; return *this; }
		stream_iterator operator+(ptrdiff_t k) const
			{ return stream_iterator(i+k); }
		ptrdiff_t operator-(const stream_iterator& it) const
			{ return i-it.i; }
		stream_iterator operator-(ptrdiff_t k) const
			{ return stream_iterator(i-k); }
		stream_iterator& operator+=(ptrdiff_t k)
			{ i += k; return *this; }
		stream_iterator& operator-=(ptrdiff_t k)
			{ i -= k; return *this; }
		stream_iterator& operator++()
			{ inc(); return *this; }
		stream_iterator operator++(int)
			{ stream_iterator ret=*this; ret.inc(); return ret; }
		stream_iterator& operator--()
			{ dec(); return *this; }
		stream_iterator operator--(int)
			{ stream_iterator ret=*this; ret.dec(); return ret; }
		bool operator<(const stream_iterator& it) const
			{ return i < it.i; }
		bool operator>(const stream_iterator& it) const
			{ return i > it.i; }
		bool operator<=(const stream_iterator& it) const
			{ return i <= it.i; }
		bool operator>=(const stream_iterator& it) const
			{ return i >= it.i; }
		bool operator==(const stream_iterator& it) const
			{ return i == it.i; }
		bool operator!=(const stream_iterator& it) const
			{ return i != it.i; }

		T& operator*()
			{ return *i; }
		const T& operator*() const
			{ return *i; }
		T* operator->()
			{ return i; }
		const T* operator->() const
			{ return i; }
	private:
		stream_iterator(T* elt) : i(elt) { }
		T* i;
		inline void inc()
		{
			if( (reinterpret_cast<char*>(i)-reinterpret_cast<char*>(NULL)) % pagesize == 0 )
			{
#ifdef WIN32
#else // WIN32
//				std::cerr << pagesize;
				madvise(reinterpret_cast<char*>(i)+2*pagesize, pagesize, MADV_WILLNEED);
//					perror("MADV_WILLNEED");
//				msync(reinterpret_cast<char*>(i)-pagesize, pagesize, MS_SYNC/*MS_ASYNC | MS_INVALIDATE*/);
//					perror("MS_SYNC");
				madvise(reinterpret_cast<char*>(i)-pagesize, pagesize, MADV_DONTNEED);
//					perror("MADV_DONTNEED");
#endif // WIN32
			}
			++i;
		}
		inline void dec()
		{
			if( (reinterpret_cast<char*>(i)-reinterpret_cast<char*>(NULL)) & (pagesize-1) == 0 )
			{
#ifdef WIN32
#else // WIN32
				madvise(reinterpret_cast<char*>(i)-3*pagesize, pagesize, MADV_WILLNEED);
				msync(reinterpret_cast<char*>(i), pagesize, MS_ASYNC | MS_INVALIDATE);
//				madvise(reinterpret_cast<char*>(i), pagesize, MADV_DONTNEED);
#endif // WIN32
			}
			--i;
		}
#ifdef WIN32
#else // WIN32
		const static int pagesize = 4096; //getpagesize();
#endif // WIN32
	};

	inline stream_iterator begin_stream()
		{ return stream_iterator(array); }
	inline stream_iterator end_stream(size_t size)
		{ return stream_iterator(array+size); }

	inline T *end()
		{ return array+m_Size; }

protected:
	// Maps array/unmap_this to size bytes of the existing file named m_szImageName, with offset of offset bytes
	void DoMap(const char *szImageName, size_t size, uint64_t offset, int type);
	// Unmaps unmap_this
	void DoUnMap();

private:
#ifdef WIN32
	HANDLE m_hMapping;
	// Used as return value from the + operator
#ifdef _MT
	map_array(T* a, size_t* pRC, HANDLE hMap, char* u, char* imageName, MUTEX mutex) : dynarray<T>(a, pRC, mutex), m_hMapping(hMap), unmap_this(u), m_szImageName(imageName)
#else // _MT
	map_array(T* a, size_t* pRC, HANDLE hMap, char* u, char* imageName) : dynarray<T>(a, pRC), m_hMapping(hMap), unmap_this(u), m_szImageName(imageName)
#endif // _MT
#else // WIN32
	size_t umsize;
#ifdef _MT
	map_array(T* a, size_t* pRC, size_t size, char* u, char* imageName, MUTEX mutex) : dynarray<T>(a, pRC, mutex), m_Size(size), unmap_this(u), m_szImageName(imageName)
#else // _MT
	map_array(T* a, size_t* pRC, size_t size, char* u, char* imageName) : dynarray<T>(a, pRC), m_Size(size), unmap_this(u), m_szImageName(imageName)
#endif // _MT
#endif // WIN32
		{ }
	char* unmap_this;
	char* m_szImageName;
	size_t m_Size;
	int m_Type;
};

#ifdef WIN32
#else // WIN32
template<class T>
const int map_array<T>::stream_iterator::pagesize;
#endif // WIN32

template<class T>
class MultiMapArray
{
public:
	typedef uint64_t size_type;

	MultiMapArray() : m_szImageName(NULL) { }
	MultiMapArray(const char* szImageName, int type=MAP_EXISTING | MAP_READ | MAP_WRITE, uint64_t size=0, uint64_t offset=0) : m_Offset(offset), m_Type(type), m_szImageName(NULL)
	{
		if( (type&3) == MAP_CREATE )
		{
			assert( size );
#ifdef WIN32
			CloseHandle(CreateFile(szImageName, 0, 0, NULL, CREATE_NEW, (type&MAP_TEMPORARY? FILE_ATTRIBUTE_TEMPORARY : 0 ) | FILE_FLAG_RANDOM_ACCESS, NULL));
#else // WIN32
			int fd = open(szImageName, O_RDWR | O_CREAT | O_LARGEFILE, 0666);
			size_t ps = getpagesize();
			uint64_t fsize = ((sizeof(T)*size)/ps+1)*ps;
			ftruncate(fd, fsize);
			void* buf = malloc(ps);
			for( uint64_t i=0; i<fsize; i+=ps )
				write(fd, buf, ps);
			free(buf);
			close(fd);
#endif // WIN32
		}
		else
		{
			assert( (type&3) == MAP_EXISTING );
			if( size == 0 )
			{
#ifdef WIN32
				LARGE_INTEGER li;
				HANDLE hFile = CreateFile(szImageName, 0, 0, NULL, OPEN_EXISTING, 0, NULL);
				GetFileSizeEx(hFile, &li);
				// TODO: Supports 64bit?
				size = li.u.LowPart/sizeof(T)-offset;
				CloseHandle(hFile);
#else // WIN32
				struct stat s;
				stat(szImageName, &s);
				// TODO: Support 64bit
				size = s.st_size/sizeof(T)-offset;
#endif // WIN32
			}
		}
		size_t slen = strlen(szImageName);
		m_szImageName = new char[slen+1];
		strcpy(m_szImageName, szImageName);
		INIT_MUTEX(m_mRC);
		INIT_MUTEX(m_mTable);
		createTables(size+m_Offset);
	}

	MultiMapArray(const MultiMapArray& src)
	{
		if( src.m_szImageName )
		{
			m_Filter = src.m_Filter;
			m_Offset = src.m_Offset;
			m_pRefCount = src.m_pRefCount;
			m_pTime = src.m_pTime;
			m_Split = src.m_Split;
			m_szImageName = src.m_szImageName;
			m_Table = src.m_Table;
			m_TableSize = src.m_TableSize;
			m_Type = src.m_Type;
			COPY_MUTEX(m_mTable, src.m_mTable);
			COPY_MUTEX(m_mRC, src.m_mRC);
			IncRefCount();
		}
		else
			m_szImageName = NULL;
	}

	MultiMapArray& operator=(const MultiMapArray& rhs)
	{
		if( this != &rhs )
		{
			if( m_szImageName )
			{
				if( DecRefCount() )
				{
					delete[] m_Table;
					if( (m_Type&3) == MAP_CREATE && m_Type & MAP_TEMPORARY )
					{
#ifdef WIN32
						DeleteFile(m_szImageName);
#else // WIN32
						remove(m_szImageName);
#endif // WIN32
					}
					delete[] m_szImageName;
					delete m_pTime;
					delete m_pRefCount;
					CLEANUP_MUTEX(m_mTable);
					CLEANUP_MUTEX(m_mRC);
				}
			}
			if( rhs.m_szImageName )
			{
				m_Filter = rhs.m_Filter;
				m_Offset = rhs.m_Offset;
				m_pRefCount = rhs.m_pRefCount;
				m_pTime = rhs.m_pTime;
				m_Split = rhs.m_Split;
				m_szImageName = rhs.m_szImageName;
				m_Table = rhs.m_Table;
				m_TableSize = rhs.m_TableSize;
				m_Type = rhs.m_Type;
				COPY_MUTEX(m_mTable, rhs.m_mTable);
				COPY_MUTEX(m_mRC, rhs.m_mRC);
				IncRefCount();
			}
			else
				m_szImageName = NULL;
		}
		return *this;
	}

	virtual ~MultiMapArray()
	{
		if( DecRefCount() )
		{
			delete[] m_Table;
			if( (m_Type&3) == MAP_CREATE && m_Type & MAP_TEMPORARY )
			{
#ifdef WIN32
				if( DeleteFile(m_szImageName) == 0 )
					std::cerr << "[MultiMapArray warning: Could't remove created temporary file '" << m_szImageName << "']" << std::flush;
#else // WIN32
				remove(m_szImageName);
#endif // WIN32
			}
			delete[] m_szImageName;
			delete m_pTime;
			delete m_pRefCount;
			CLEANUP_MUTEX(m_mTable);
			CLEANUP_MUTEX(m_mRC);
		}
	}

	inline T& operator[](uint64_t i)
		{ assert( m_szImageName ); size_t page = (size_t)((i+m_Offset)>>m_Split); TouchPage(page); T& ret = m_Table[page].Page[(size_t)(i+m_Offset) & m_Filter]; return ret; }

	inline MultiMapArray<T> operator+(int64_t offset) const
		{ assert( m_szImageName ); MultiMapArray<T> a(*this); a.m_Offset += offset; return a; }

	class MMAIt : public std::iterator<std::random_access_iterator_tag, T, int64_t>
	{
		friend class MultiMapArray<T>;
	public:
		inline MMAIt() : i(0), a(NULL)
			{ }
		inline MMAIt operator+(int64_t k) const
			{ return MMAIt(i+k, a); }
		inline MMAIt operator-(int64_t k) const
			{ return MMAIt(i-k, a); }
		inline int64_t operator-(const MMAIt& it) const
			{ assert( a == it.a ); return i-it.i; }
		inline MMAIt operator+=(int64_t k)
			{ i += k; return *this; }
		inline MMAIt operator-=(int64_t k)
			{ i -= k; return *this; }
		inline MMAIt operator++()
			{ i++; return *this; }
		inline MMAIt operator++(int)
			{ MMAIt ret=*this; i++; return ret; }
		inline MMAIt operator--()
			{ i--; return *this; }
		inline MMAIt operator--(int)
			{ MMAIt ret=*this; i--; return ret; }
		inline bool operator<(const MMAIt& it) const
			{ return (a == it.a) && (i < it.i); }
		inline bool operator>(const MMAIt& it)const
			{ return (a == it.a) && (i > it.i); }
		inline bool operator<=(const MMAIt& it)const
			{ return (a == it.a) && (i <= it.i); }
		inline bool operator>=(const MMAIt& it)const
			{ return (a == it.a) && (i >= it.i); }
		inline bool operator==(const MMAIt& it)const
			{ return (a == it.a) && (i == it.i); }
		inline bool operator!=(const MMAIt& it)const
			{ return (a != it.a) || (i != it.i); }
		inline T& operator*()
		{
			size_t page = (size_t)((i+a->m_Offset)>>a->m_Split);
			a->TouchPage(page);
			T& ret = a->m_Table[page].Page[(size_t)(i+a->m_Offset) & a->m_Filter];
			return ret;
		}
	private:
		MMAIt(uint64_t i, MultiMapArray<T>* a) : i(i), a(a)
			{ }
		uint64_t i;
		MultiMapArray<T>* a;
	};

	inline MMAIt begin()
		{ return MMAIt(0, this); }
	inline MMAIt end(uint64_t e)
		{ return MMAIt(e, this); }

	typedef MMAIt iterator;

#ifdef MAPPINGSTAT
	static uint64_t PageFaults;
#endif // MAPPINGSTAT
private:
	static const size_t nActivePages = 16;

	void createTables(uint64_t size)
	{
		int AddrLen = 0;
		for( uint64_t s=size-1; s; s>>=1 )
			AddrLen++;
		m_Split = (AddrLen+1)/2;

		m_Split += 3;

		m_Filter = (1<<m_Split)-1;
		m_TableSize = (size_t)((size-1)>>m_Split)+1;
		m_Table = new TableEntry[m_TableSize];
		m_pTime = new uint64_t (0);
		m_pRefCount = new size_t(1);
		for( size_t page=0; page<nActivePages && page<m_TableSize; page++ )
			fetchPage(page);
	}
	inline void TouchPage(size_t page)
	{
		assert( m_szImageName ); // Accessing a destroyed array?
		assert( page < m_TableSize ); // Index too large?
		if( m_Table[page].Time )
		{
			// Only update time, if it's not the MRU
			LOCK_MUTEX(m_mTable);
			if( m_Table[page].Time < *m_pTime )
			{
				// Assume m_Time doesn't overflow
				*m_pTime = *m_pTime+1;
				m_Table[page].Time = *m_pTime;
			}
			UNLOCK_MUTEX(m_mTable);
		}
		else
			PageFault(page);
	}
	void PageFault(size_t page)
	{
		// Find LRU
		LOCK_MUTEX(m_mTable);
#ifdef MAPPINGSTAT
		PageFaults++;
#endif // MAPPINGSTAT
		size_t i = 0;
		//  Find first mapped page
		while( m_Table[i].Time == 0 )
			 i++;
		size_t LRU = i;
		for( ; i<m_TableSize; i++ )
			if( m_Table[i].Time && m_Table[i].Time < m_Table[LRU].Time )
				LRU = i;
#ifdef _DEBUG
//		std::cout << (unsigned long)LRU << "<->" << (unsigned long)page << std::endl;
#endif // _DEBUG
		// Evict page
		m_Table[LRU].Page = map_array<T>();
		m_Table[LRU].Time = 0;
		// Get page
		fetchPage(page);
		UNLOCK_MUTEX(m_mTable);
	}
	inline void fetchPage(size_t page)
	{
		m_Table[page].Page = map_array<T>(m_szImageName, MAP_EXISTING | m_Type & (MAP_TEMPORARY | MAP_READ | MAP_WRITE), m_Filter+1, (m_Filter+1)*page+m_Offset);
		(*m_pTime)++;
		m_Table[page].Time = *m_pTime;
	}

	uint64_t m_Offset;
	size_t m_Filter;
	int m_Split;
	int m_Type;

	inline bool DecRefCount()
	{
		if( m_szImageName )
		{
			LOCK_MUTEX(m_mRC);
			(*m_pRefCount)--;
			UNLOCK_MUTEX(m_mRC);
			return *m_pRefCount == 0;
		}
		return false;
	}
	inline void IncRefCount()
	{
		assert( m_szImageName );
		LOCK_MUTEX(m_mRC);
		(*m_pRefCount)++;
		UNLOCK_MUTEX(m_mRC);
	}

	class TableEntry
	{
	public:
		inline TableEntry() : Time(0) { }
		map_array<T> Page;
		uint64_t Time;
	};
	TableEntry* m_Table;

	size_t m_TableSize;
	char* m_szImageName;
	uint64_t* m_pTime;

	size_t* m_pRefCount;
	DECLARE_MUTEX(m_mTable)
	DECLARE_MUTEX(m_mRC)
};


#ifdef MAPPINGSTAT
template<class T>
uint64_t MultiMapArray<T>::PageFaults;
template<class T>
uint64_t MapArray<T>::mappingcount = 0;
#endif // MAPPINGSTAT

template<class T>
class SeqMapArray
{
public:
	typedef uint64_t size_type;

	SeqMapArray() : m_Begin(0), m_End(0), delete_file(false) { }

	SeqMapArray(uint64_t size, const char* szImageName) : m_Begin(0), m_Offset(0), m_Size(size), m_szImageName(szImageName), delete_file(true)
	{
#ifdef WIN32
		CloseHandle(CreateFile(szImageName, GENERIC_READ | GENERIC_WRITE, 0, NULL, CREATE_NEW, FILE_FLAG_SEQUENTIAL_SCAN, NULL));
#else
		int fd = open(m_szImageName, O_RDWR | O_CREAT | O_LARGEFILE, 0666);
		size_t ps = getpagesize();
		size = ((sizeof(T)*size)/ps+1)*ps;
		ftruncate(fd, size); // HACK: We should realy fail instead, if the file exists
		void* buf = malloc(ps);
		for( size_t i=0; i<size; i+=ps )
			write(fd, buf, ps);
		free(buf);
		close(fd);
#endif // WIN32
		int AddrLen = 0;
		for( uint64_t s=size-1; s; s>>=1 )
			AddrLen++;
		m_MapSize = (1<<((AddrLen+1)/2))-1;
		ReMap(0);
	}

	SeqMapArray(const char* szImageName, uint64_t offset=0, uint64_t size=0) : m_Begin(0), m_Offset(offset), m_szImageName(szImageName), delete_file(false)
	{
		// TODO: get size, if size == 0
		m_Size = size;
		int AddrLen = 0;
		for( uint64_t s=size-1; s; s>>=1 )
			AddrLen++;
		m_MapSize = (1<<((AddrLen+1)/2))-1;
//		m_Map.ReMap(0);
	}

	virtual ~SeqMapArray()
	{
		if( delete_file )
		{
#ifdef WIN32
			DeleteFile(m_szImageName);
#else // WIN32
			remove(m_szImageName);
#endif // WIN32
		}
	}

	inline T& operator[](uint64_t i)
		{ if( i>=m_End || i<m_Begin ) ReMap(i); return m_Map[(size_t)(i-m_Begin)]; }

	class SMAIt : public std::iterator<std::random_access_iterator_tag, T, int64_t>
	{
		friend class SeqMapArray<T>;
	public:
		SMAIt() : i(0), a(NULL)
			{ }
		SMAIt operator+(int64_t k) const
			{ return SMAIt(i+k, a); }
		SMAIt operator++()
			{ i++; return *this; }
		SMAIt operator++(int)
			{ ++i; return *this; }
		SMAIt operator+=(int64_t k)
			{ i += k; return *this; }
		SMAIt operator-(int64_t k) const
			{ return SMAIt(i-k, a); }
		SMAIt operator--()
			{ i--; return *this; }
		SMAIt operator--(int)
			{ --i; return *this; }
		SMAIt operator-=(int64_t k)
			{ i -= k; return *this; }
		int64_t operator-(const SMAIt& it) const
			{ assert( a == it.a ); return i-it.i; }
		bool operator<(const SMAIt& it) const
			{ return (a == it.a) && (i < it.i); }
		bool operator>(const SMAIt& it)const
			{ return (a == it.a) && (i > it.i); }
		bool operator<=(const SMAIt& it)const
			{ return (a == it.a) && (i <= it.i); }
		bool operator>=(const SMAIt& it)const
			{ return (a == it.a) && (i >= it.i); }
		bool operator==(const SMAIt& it)const
			{ return (a == it.a) && (i == it.i); }
		bool operator!=(const SMAIt& it)const
			{ return (a != it.a) || (i != it.i); }
		T& operator*()
		{
			return a->operator[](i);
		}
	private:
		SMAIt(uint64_t i, SeqMapArray<T>* a) : i(i), a(a)
			{ }
		uint64_t i;
		SeqMapArray<T>* a;
	};

	SMAIt begin()
		{ return SMAIt(0, this); }
	SMAIt end(uint64_t e)
		{ return SMAIt(e, this); }

	typedef SMAIt iterator;

protected:
	void ReMap(uint64_t i)
	{
		m_Begin = i;
		if( i+m_MapSize < m_Size )
		{
//			m_Map.remap(m_szImageName, m_Offset+m_Begin, m_MapSize);
			m_End = i+m_MapSize;
		}
		else
		{
//			m_Map.remap(m_szImageName, m_Offset+m_Begin, m_Size-i);
			m_End = m_Size;
		}
	}

	uint64_t m_Begin, m_End;
	uint64_t m_Offset, m_Size;
	size_t m_MapSize;
	map_array<T> m_Map;
	const char* m_szImageName;
	bool delete_file;
};

} // namespace auto_array

#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#include "aarray.timpl.h"

#endif // __FARRAY_H_INCLUDED_
