/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __ELEMENTS_H_INCLUDED_
#define __ELEMENTS_H_INCLUDED_

#include <ostream>
#include <stdexcept>
#include <limits>

template<class T>
class factory;

class element
{
	friend class factory<element>;
	friend std::ostream& operator<<(std::ostream& os, const element& e);
public:
	element() : key(), ptr() { };
	element(unsigned long int key, void* ptr) : key(key), ptr(ptr) { };
	inline bool operator<(const element& rhs) const
		{ return key<rhs.key; }
	inline bool operator>(const element& rhs) const
		{ return key>rhs.key; }
	inline bool operator!=(const element& rhs) const
		{ return key!=rhs.key; }
	inline bool operator==(const element& rhs) const
		{ return key==rhs.key; }

	inline element operator+=(const element& rhs)
		{ key+=rhs.key; return *this; }
private:
	unsigned long int key;
	void* ptr;
};
std::ostream& operator<<(std::ostream& os, const element& e)
{
	return os << '(' << e.key << ',' << e.ptr << ')';
}
namespace std
{
template<>
class std::numeric_limits<element>
{
public:
	inline static element min()
	{
		return element(std::numeric_limits<unsigned long int>::min(), static_cast<char*>(NULL)+0xF00DF00D);
	}
	inline static element max()
	{
		return element(std::numeric_limits<unsigned long int>::max(), static_cast<char*>(NULL)+0xF00DF00D);
	}
};
} // namespace std

#include <cstdlib> // for strncmp

class record
{
	friend class factory<record>;
	friend std::ostream& operator<<(std::ostream& os, const record& e);
public:
	inline bool operator<(const record& rhs) const
		{ return strncmp(rec,rhs.rec,record_size) < 0; }
	inline bool operator>(const record& rhs) const
		{ return strncmp(rec,rhs.rec,record_size) > 0; }
	inline bool operator!=(const record& rhs) const
		{ return strncmp(rec,rhs.rec,record_size) != 0; }
	inline bool operator==(const record& rhs) const
		{ return strncmp(rec,rhs.rec,record_size) == 0; }

	inline record operator+=(const record& rhs)
	{
		for( int i=0; i!=record_size; ++i )
			rec[i] += rhs.rec[i];
		return *this;
	}

	inline char& operator[](int i)
		{ return rec[i]; }
private:
	enum tag_size { record_size = 100 };
	char rec[record_size];
};
std::ostream& operator<<(std::ostream& os, const record& e)
{
	os << '[';
	for( int i=0; i!=record::record_size; ++i )
		os << e.rec[i];
	return os << ']';
}

namespace std
{
template<>
class numeric_limits<record>
{
public:
	inline static record max()
	{
		record ret;
		for( int i=0; i!=100; ++i )
			ret[i] = std::numeric_limits<char>::max();
		return ret;
	}
	inline static record min()
	{
		record ret;
		for( int i=0; i!=100; ++i )
			ret[i] = std::numeric_limits<char>::min();
		return ret;
	}
};
} // namespace std

#include <cmath>

template<class T>
class factory
{
public:
	inline static T from_scalar(double d) throw( std::invalid_argument )
	{
		if( d < 0.0 || d > 1.0 )
			throw std::invalid_argument("scalar must be positive and less than one");
		return static_cast<T>(d*(std::numeric_limits<T>::max()-std::numeric_limits<T>::min()))+std::numeric_limits<T>::min();
	}
	inline static void increment(T& t)
		{ ++t; }
};

template<>
class factory<element>
{
public:
	inline static element from_scalar(double d) throw( std::invalid_argument )
		{ return element(factory<unsigned long int>::from_scalar(d), static_cast<char*>(NULL)+0xBAADF00D); }
	inline static void increment(element& t)
		{ ++t.key; }
};

template<>
class factory<record>
{
public:
	inline static record from_scalar(double d) throw( std::invalid_argument )
	{
		record ret;
		if( d < 0.0 || d > 1.0 )
			throw std::invalid_argument("scalar must be positive and less than one");
		for( int i=0; i!=record::record_size; ++i )
		{
			d *= std::numeric_limits<char>::max()-std::numeric_limits<char>::min();
			ret.rec[i] = static_cast<char>(d+std::numeric_limits<char>::min());
			d -= floor(d);
		}
		return ret;
	}
	inline static void increment(record& t)
	{
		int i = record::record_size-1;
		for( ++t.rec[i]; t.rec[i]==std::numeric_limits<char>::min() && i; --i );
	}
};

#endif // __ELEMENTS_H_INCLUDED_
