/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

namespace auto_array
{

template<class T, class MemMan>
inline DynArray<T,MemMan>::DynArray(const DynArray<T, MemMan>& src) : FArray<T>(src)
{
	if( array )
	{
		m_pRefCount = src.m_pRefCount;
		COPY_MUTEX(mutex, src.mutex);
		IncRefCount();
	}
}

template<class T, class MemMan>
inline DynArray<T,MemMan>::~DynArray()
{
	if( array )
	{
		if( DecRefCount() )
		{
			MemMan::DAlloca(array);
			MemMan::DAlloc(m_pRefCount);
			CLEANUP_MUTEX(mutex);
		}
	}
}

template<class T, class MemMan>
inline DynArray<T, MemMan>& DynArray<T, MemMan>::operator=(DynArray<T, MemMan>& rhs)
{
	if( this != &rhs )
	{
		if( array )
		{
			if( DecRefCount() )
			{
				MemMan::DAlloca(array);
				MemMan::DAlloc(m_pRefCount);
				CLEANUP_MUTEX(mutex);
			}
		}
		if( rhs.array )
		{
			array = rhs.array;
			m_pRefCount = rhs.m_pRefCount;
			COPY_MUTEX(mutex, rhs.mutex);
			IncRefCount();
		}
		else
			array = NULL;
	}
	return *this;
}

template<class T, class MemMan>
inline bool DynArray<T, MemMan>::DecRefCount()
{
	if( array )
	{
		LOCK_MUTEX(m_mRC);
		(*m_pRefCount)--;
		UNLOCK_MUTEX(m_mRC);
		return *m_pRefCount == 0;
	}
	return false;
}

template<class T, class MemMan>
inline void DynArray<T, MemMan>::IncRefCount()
{
	assert( array );
	LOCK_MUTEX(m_mRC);
	(*m_pRefCount)++;
	UNLOCK_MUTEX(m_mRC);
}


#ifdef WIN32

template<class T>
void MapArray<T>::DoMap(const char* szImageName, size_t size, _uint64_t offset, int type)
{
	assert( szImageName );
	assert( array == NULL );

	HANDLE hFile;
	switch( type & 3 )
	{
	case MAP_CREATE:
		assert( size );
		assert( offset == 0 );
		hFile = CreateFile(szImageName, (type&MAP_READ? GENERIC_READ : 0) | (type&MAP_WRITE? GENERIC_WRITE : 0), 0, NULL, OPEN_ALWAYS, (type&MAP_TEMPORARY? FILE_ATTRIBUTE_TEMPORARY | FILE_FLAG_DELETE_ON_CLOSE : 0 ) | (type&MAP_STREAMING? FILE_FLAG_SEQUENTIAL_SCAN : FILE_FLAG_RANDOM_ACCESS ) | (type&MAP_STREAMING && !(type&MAP_READ)? FILE_FLAG_OVERLAPPED | FILE_FLAG_NO_BUFFERING : 0), NULL);
		break;
	case MAP_EXISTING:
		hFile = CreateFile(szImageName, (type&MAP_READ? GENERIC_READ : 0) | (type&MAP_WRITE? GENERIC_WRITE : 0), 0, NULL, OPEN_EXISTING, (type&MAP_TEMPORARY? FILE_ATTRIBUTE_TEMPORARY : 0 ) | (type&MAP_STREAMING? FILE_FLAG_SEQUENTIAL_SCAN : FILE_FLAG_RANDOM_ACCESS ) | (type&MAP_STREAMING && !(type&MAP_READ)? FILE_FLAG_OVERLAPPED | FILE_FLAG_NO_BUFFERING : 0), NULL);
		break;
	default:
		hFile = 0;
		assert( false );
	}

	m_hMapping = CreateFileMapping(hFile, NULL, (type&MAP_WRITE? PAGE_READWRITE : PAGE_READONLY), (DWORD)((size+offset)>>32), (DWORD)(size+offset), NULL);
	CloseHandle(hFile);

	SYSTEM_INFO si;
	GetSystemInfo(&si);
	DWORD ag = si.dwAllocationGranularity;
	_uint64_t foffset = offset/ag*ag;
	unmap_this = (char*)MapViewOfFile(m_hMapping, (type&MAP_WRITE? FILE_MAP_ALL_ACCESS : FILE_MAP_READ), (DWORD)(foffset>>32), (DWORD)foffset, (DWORD)(size+offset-foffset));
	array = (T*)(unmap_this+(offset-foffset));

	if( unmap_this == NULL )
	{
		std::cerr << "Fatal error: unable to create mapping of file '" << szImageName << "'!" << std::endl;
#ifdef MAPPINGSTAT
		std::cerr << "Mapping count: " << (unsigned int)mappingcount << std::endl;
#endif // MAPPINGSTAT
		abort();
	}
	this->m_pRefCount = new size_t(1);
/*	if( (type&3) == MAP_CREATE && (type&MAP_TEMPORARY) )
	{
		size_t slen = strlen(szImageName);
		m_szImageName = new char[slen+1];
		strcpy(m_szImageName, szImageName);
	}
	else
		m_szImageName = NULL;
*/
#ifdef MAPPINGSTAT
	mappingcount++;
#endif // MAPPINGSTAT
}

template<class T>
void MapArray<T>::DoUnMap()
{
	assert( *m_pRefCount == 0 );
	delete m_pRefCount;
#ifdef MAPPINGSTAT
	mappingcount--;
#endif // MAPPINGSTAT

	UnmapViewOfFile(unmap_this);
	CloseHandle(m_hMapping);
/*	if( m_szImageName )
	{
		if( DeleteFile(m_szImageName) == 0 )
			std::cerr << "[MapArray warning: Could't remove created temporary file '" << m_szImageName << "']" << std::flush;
		delete[] m_szImageName;
	}
*/}

#else // WIN32

template<class T>
void MapArray<T>::DoMap(const char* szImageName, size_t size, _uint64_t offset, int type)
{
	assert( szImageName );
	assert( array == NULL );

	int fd = open(szImageName, ((type&3) == MAP_CREATE? O_CREAT : 0) | O_RDWR | O_LARGEFILE, 0666);
	if( fd == -1 )
		unmap_this = NULL;
	else
	{
		size_t ps = getpagesize();
		_uint64_t foffset = offset/ps*ps;
		size = ((size+(offset-foffset))/ps+1)*ps;
#ifdef MAP_AUTOGROW
		unmap_this = (char*)mmap(NULL, size, (type&MAP_READ? PROT_READ : 0) | (type&MAP_WRITE? PROT_WRITE : 0), MAP_SHARED | MAP_AUTOGROW, fd, foffset);
#else // MAP_AUTOGROW
		if( (type&3) == MAP_CREATE )
		{
			assert( size );
			assert( offset == 0 );
			ftruncate(fd, size);
/*			void* buf = malloc(ps);
			for( size_t i=0; i<size; i+=ps )
				if( write(fd, buf, ps) < (int)ps )
				{
					std::cerr << "Fatal error: unable to create file '" << szImageName << "'!" << std::endl;
					close(fd);
					unlink(szImageName);
					abort();
				}
			free(buf);
*/		}
		unmap_this = (char*)mmap(NULL, size, (type&MAP_READ? PROT_READ : 0) | (type&MAP_WRITE? PROT_WRITE : 0), MAP_SHARED, fd, foffset);
#endif // MAP_AUTOGROW
		m_Size = size;
		array = (T*)(unmap_this+(offset-foffset));
		close(fd);
		if( unmap_this == MAP_FAILED )
			unmap_this = NULL;
	}

	if( unmap_this == NULL )
	{
		std::cerr << "Fatal error: unable to create mapping of file '" << szImageName << "'!" << std::endl;
#ifdef MAPPINGSTAT
		std::cerr << "Mapping count: " << (unsigned int)mappingcount << std::endl;
#endif // MAPPINGSTAT
		unlink(szImageName);
		abort();
	}
//	if( type & MAP_STREAMING )
//		madvise(unmap_this, size, MADV_SEQUENTIAL);
	m_pRefCount = new size_t(1);
	if( (type&3) == MAP_CREATE && (type&MAP_TEMPORARY) )
	{
		size_t slen = strlen(szImageName);
		m_szImageName = new char[slen+1];
		strcpy(m_szImageName, szImageName);
	}
	else
		m_szImageName = NULL;
#ifdef MAPPINGSTAT
	mappingcount++;
#endif // MAPPINGSTAT
}

template<class T>
void MapArray<T>::DoUnMap()
{
	assert( *m_pRefCount == 0 );
	delete m_pRefCount;
#ifdef MAPPINGSTAT
	mappingcount--;
#endif // MAPPINGSTAT

	if( (m_Type & MAP_TEMPORARY) && !(m_Type & MAP_EXISTING) )
		madvise(unmap_this, m_Size, MADV_DONTNEED);
	munmap(unmap_this, m_Size);
	if( m_szImageName )
	{
		unlink(m_szImageName);
		delete[] m_szImageName;
	}
}

#endif // WIN32

} // namespace auto_array
