/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __FMERGER_H_INCLUDED_
#define __FMERGER_H_INCLUDED_

#include <cmath>
#include "BasicMerger.h"
#include "FStream.h"


template<class T, class InStreamIt, class MemMan>
class Funnel
{
private:
	typedef FBuffer<T, DynArray<T, MemMan> > _Buffer;
	typedef _uint64_t size_type;

public:
	explicit Funnel(size_t k) : m_bEmpty(false), k(k)
	{
		assert( k > 1 ); // Punish the ones who want to merge a single stream
		if( k < 4 )
			for( size_t i=0; i<k; i++ )
				u.S[i] = in+i;
		else
		{
			const size_t f = (size_t)floor(sqrt((double)k));
			const size_t c = (size_t)ceil(sqrt((double)k));

			u.r.L = MemMan::template Allocap<Funnel<T, InStreamIt, MemMan>* >(c);
			u.r.B = MemMan::template Alloca<_Buffer>(c);

			size_t kp = k;
			u.r.l = 0;
			while( kp > 0 && !(kp%f == 0 && kp/f+u.r.l<=c) )
			{
				u.r.L[u.r.l] = MemMan::template Alloc<Funnel<T, InStreamIt, MemMan> >(c, in+(k-kp));
				kp -= c;
				u.r.l++;
			}
			while( kp > 0 )
			{
				u.r.L[u.r.l] = MemMan::template Alloc<Funnel<T, InStreamIt, MemMan> >(f, in+(k-kp));
				kp -= f;
				u.r.l++;
			}
			assert( kp == 0 );

			// Construct the buffers
			for( size_t i=0; i<u.r.l; i++ )
				if( u.r.L[i]->k > u.r.l )
				{
					_Buffer b(2*u.r.L[i]->k*u.r.L[i]->k*u.r.L[i]->k);
					u.r.B[i] = b;
				}
				else
				{
					_Buffer b(2*u.r.l*u.r.l*u.r.l);
					u.r.B[i] = b;
				}
			// Construct the head merger
			u.r.R = MemMan::template Alloc<Funnel<T, _Buffer, MemMan> >(u.r.l, u.r.B);
		}
	}
	virtual ~Funnel()
	{
		if( k > 3 )
		{
			MemMan::DAlloc(u.r.R);
			for( size_t i=u.r.l; i>0; i-- )
				MemMan::DAlloc(u.r.L[i-1]);
			MemMan::DAlloca(u.r.B);
			MemMan::DAllocap(u.r.L);
		}
	}

	template<class OutStreamIt>
	inline OutStreamIt invoke(OutStreamIt outbegin, OutStreamIt outend)
	{
		m_bEmpty = false;
		if( k == 2 )
			outbegin = LazyMerge(u.S->first, u.S->second, (u.S+1)->first, (u.S+1)->second, outbegin, outend);
		else if( k == 3 )
			outbegin = LazyMerge(u.S->first, u.S->second, (u.S+1)->first, (u.S+1)->second, (u.S+2)->first, (u.S+2)->second, outbegin, outend);
		else
			outbegin = rec_invoke(outbegin, outend);
		// If we didn't deliver, we must've run dry
		m_bEmpty = outbegin < outend;
		return outbegin;
	}

	inline bool empty() const
		{ return m_bEmpty; }
	inline size_t degree() const
		{ return k; }

	static size_t size(const size_t k)
	{
		assert( k > 1 ); // Punish the ones who want to merge a single stream
		size_t s = sizeof(Funnel<T, InStreamIt, MemMan>);
		if( k < 4 )
			return s;
		else
		{
			const size_t f = (size_t)floor(sqrt((double)k));
			const size_t c = (size_t)ceil(sqrt((double)k));
			size_t* ks = new size_t[c];
			s += sizeof(Funnel<T, InStreamIt, MemMan>*)*c;
			s += sizeof(_Buffer)*c;
			size_t kp = k;
			size_t l = 0;
			while( kp > 0 && !(kp%f == 0 && kp/f+l<=c) )
			{
				ks[l] = c;
				s += size(c);
				kp -= c;
				l++;
			}
			while( kp > 0 )
			{
				ks[l] = f;
				s += size(f);
				kp -= f;
				l++;
			}
			assert( kp == 0 );
			for( size_t i=0; i<l; i++ )
				if( ks[i] > l )
					s += sizeof(T)*((2*ks[i]*ks[i]*ks[i])+1)+sizeof(size_t);
				else
					s += sizeof(T)*((2*l*l*l)+1)+sizeof(size_t);
			s += size(l);
			delete[] ks;
			return s;
		}
	}

protected:
	template<class OutStreamIt>
	OutStreamIt rec_invoke(OutStreamIt outbegin, OutStreamIt outend)
	{
		size_type count = 0;
		// Keep invoking R until we've run dry, or N elements has been output
		size_t k3 = u.r.R->degree();
		k3 = k3*k3*k3;
		do
		{
			// Make sure the buffers are (half) filled
			for( size_t i=0; i<u.r.l; i++ )
			{
				size_type halfSize = u.r.B[i].getSize()/2;
				if( u.r.B[i].elementCount() < halfSize && !u.r.L[i]->empty() )
					u.r.L[i]->invoke(u.r.B[i].insertion_iterator(), u.r.B[i].insertion_iterator()+halfSize);
			}
			// If there're still elements left, invoke the head
			if( N-count < k3 ) // Don't extract too much, though
				count += u.r.R->invoke(outbegin+count, outend);
			else
				count += u.r.R->invoke(outbegin+count, outbegin+k3);
		}
		while( count<N && !u.r.R->empty() );
		return outbegin+count;
	}

	union tagUnion
	{
		struct tagRecursive
		{
			Funnel<T, InStreamIt, MemMan>** L;
			_Buffer* B;
			Funnel<T, _Buffer, MemMan>* R;
			size_t l;
		} r;
		InStreamIt S[3];
	} u;

	bool m_bEmpty;
	size_t k;
};




template<class T, class InStreamIt, class MergerTree>
class FMerger
{
protected:
	class FMergerCache
	{
	public:
		explicit FMergerCache(size_t maxorder) : cache(new FMerger*[maxorder+1]), size(maxorder+1), oversized(NULL)
		{
			for( size_t i=0; i<size; i++ )
				cache[i] = NULL;
		}
		virtual ~FMergerCache()
		{
			for( size_t i=0; i<size; i++ )
				delete cache[i];
			delete[] cache;
			delete oversized;
		}
		FMerger* get_merger(size_t order)
		{
			if( order < size )
				if( cache[order] )
					return cache[order];
				else
				{
					cache[order] = new FMerger(order);
					return cache[order];
				}
			else
			{
				delete oversized;
				oversized = new FMerger(order);
				return oversized;
			}
		}
	private:
		FMerger** cache;
		size_t size;
		FMerger* oversized;
	};

public:
	typedef FMergerCache Cache;

	explicit FMerger(size_t k) : k(k), m_MergerTree(k)
		{ }
	explicit FMerger(InStreamIt begin, InStreamIt end) : k(end-begin)
	{
		ITERATORS_MUST_POINT_TO_PAIRS_OF_ITERATORS_OF_SAME_TYPE(*begin);
	}
	void set_input_streams(InStreamIt b, InStreamIt e)
	{
		ITERATORS_MUST_POINT_TO_PAIRS_OF_ITERATORS_OF_SAME_TYPE(*begin);
	}

	template<class OutStreamIt>
	inline OutStreamIt invoke(OutStreamIt outbegin, OutStreamIt outend)
	{
		return m_MergerTree.invoke();
	}

protected:
	size_t k;
	MergerTree m_MergerTree;
};
























/*

template<class T, class InStreamClass, class MemMan = HeapMemMan>
class FMerger
{
private:
	typedef FBuffer<T, DynArray<T, MemMan> > _Buffer;
	typedef _uint64_t size_type;

public:
	FMerger()
		{ }
	// Constructs a kMerger, that merges k streams in the array in
	FMerger(const size_t k, InStreamClass* in) : k(k)
	{
		assert( k > 1 ); // Punish the ones who want to merge a single stream
		m_bEmpty = false; // assumption
		if( k < 4 )
			for( size_t i=0; i<k; i++ )
				u.S[i] = in+i;
		else
		{
			const size_t f = (size_t)floor(sqrt((double)k));
			const size_t c = (size_t)ceil(sqrt((double)k));

			u.r.L = MemMan::template Allocap<FMerger<T, InStreamClass, MemMan>* >(c);
			u.r.B = MemMan::template Alloca<_Buffer>(c);

			size_t kp = k;
			u.r.l = 0;
			while( kp > 0 && !(kp%f == 0 && kp/f+u.r.l<=c) )
			{
				u.r.L[u.r.l] = MemMan::template Alloc<FMerger<T, InStreamClass, MemMan> >(c, in+(k-kp));
				kp -= c;
				u.r.l++;
			}
			while( kp > 0 )
			{
				u.r.L[u.r.l] = MemMan::template Alloc<FMerger<T, InStreamClass, MemMan> >(f, in+(k-kp));
				kp -= f;
				u.r.l++;
			}
			assert( kp == 0 );

			// Construct the buffers
			for( size_t i=0; i<u.r.l; i++ )
				if( u.r.L[i]->k > u.r.l )
				{
					_Buffer b(2*u.r.L[i]->k*u.r.L[i]->k*u.r.L[i]->k);
					u.r.B[i] = b;
				}
				else
				{
					_Buffer b(2*u.r.l*u.r.l*u.r.l);
					u.r.B[i] = b;
				}
			// Construct the head merger
			u.r.R = MemMan::template Alloc<FMerger<T, _Buffer, MemMan> >(u.r.l, u.r.B);
		}
	}
	~FMerger()
	{
		if( k > 3 )
		{
			MemMan::DAlloc(u.r.R);
			for( size_t i=u.r.l; i>0; i-- )
				MemMan::DAlloc(u.r.L[i-1]);
			MemMan::DAlloca(u.r.B);
			MemMan::DAllocap(u.r.L);
		}
	}
	static size_t size(const size_t k)
	{
		assert( k > 1 ); // Punish the ones who want to merge a single stream
		size_t s = sizeof(FMerger<T, InStreamClass, MemMan>);
		if( k < 4 )
			return s;
		else
		{
			const size_t f = (size_t)floor(sqrt((double)k));
			const size_t c = (size_t)ceil(sqrt((double)k));
			size_t* ks = new size_t[c];
			s += sizeof(FMerger<T, InStreamClass, MemMan>*)*c;
			s += sizeof(_Buffer)*c;
			size_t kp = k;
			size_t l = 0;
			while( kp > 0 && !(kp%f == 0 && kp/f+l<=c) )
			{
				ks[l] = c;
				s += size(c);
				kp -= c;
				l++;
			}
			while( kp > 0 )
			{
				ks[l] = f;
				s += size(f);
				kp -= f;
				l++;
			}
			assert( kp == 0 );
			for( size_t i=0; i<l; i++ )
				if( ks[i] > l )
					s += sizeof(T)*((2*ks[i]*ks[i]*ks[i])+1)+sizeof(size_t);
				else
					s += sizeof(T)*((2*l*l*l)+1)+sizeof(size_t);
			s += size(l);
			delete[] ks;
			return s;
		}
	}

	template<class OutStreamClass>
	inline size_type invoke(OutStreamClass& out, const size_type N)
	{
		size_type count = 0;
		m_bEmpty = false;
		if( k == 2 )
			count = invoke(*u.S[0], *u.S[1], out, N);
		else if( k == 3 )
			count = invoke(*u.S[0], *u.S[1], *u.S[2], out, N);
		else
			count = _invoke(out, N);
		// If we didn't deliver, we must've run dry
		m_bEmpty = count < N;
		return count;
	}

	inline bool empty() const
		{ return m_bEmpty; }
	inline size_t degree() const
		{ return k; }
private:
	// Merges n elements from in1 and in2 into out. Returns the number of elements actually sent to out
	template<class OutStreamClass>
	static size_type invoke(InStreamClass& in1, InStreamClass& in2, OutStreamClass& out, const size_type N)
	{
		T e;
		if( in1.empty() )
			return in2.invoke(out, N);
		if( in2.empty() )
			return in1.invoke(out, N);
		size_type count=0;
		while( count < N )
		{
			if( in1.peek() < in2.peek() )
			{
				in1 >> e;
				out << e;
				count++;
				if( in1.empty() )
					return count+in2.invoke(out, N-count);
			}
			else
			{
				in2 >> e;
				out << e;
				count++;
				if( in2.empty() )
					return count+in1.invoke(out, N-count);
			}
		}
		return count;
	}
	// Merges n elements from in1, in2 and in3 into out. Returns the number of elements actually sent to out
	template<class OutStreamClass>
	static size_type invoke(InStreamClass& in1, InStreamClass& in2, InStreamClass& in3, OutStreamClass& out, const size_type N)
	{
		T e;
		if( in1.empty() )
			return invoke(in2, in3, out, N);
		if( in2.empty() )
			return invoke(in1, in3, out, N);
		if( in3.empty() )
			return invoke(in1, in2, out, N);
		size_type count=0;
		while( count < N )
		{
			if( in1.peek() < in2.peek() )
				if( in1.peek() < in3.peek() )
				{
					in1 >> e;
					out << e;
					count++;
					if( in1.empty() )
						return count+invoke(in2, in3, out, N-count);
				}
				else
				{
					in3 >> e;
					out << e;
					count++;
					if( in3.empty() )
						return count+invoke(in1, in2, out, N-count);
				}
			else
				if( in2.peek() < in3.peek() )
				{
					in2 >> e;
					out << e;
					count++;
					if( in2.empty() )
						return count+invoke(in1, in3, out, N-count);
				}
				else
				{
					in3 >> e;
					out << e;
					count++;
					if( in3.empty() )
						return count+invoke(in1, in2, out, N-count);
				}
		}
		return N;
	}

#pragma warning(push)
#pragma warning(disable: 4714) // invoke() will not be inlined
	template<class OutStreamClass>
	size_type _invoke(OutStreamClass& out, const size_type N)
	{
		size_type count = 0;
		// Keep invoking R until we've run dry, or N elements has been output
		size_t k3 = u.r.R->degree();
		k3 = k3*k3*k3;
		do
		{
			// Make sure the buffers are (half) filled
			for( size_t i=0; i<u.r.l; i++ )
			{
				size_type halfSize = u.r.B[i].getSize()/2;
				if( u.r.B[i].elementCount() < halfSize && !u.r.L[i]->empty() )
					u.r.L[i]->invoke(u.r.B[i], halfSize);
			}
			// If there're still elements left, invoke the head
			if( N-count < k3 ) // Don't extract too much, though
				count += u.r.R->invoke(out, N-count);
			else
				count += u.r.R->invoke(out, k3);
		}
		while( count<N && !u.r.R->empty() );
		return count;
	}
#pragma warning(pop)

protected:
	union tagUnion
	{
		struct tagRecursive
		{
			FMerger<T, InStreamClass, MemMan>** L;
			_Buffer* B;
			FMerger<T, _Buffer, MemMan>* R;
			size_t l;
		} r;
		InStreamClass* S[3];
	} u;

	bool m_bEmpty;
	size_t k;
};


template<class T, class InStreamClass>
class FFMerger
{
private:
	typedef _uint64_t size_type;
public:
	FFMerger(const size_t k, InStreamClass* in) : sm(FMerger<T, InStreamClass, StackMemMan>::size(k)-sizeof(FMerger<T, InStreamClass, StackMemMan>)), m(k, in)
		{ }
	template<class OutStreamClass>
	inline size_type invoke(OutStreamClass& out, const size_type N)
		{ return m.invoke(out, N); }
	inline bool empty() const
		{ return m.empty(); }
	inline size_t degree() const
		{ return m.degree(); }

private:
	StackMemMan::StackMem sm;
	FMerger<T, InStreamClass, StackMemMan> m;
};

*/
/*
template<class T, class InStreamClass, class MemMan = HeapMemMan>
class LFMerger
{
public:
	LFMerger(const size_t k, InStreamClass* in) : k(k), in(in)
	{
//		Layout(log(k));
	}
	~LFMerger()
		{ }

	void Layout(size_t logk)
	{
		if( k )
		{
            Layout(split(logk));
			size_t sms = 1<<(logk-split(logk));
			for( size_t i=0; i<1<<split(logk); i++ )
			{
				lm.Allocate(sizeof(T)*sms*sms*sms);
				lm.Layout((k+1)/2);
			}
		}
	}

	static void Layout(LayoutManager<LFMerger>& lm, size_t k)
	{
		if( k )
		{
            lm.Layout(k/2);
			size_t sms = 1<<((k+1)/2);
			for( size_t i=0; i<k/2; i++ )
			{
				lm.Allocate(sizeof(T)*sms*sms*sms);
				lm.Layout((k+1)/2);
			}
		}
	}

	template<class OutStreamClass>
	inline index_t invoke(OutStreamClass& out, const index_t N)
		{ return root().invoke(out, N); }
	inline bool empty() const
		{ return false; }
	inline size_t degree() const
		{ return k; }

private:
	class Buffer
	{
		Buffer(size_t logk, void* start) : logk(logk), start(start)
			{ }

	private:
		size_t logk;
		void* start;
	};

	class Merger
	{
	public:
		Merger(size_t logk, void* start) : logk(logk), start(start)
			{ }

		Merger top()
			{ return Merger(split(logk), start); }
		Merger bottom(size_t i)
			{ size_t sms = memUsage(logk-split(logk)); size_t bs = MemUsageClass::buffer().size(logk); return Merger(split(logk), (char*)start+memUsage(split(logk))+(i+1)*bs+i*sms); }
		Buffer buffer(size_t i)
			{ size_t sms = memUsage(logk-split(logk)); size_t bs = MemUsageClass::buffer().size(logk); return Merger(split(logk), (char*)start+memUsage(split(logk))+i*bs+i*sms); }

		template<class OutStreamClass>
		index_t invoke(OutStreamClass& out, index_t N)
		{
			return 2;
		}			

	private:
		size_t logk;
		void* start;
	};

	Merger root()
	{
		return Merger(1, NULL);
	}

	static size_t split(size_t hight)
		{ return hight/2; }
	static size_t log(size_t n)
		{ size_t logn = 0; for( size_t i=n-1; i; logn++, i>>=1 ); return logn; }
	static class MemUsageClass
	{
	public:
		MemUsageClass()
			{ }
		size_t operator()(size_t logk)
		{
			size_t s = 0;
			if( logk > 1 )
			{
				s += operator()(split(logk)); // top merger
				size_t tmd = 1<<(split(logk)/2); // top merger degree
				s += tmd*operator()(logk-split(logk)); // bottom mergers
				s += buffer().size(logk); // buffers
			}
			return s;
		}
		static size_t buffer().size(size_t logk)
			{ size_t sms = 1<<(logk-split(logk)/2); return sizeof(T)*sms*sms*sms + sizeof(void*); }
	} memUsage;

	size_t k;
	InStreamClass* in;
};
*/

#endif // __FMERGER_H_INCLUDED_
