/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __FSTREAM_H_INCLUDED_
#define __FSTREAM_H_INCLUDED_

/* Interface:
template<class T>
class FStream
{
public:
	virtual FStream& operator<<(const T& e) = 0;
	virtual FStream& operator>>(T& e) = 0;
	virtual T& peek() = 0;
	virtual bool empty() = 0;
	virtual ~FStream()
		{ }
};*/

template<class T, class ArrayClass /*= DynArray<T>*/ >
class FBuffer
{
public:
	typedef typename ArrayClass::size_type size_type;

	FBuffer() : begin(0), end(0), size(1)
		{ }
	explicit FBuffer(size_type size) : m_Array(size+1), begin(0), end(0), size(size+1)
		{ }
	FBuffer(const char* szImageName, int flags, size_type size) : m_Array(szImageName, flags, size+1), begin(0), end(0), size(size+1)
		{ }
	FBuffer(const ArrayClass& array, size_type size) : m_Array(array), begin(0), end(size), size(size+1)
		{ }
	inline FBuffer<T, ArrayClass>& operator<<(const T& e)
	{
		assert( (end+1)%size != begin );
		m_Array[end] = e;
		end = end+1==size? 0 : end+1; // = (end+1) % size;
		return *this;
	}
	inline FBuffer<T, ArrayClass>& operator>>(T& e)
	{
		assert( end != begin );
		e = m_Array[begin];
		begin = begin+1==size? 0 : begin+1; // = (begin+1) % size;
		return *this;
	}
	template<class OutStreamClass>
	inline size_type invoke(OutStreamClass& out, const size_type N)
	{
		if( begin <= end )
		{
			if( end-begin <= N )
			{
				size_type count = end-begin;
				while( begin < end )
				{
					out << m_Array[begin];
					begin++;
				}
				return count;
			}
			else
			{
				for( size_type count=0; count<N; begin++, count++ )
					out << m_Array[begin];
				return N;
			}
		}
		else
		{
			if( begin+N >= size )
			{
				size_type count = size-begin;
				while( begin < size )
				{
					out << m_Array[begin];
					begin++;
				}
				begin = 0;
				if( end < N-count )
				{
					count += end;
					while( begin < end )
					{
						out << m_Array[begin];
						begin++;
					}
				}
				else
					for( ; count<N; begin++, count++ )
						out << m_Array[begin];
				return count;
			}
			else
			{
				for( size_type count=0; count<N; begin++, count++ )
					out << m_Array[begin];
				return N;
			}
		}
	}
	inline bool empty()
		{ return end == begin; }
	inline T& peek()
		{ return m_Array[begin]; }

	inline size_type elementCount()
		{ return end>=begin? end-begin : size-begin+end; /* (end-begin+size)%size; */ }
	inline size_type getSize()
		{ return size-1; }
	inline void reset()
		{ end = begin; }
	// Clears the buffer and sets the size to newSize. Only supported on arrays with single-argument constructors
	inline void setSize(size_type newSize)
	{
		begin = 0;
		end = 0;
		size = newSize;
		m_Array = ArrayClass(newSize+1);
	}

private:
	ArrayClass m_Array;
	size_type begin, end, size;
};


template<class T>
class Verify
{
public:
	Verify(const T& minElement) : last(minElement), c(0), sum(0)
		{ }

	Verify<T>& operator<<(const T& e)
	{
		if( e < last )
		{
			std::cerr << "Unordered: " << e << " < " << last << " at " << (unsigned int)c << std::endl;
			abort();
		}
//		if( e == last )
//			std::cerr << "Dulicate: " << e << std::endl;
		last = e;
		c++;
		sum += e;
		return *this;
	}

	void reset(const T& minElement)
		{ c = minElement; sum = 0; }
	size_t insertionCount()
		{  return c; }
	size_t checkSum()
		{  return sum; }

protected:
	T last;
	size_t c;
	size_t sum;
};


template<class T>
class Term
{
public:
	Term(std::ostream* os, std::istream* is) : m_pOStream(os), m_pIStream(is)
		{ }
	Term(std::iostream* s) : m_pOStream(s), m_pIStream(s)
		{ }

	Term<T>& operator<<(const T& e)
	{
		*m_pOStream << e << std::endl;
		return *this;
	}
	Term<T>& operator>>(T& e)
	{
		*m_pIStream >> e;
		return *this;
	}
	template<class OutStreamClass>
	_uint64_t invoke(OutStreamClass* out, const _uint64_t N)
	{
		for( _uint64_t count=0; count<N; count++ )
		{
			*m_pIStream >> dummy;
			*out << dummy;
		}
		return N;
	}
	bool empty()
		{ return false; }
	T& peek()
	{
		*m_pIStream >> dummy;
		return dummy;
	}

protected:
	std::istream* m_pIStream;
	std::ostream* m_pOStream;
	static T dummy;
};

template<class T>
T Term<T>::dummy = T();

template<class Stream, class T>
class OutStreamIterator
{
public:
	OutStreamIterator(Stream& s) : m_Stream(s), offset(0)
		{ }
	T& operator*()
		{ return tmp; }
	void operator++()
	{
		offset++;
		s << elt;
	}
	OutStreamIterator<Stream,T>& operator+=(size_t k)
	{
		offset += k;
		return *this;
	}
	OutStreamIterator<Stream,T>& operator=(T& lhs)
	{
		elt = lhs;
		return *this;
	}

	bool operator==(const OutStreamIterator<Stream,T>& rhs) const
		{ return offset == rhs.offset; }

private:
	Stream& m_Stream;
	T elt;
	size_t offset;
};

template<class T>
class Sink
{
public:
	Sink() : offset(0)
		{ }
	inline Sink<T>& operator++()
	{
		offset++;
		return *this;
	}
	inline Sink<T>& operator++(int)
	{
		++offset;
		return *this;
	}
	inline Sink<T>& operator+=(size_t k)
	{
		offset += k;
		return *this;
	}
	inline Sink<T> operator+(size_t k) const
	{
		return Sink<T>(offset+k);
	}
	inline int operator-(const Sink<T>& rhs) const
	{
		return offset-rhs.offset;
	}
	bool operator==(const Sink<T>& rhs) const
		{ return offset == rhs.offset; }
	bool operator!=(const Sink<T>& rhs) const
		{ return offset != rhs.offset; }
	T& operator*()
		{ return dummy; }
private:
	Sink(size_t o) : offset(o)
		{ }
	size_t offset;
	T dummy;
};

#endif // __FSTREAM_H_INCLUDED_
