/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __STREAMMERGER_H_INCLUDED_
#define __STREAMMERGER_H_INCLUDED_

#include <memory>
#include "int64.h"

#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER

template<class T, class InStream>
class mem_splitter
{
public:
	static inline size_t prefered_order_output(_uint64_t N)
	{
//		run_size = (ItDiff)((_int64_t)(M/sizeof(T)-1)*N/(N+M-1));
//		return N/run_size;
		(void)N;
		return M/(B+sizeof(InStream));
	}
	static inline size_t prefered_order_input(_uint64_t N)
	{
//		run_size = (ItDiff)((_int64_t)(M/sizeof(T)-1)*N/(N+M-1));
//		return N/run_size;
		(void)N;
		return M/(B+sizeof(InStream));
	}
	static inline _uint64_t switch_to_cache_aware()
	{
		return M/sizeof(T);
	}
private:
	static const size_t M = 214748364;
	static const size_t B = 131072;
};

template<class T, class InStream>
class cache_splitter
{
public:
	static inline size_t prefered_order(_uint64_t N)
	{
//		run_size = (ItDiff)((_int64_t)(M/sizeof(T)-1)*N/(N+M-1));
//		return N/run_size;
		(void)N;
		return M/(B+sizeof(InStream));
	}
	static inline _uint64_t switch_to_cache_aware()
	{
		return M/sizeof(T);
	}
private:
	static const size_t M = 240000;
	static const size_t B = 128;
};

template<class T, int M, int B, class InStream, class Allocator=std::allocator<T>, class Refiller = NopRefill<typename InStream::iterator> >
class HeapMerger
{
protected:
	class HeapMergerCache
	{
	public:
		HeapMergerCache() : pMerger(NULL) { }
		explicit inline HeapMergerCache(size_t maxorder) : pMerger(NULL) { }
		virtual ~HeapMergerCache() { }

		inline HeapMerger* get_merger(size_t order)
		{
			delete pMerger;
			pMerger = new HeapMerger(order);
			return pMerger;
		}
	private:
		HeapMerger* pMerger;
	};

	class PQElement
	{
	public:
		inline PQElement()
			{ }
		inline void operator=(InStream rhs)
			{ pair = rhs; first = *pair.head(); }
		inline bool operator>(const PQElement& rhs) const
			{ return first > rhs.first; }
		inline T extract()
		{
			T ret(first);
			pair.head()++;
			if( !empty() )
				first = *pair.head();
			return ret;
		}
		inline bool empty()
			{ return pair.head() == pair.tail(); }

		inline typename InStream::iterator begin()
			{ return pair.head(); }
		inline typename InStream::iterator& head()
			{ return pair.head(); }
		inline const typename InStream::iterator& tail() const
			{ return pair.tail(); }

		InStream pair;
		T first;
	};
public:
	typedef HeapMergerCache cache;
	typedef Allocator allocator;
	typedef InStream stream;
	typedef T value_type;
	enum { order = 2 };
	template<class NewRefiller>
	struct rebind { typedef HeapMerger<T, M, B, InStream, Allocator, NewRefiller> other; };

	explicit inline HeapMerger(size_t k) : m_bChanged(true)
	{
		pqend = pqbegin = alloc.allocate(k);
		for( PQElement dflt, *p=pqbegin; p<pqbegin+k; ++p )
			alloc.construct(p, dflt);
	}
	explicit inline HeapMerger(size_t k, Allocator& a) : alloc(a), m_bChanged(true)
	{
		pqend = pqbegin = alloc.allocate(k);
		for( PQElement dflt, *p=pqbegin; p<pqbegin+k; ++p )
			alloc.construct(p, dflt);
	}
	virtual ~HeapMerger()
		{ delete[] pqbegin; }
	static inline size_t prefered_order(_uint64_t N)
	{
//		run_size = (ItDiff)((_int64_t)(M/sizeof(T)-1)*N/(N+M-1));
//		return N/run_size;
		(void)N;
		return M/(B+sizeof(InStream));
	}
	static inline size_t min_order()
	{
		return 2;
	}
	static inline _uint64_t switch_to_cache_aware()
		{ return M/sizeof(T); }
/*	inline InStream get_input_stream(size_t k) const
		{ return pqbegin[k].pair; }
	inline void set_input_stream(size_t k, const InStream& s)
		{ pqbegin[k] = s; m_bChanged = true; }
	inline size_t order() const
		{ return pqend-pqbegin; }
*/	inline void set_refiller(const Refiller& r)
		{ m_Refiller = r; }
	inline const Refiller& get_refiller() const
		{ return m_Refiller; }

	// New interface:
	typedef const PQElement* const_stream_iterator;
	typedef PQElement* stream_iterator;
	inline void add_stream(const InStream& stream)
		{ *pqend++ = stream; }
	inline const_stream_iterator begin() const
		{ return pqbegin; }
	inline const_stream_iterator end() const
		{ return pqend; }
	inline stream_iterator begin()
		{ return pqbegin; }
	inline stream_iterator end()
		{ return pqend; }
	static inline size_t size_of(int k)
		{ return k*sizeof(PQElement); }
	inline void reset()
		{ m_bChanged = true, pqend = pqbegin; }

	template<class OutStreamIt>
	inline OutStreamIt empty(OutStreamIt outbegin, OutStreamIt outend)
		{ (void)outend; return outbegin; }

	template<class OutStreamIt>
	inline OutStreamIt operator()(OutStreamIt outbegin, OutStreamIt outend)
	{
		if( m_bChanged )
			std::make_heap(pqbegin, pqend, std::greater<PQElement>());
		m_bChanged = false;
		while( pqbegin!=pqend && outbegin!=outend )
		{
			std::pop_heap(pqbegin, pqend, std::greater<PQElement>());
			PQElement* least = pqend-1;
			m_Refiller(this, least->pair.head(), least->pair.head()+1);
			*outbegin = least->extract();
			++outbegin;
			if( !least->empty() )
				std::push_heap(pqbegin, pqend, std::greater<PQElement>());
			else
				pqend--;
		}
		return outbegin;
	}

protected:
	typename Allocator::template rebind<PQElement>::other alloc;
	PQElement *pqbegin, *pqend;
	bool m_bChanged;
	Refiller m_Refiller;
};




template<class T, int M, int B, class InStream>
class LoserMerger
{
protected:
	class LoserMergerCache
	{
	public:
		LoserMergerCache() : pMerger(NULL) { }
		explicit inline LoserMergerCache(size_t maxorder) : pMerger(NULL) { }
		virtual ~LoserMergerCache() { }
		inline LoserMerger* get_merger(size_t order)
		{
			delete pMerger;
			pMerger = new LoserMerger(order);
			return pMerger;
		}
	private:
		LoserMerger* pMerger;
	};

	struct External
	{
		InStream s;
	};
	struct Internal
	{
		T value;
		External* origin;
	};

public:
	typedef LoserMergerCache Cache;

	explicit LoserMerger(size_t k)
	{
		exbegin = new External[k];
		exend = exbegin+k;
		inbegin = new Internal[k];
		inend = inbegin+k;
	}
	static size_t prefered_order(_uint64_t N)
	{
		(void)N;
		return M/(B+sizeof(External)+sizeof(Internal));
	}
	static _uint64_t switch_to_cache_aware()
		{ return M/sizeof(T); }
	inline InStream& input_stream(size_t k)
		{ return exbegin[k].s; }
	inline const InStream& input_stream(size_t k) const
		{ return exbegin[k].s; }

	template<class OutStreamIt>
	inline OutStreamIt invoke(OutStreamIt outbegin, OutStreamIt outend)
	{
		build();
		(void)outbegin;
		return outbegin;
	}

protected:
	void build()
	{
	}

	External *exbegin, *exend;
	Internal *inbegin, *inend;
	Internal winner;
};

#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __STREAMMERGER_H_INCLUDED_
