/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __LAZYMERGER_H_INCLUDED_
#define __LAZYMERGER_H_INCLUDED_

#include <cmath>
#include <vector>
#include "BasicMerger.h"
#include "int64.h"

#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER
#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__


template<class It> class NopRefill { public: template<class Merger> inline bool operator()(const Merger*, It, It)
	{ return false; } };

template<class It>
class InputStream
{
public:
	typedef It iterator;
	typedef It pointer;

	inline InputStream() { }
	inline InputStream(const It& b, const It& e) : b(b), end(e) { }

	inline It begin()
		{ return b; }

	inline It& head()
		{ return b; }
	inline const It& tail() const
		{ return end; }

private:
	It b, end;
};

template<class MergeTree, class InStream, class Allocator = std::allocator<typename MergeTree::value_type>, class Refiller = NopRefill<typename InStream::iterator> >
class LazyMerger
{
public:
	class LazyMergerCache
	{
	public:
		inline LazyMergerCache()
		{
			for( It i=m_Cache.begin(); i!=m_Cache.end(); i++ )
				*i = NULL;
		}
		virtual ~LazyMergerCache()
		{
			for( It i=m_Cache.begin(); i!=m_Cache.end(); i++ )
				delete *i;
		}
		LazyMerger* get_merger(size_t order)
		{
			order = _clog(order);
			if( m_Cache.size() < order+1 )
				m_Cache.resize(order+1, NULL);
			if( m_Cache[order] == NULL )
				m_Cache[order] = new LazyMerger(1<<order);
			else
				m_Cache[order]->reset();
			return m_Cache[order];
		}
		void flush()
		{
			for( It i=m_Cache.begin(); i!=m_Cache.end(); i++ )
			{
				delete *i;
				*i = NULL;
			}
		}
	private:
		typedef typename std::vector<LazyMerger*>::iterator It;
		std::vector<LazyMerger*> m_Cache;
	};

	typedef typename MergeTree::MergerLoc MergerLoc;
	typedef typename MergeTree::Buffer Buffer;

	typedef LazyMergerCache cache;
	typedef Allocator allocator;
	typedef InStream stream;
	typedef typename MergeTree::value_type value_type;
	enum { order = 2 };
	template<class NewRefiller>
	struct rebind { typedef LazyMerger<MergeTree, InStream, Allocator, NewRefiller> other; };

	explicit inline LazyMerger(size_t order) : m_MergeTree(_clog(order)), input_streams(new InStream[order]), last_stream(input_streams+order)
		{ assert( order > 3 ); reset_input(); }
	explicit inline LazyMerger(size_t order, Allocator) : m_MergeTree(_clog(order)), input_streams(new InStream[order]), last_stream(input_streams+order)
		{ assert( order > 3 ); reset_input(); }
	virtual ~LazyMerger()
		{ delete[] input_streams; }
	inline void set_refiller(const Refiller& r)
		{ m_Refiller = r; }
	inline const Refiller& get_refiller() const
		{ return m_Refiller; }

	static inline size_t prefered_order(_uint64_t N);
	static inline size_t min_order()
	{
		return 2;
	}
	static inline _uint64_t switch_to_cache_aware();

//	inline InStream get_input_stream(size_t k) const
//		{ return input_streams[k]; }
//	inline void set_input_stream(size_t k, const InStream& s)
//		{ input_streams[k] = s; }
//	inline size_t order() const
//		{ return last_stream-input_streams; }

	// New interface:
	typedef const InStream* const_stream_iterator;
	typedef InStream* stream_iterator;
	inline void add_stream(const InStream& stream)
		{ *last_stream++ = stream; }
	inline const_stream_iterator begin() const
		{ return input_streams; }
	inline const_stream_iterator end() const
		{ return last_stream; }
	inline stream_iterator begin()
		{ return input_streams; }
	inline stream_iterator end()
		{ return last_stream; }
	static size_t size_of(int)
		{ return 0; }
	inline void reset()
		{ reset_input(); }


	template<class OutStreamIt>
	inline OutStreamIt empty(OutStreamIt outbegin, OutStreamIt outend)
		{ return m_MergeTree.empty(outbegin, outend); }

	template<class OutStreamIt>
	OutStreamIt operator()(OutStreamIt outbegin, OutStreamIt outend)
	{
		register MergerLoc lm;
		LazyMergeFillLeft fillleft(lm);
		LazyMergeFillRight fillright(lm);
		LazyMergeFillUp fillup(lm);

//		assert(!( outend < outbegin ));
		m_MergeTree.ResetMergers();
		bool smallmerger = warmup(); //smallmerger = true;
		m_MergeTree.make_root(lm);
		godownleft(lm);
		if( smallmerger )
			for( ;; )
			{
#ifndef NDEBUG
				assert( !lm.left_exhausted() );
				assert( !lm.right_exhausted() );
				if( lm.is_leaf() )
				{
					if( lm.right_exhausted() )
						assert( input_streams[lm.bfs_index()+1].head() == input_streams[lm.bfs_index()+1].tail() );
					if( input_streams[lm.bfs_index()+1].head() == input_streams[lm.bfs_index()+1].tail() )
						assert( lm.right_exhausted() );
					if( lm.left_exhausted() )
						assert( input_streams[lm.bfs_index()].head() == input_streams[lm.bfs_index()].tail() );
					if( input_streams[lm.bfs_index()].head() == input_streams[lm.bfs_index()].tail() )
						assert( lm.left_exhausted() );
				}
				else
				{
					if( lm.right_exhausted() )
						assert( lm.right_buffer().tail() == lm.right_buffer().head() );
					if( lm.left_exhausted() )
						assert( lm.left_buffer().tail() == lm.left_buffer().head() );
				}
#endif // NDEBUG
				if( !lm.is_leaf() && lm.right_buffer().tail() == lm.right_buffer().head() )
				{
					lm.go_right();
					lm.buffer().tail() = lm.buffer().head();
					lm.buffer().head() = lm.buffer().begin();
					godownleft(lm);
				}
				else
				{
					bool isroot = lm.is_root();
					if( domerge(lm,outbegin,outend) )
						return outbegin;
					if( isroot )
						break;
				}
			}
		else
			for( ;; )
			{
#ifndef NDEBUG
				if( lm.is_leaf() )
				{
					if( lm.right_exhausted() )
						assert( input_streams[lm.bfs_index()+1].head() == input_streams[lm.bfs_index()+1].tail() );
					if( input_streams[lm.bfs_index()+1].head() == input_streams[lm.bfs_index()+1].tail() )
						assert( lm.right_exhausted() );
					if( lm.left_exhausted() )
						assert( input_streams[lm.bfs_index()].head() == input_streams[lm.bfs_index()].tail() );
					if( input_streams[lm.bfs_index()].head() == input_streams[lm.bfs_index()].tail() )
						assert( lm.left_exhausted() );
				}
				else
				{
					if( lm.right_exhausted() )
						assert( lm.right_buffer().tail() == lm.right_buffer().head() );
					if( lm.left_exhausted() )
						assert( lm.left_buffer().tail() == lm.left_buffer().head() );
				}
#endif // NDEBUG
				if( lm.right_exhausted() )
					if( lm.left_exhausted() )
						if( lm.is_root() )
							return outbegin;
						else
							fillup();
					else
					{
						if( copyleft(outbegin, outend, lm, input_streams, m_Refiller) )
							break;
					}

				else if( !lm.is_leaf() && lm.right_buffer().tail() == lm.right_buffer().head() )
				{
					void *dummy;
					fillright(dummy, dummy);
					while( !lm.is_leaf() ) //&& !lm.left_exhausted() && lm.left_buffer().tail() == lm.left_buffer().head() )
						fillleft(dummy, dummy);
				}

				else if( lm.left_exhausted() )
				{
					bool isroot = lm.is_root();
					if( copyright(outbegin, outend, lm, input_streams, m_Refiller) )
						return outbegin;
					if( isroot )
						break;
				}

				else
				{
					bool isroot = lm.is_root();
					if( domerge(lm,outbegin,outend) )
						return outbegin;
					if( isroot )
						break;
				}
			}

		for( ;; )
		{
#ifndef NDEBUG
			if( lm.is_leaf() )
			{
				if( lm.right_exhausted() )
					assert( input_streams[lm.bfs_index()+1].head() == input_streams[lm.bfs_index()+1].tail() );
				if( input_streams[lm.bfs_index()+1].head() == input_streams[lm.bfs_index()+1].tail() )
					assert( lm.right_exhausted() );
				if( lm.left_exhausted() )
					assert( input_streams[lm.bfs_index()].head() == input_streams[lm.bfs_index()].tail() );
				if( input_streams[lm.bfs_index()].head() == input_streams[lm.bfs_index()].tail() )
					assert( lm.left_exhausted() );
			}
			else
			{
				if( lm.right_exhausted() )
					assert( lm.right_buffer().tail() == lm.right_buffer().head() );
				else if( lm.right_buffer().tail() == lm.right_buffer().head() )
				{
					lm.go_right();
					assert( !lm.left_exhausted() || !lm.right_exhausted() );
					lm.go_up();
					assert( lm.left_buffer().tail() != lm.left_buffer().head() );
				}
				if( lm.left_exhausted() )
					assert( lm.left_buffer().tail() == lm.left_buffer().head() );
				else if( lm.left_buffer().tail() == lm.left_buffer().head() )
				{
					lm.go_left();
					assert( !lm.left_exhausted() || !lm.right_exhausted() );
					lm.go_up();
					assert( lm.right_buffer().tail() != lm.right_buffer().head() );
				}
			}
#endif // NDEBUG
			if( lm.right_exhausted() )
				if( lm.left_exhausted() )
					if( lm.is_root() )
						return outbegin;
					else
						fillup();
				else
				{
					if( copyleft(outbegin, outend, lm, input_streams, m_Refiller) )
						return outbegin;
				}

			else if( lm.left_exhausted() )
			{
				if( copyright(outbegin, outend, lm, input_streams, m_Refiller) )
					return outbegin;
			}

			else
				if( domerge(lm,outbegin,outend) )
					return outbegin;
		}
	}

protected:
	template<class OutStreamIt>
	inline bool domerge(MergerLoc& lm, OutStreamIt& outbegin, const OutStreamIt& outend)
	{
		LazyMergeFillLeft fillleft(lm);
		LazyMergeFillLeftLeaf fillleftleaf(lm, input_streams);
		LazyMergeFillRight fillright(lm);
		LazyMergeFillRightLeaf fillrightleaf(lm, input_streams);
		LazyMergeFillUp fillup(lm);
		if( lm.is_leaf() )
		{
			InStream *is1 = input_streams+lm.bfs_index(), *is2 = is1+1;
			typename InStream::iterator htmp, h1 = is1->head(), h2 = is2->head();
			bool abort = false;

/*			LazyMerge(is1->head(), is1->tail(), is2->head(), is2->tail(), lm.buffer().head(), lm.buffer().tail(), 0, fillleftleaf, fillrightleaf, fillup);
			typename Stream::iterator h2tmp = is2->head(); is2->head() = h2;
			m_Refiller(this, h1, is1->head());
			is2->head() = h2tmp;
			m_Refiller(this, h2, is2->head());
*/
			LazyMerge(h1, is1->tail(), h2, is2->tail(), lm.buffer().head(), lm.buffer().tail(), 0, fillleftleaf, fillrightleaf, fillup);
			htmp = is1->head();
			is1->head() = h1;
			abort = abort || m_Refiller(this, htmp, is1->head());

			htmp = is2->head();
			is2->head() = h2;
			abort = abort || m_Refiller(this, htmp, is2->head());
			return abort;
		}
		else if( lm.is_root() )
		{
			LazyMerge(lm.left_buffer().head(), lm.left_buffer().tail(), lm.right_buffer().head(), lm.right_buffer().tail(), outbegin, outend, 0, fillleft, fillright);
			if( outbegin == outend )
			{
				if( !lm.is_root() ) // One of children of the root got empty at the same time outbegin hit outend (and fill{left|right} was called
					lm.buffer().head() = lm.buffer().tail();
				return true;
			}
//			assert( outbegin < outend );
			return false;
		}
		else
		{
			LazyMerge(lm.left_buffer().head(), lm.left_buffer().tail(), lm.right_buffer().head(), lm.right_buffer().tail(), lm.buffer().head(), lm.buffer().tail(), 0, fillleft, fillright, fillup);
			return false;
		}
	}

	template<class OutStreamIt>
	bool copyright(OutStreamIt& outbegin, const OutStreamIt& outend, MergerLoc& lm, InStream* begin, Refiller& f)
	{
		LazyMergeFillUp fillup(lm);
		LazyMergeFillRight fillright(lm);
		LazyMergeFillRightLeaf fillrightleaf(lm, begin);
		assert( lm.left_exhausted() );
		if( lm.is_leaf() )
		{
			InStream *is = input_streams+lm.bfs_index()+1;
			typename InStream::iterator h = is->head();
			_LazyCopy(is->head(), is->tail(), lm.buffer().head(), lm.buffer().tail(), fillrightleaf, fillup);
			return f(this, h, is->head());
		}
		else if( lm.is_root() )
		{
			_LazyCopy(lm.right_buffer().head(), lm.right_buffer().tail(), outbegin, outend, fillright);
			if( outbegin == outend ) // if( outbegin >= outend )
			{
				if( !lm.is_root() ) // One of the roots children got empty at the same time outbegin hit outend (and fill{left|right} was called
					lm.buffer().head() = lm.buffer().tail();
				return true;
			}
			return false;
		}
		else
		{
			_LazyCopy(lm.right_buffer().head(), lm.right_buffer().tail(), lm.buffer().head(), lm.buffer().tail(), fillright, fillup);
			return false;
		}
	}
	template<class OutStreamIt>
	bool copyleft(OutStreamIt& outbegin, const OutStreamIt& outend, MergerLoc& lm, InStream* begin, Refiller& f)
	{
		LazyMergeFillUp fillup(lm);
		LazyMergeFillLeft fillleft(lm);
		LazyMergeFillLeftLeaf fillleftleaf(lm, begin);
		assert( lm.right_exhausted() );
		if( lm.is_leaf() )
		{
			InStream *is = input_streams+lm.bfs_index();
			typename InStream::iterator h = is->head();
			_LazyCopy(is->head(), is->tail(), lm.buffer().head(), lm.buffer().tail(), fillleftleaf, fillup);
			return f(this, h, is->head());
		}
		else if( lm.is_root() )
		{
			_LazyCopy(lm.left_buffer().head(), lm.left_buffer().tail(), outbegin, outend, fillleft);
			if( outbegin == outend ) // if( outbegin >= outend )
			{
				if( !lm.is_root() ) // One of the roots children got empty at the same time outbegin hit outend (and fill{left|right} was called
					lm.buffer().head() = lm.buffer().tail();
				return true;
			}
			return false;
		}
		else
		{
			_LazyCopy(lm.left_buffer().head(), lm.left_buffer().tail(), lm.buffer().head(), lm.buffer().tail(), fillleft, fillup);
			return false;
		}
	}


	inline void reset_input();
	inline static void godownleft(MergerLoc& lm);
	_int64_t recwarmup(MergerLoc& lm);
	bool warmup();

	class LazyMergeFillLeft
	{
	public:
		inline LazyMergeFillLeft(MergerLoc& lm) : m_locMerger(lm) { }
		template<class It>
		inline bool operator()(It&, It&)
		{
			m_locMerger.go_left();
			assert( m_locMerger.buffer().head() == m_locMerger.buffer().tail() );
			if( m_locMerger.right_exhausted() && m_locMerger.left_exhausted() )
			{
				m_locMerger.go_up();
				m_locMerger.set_left_exhausted();
				return false;
			}
			else
			{
				m_locMerger.buffer().tail() = m_locMerger.buffer().head();
				m_locMerger.buffer().head() = m_locMerger.buffer().begin();
				return false;
			}
		}
	protected:
		void operator=(LazyMergeFillLeft& rhs) { }
		MergerLoc& m_locMerger;
	};
	class LazyMergeFillRight
	{
	public:
		inline LazyMergeFillRight(MergerLoc& lm) : m_locMerger(lm) { }
		template<class It>
		inline bool operator()(It&, It&)
		{
			m_locMerger.go_right();
			assert( m_locMerger.buffer().head() == m_locMerger.buffer().tail() );
			if( m_locMerger.right_exhausted() && m_locMerger.left_exhausted() )
			{
				m_locMerger.go_up();
				m_locMerger.set_right_exhausted();
				return false;
			}
			else
			{
				m_locMerger.buffer().tail() = m_locMerger.buffer().head();
				m_locMerger.buffer().head() = m_locMerger.buffer().begin();
				return false;
			}
		}
	protected:
		void operator=(LazyMergeFillRight& rhs) { }
		MergerLoc& m_locMerger;
	};
	class LazyMergeFillLeftLeaf
	{
	public:
		inline LazyMergeFillLeftLeaf(MergerLoc& lm, const InStream* b) : m_locMerger(lm), b(b) { }
		template<class It>
		inline bool operator()(It&, It& )
		{
			assert( m_locMerger.is_leaf() );
			m_locMerger.set_left_exhausted();
			m_locMerger.buffer().tail() = m_locMerger.buffer().head();
			m_locMerger.buffer().head() = m_locMerger.buffer().begin();
			m_locMerger.go_up();
			return false;
		}
	protected:
		void operator=(LazyMergeFillLeftLeaf& rhs) { }
		MergerLoc& m_locMerger;
		const InStream* b;
	};
	class LazyMergeFillRightLeaf
	{
	public:
		inline LazyMergeFillRightLeaf(MergerLoc& lm, const InStream* b) : m_locMerger(lm), b(b) { }
		template<class It>
		inline bool operator()(It&, It&)
		{
			assert( m_locMerger.is_leaf() );
			m_locMerger.set_right_exhausted();
			m_locMerger.buffer().tail() = m_locMerger.buffer().head();
			m_locMerger.buffer().head() = m_locMerger.buffer().begin();
			m_locMerger.go_up();
			return false;
		}
	protected:
		void operator=(LazyMergeFillRightLeaf& rhs) { }
		MergerLoc& m_locMerger;
		const InStream* b;
	};
	class LazyMergeFillUp
	{
	public:
		inline LazyMergeFillUp(MergerLoc& lm) : m_locMerger(lm) { }
		inline void operator()()
		{
			assert( !m_locMerger.is_root() );
			m_locMerger.buffer().tail() = m_locMerger.buffer().head();
			m_locMerger.buffer().head() = m_locMerger.buffer().begin();
			m_locMerger.go_up();
		}
	protected:
		void operator=(LazyMergeFillUp& rhs) { }
		MergerLoc& m_locMerger;
	};

	MergeTree m_MergeTree;
	InStream *input_streams, *last_stream;
	Refiller m_Refiller;
};




// A desaster waiting to happen: (It's horribly outdated and unmaintained, so try and instatiate it if you dare... :)

template<class T, class InStream, class MergeTree>
class RecLazyMerger
{
protected:
	class LazyMergerCache
	{
	public:
		LazyMergerCache()
		{
			for( It i=m_Cache.begin(); i!=m_Cache.end(); i++ )
				*i = NULL;
		}
		virtual ~LazyMergerCache()
		{
			for( It i=m_Cache.begin(); i!=m_Cache.end(); i++ )
				delete *i;
		}
		RecLazyMerger* get_merger(size_t order)
		{
			order = _clog(order);
			if( m_Cache.size() < order+1 )
				m_Cache.resize(order+1, NULL);
			if( m_Cache[order] == NULL )
				m_Cache[order] = new RecLazyMerger(1<<order);
			else
				m_Cache[order]->reset_input();
			return m_Cache[order];
		}
		void flush()
		{
			for( It i=m_Cache.begin(); i!=m_Cache.end(); i++ )
			{
				delete *i;
				*i = NULL;
			}
		}
	private:
		typedef typename std::vector<RecLazyMerger*>::iterator It;
		std::vector<RecLazyMerger*> m_Cache;
	};

	static const int BASIC_ORDER = 2;

public:
	typedef LazyMergerCache Cache;

	explicit RecLazyMerger(size_t order) : m_MergeTree(_clog(order+(order&1))), begin(new InStream[order+(order&1)]), end(begin+order)
	{
//		assert( order > 3 );
		if( order > 3 )
		{
#ifndef NDEBUG
			assert_valid();
#endif // NDEBUG
			if( order & 1 )
				begin[order].second = begin[order].first;
		}
	}
	virtual ~RecLazyMerger()
		{ delete[] begin; }
	static inline size_t prefered_order(_uint64_t N)
	{
		return (size_t)floor(pow((double)N, 1.0/3)+.5);
	}
	static inline _uint64_t switch_to_cache_aware()
	{
		return 26;
	}
	inline InStream& input_stream(size_t k)
	{
		return begin[k];
	}
	inline const InStream& input_stream(size_t k) const
	{
		return begin[k];
	}
	template<class OutStreamIt>
	OutStreamIt invoke(OutStreamIt outbegin, OutStreamIt outend)
	{
		m_MergeTree.ResetMergers();
		m_MergeTree.make_root(m_locMerger);
		for( size_t f=1<<(_clog(end-begin)-1); f>1; f>>=1 )
		{
			if( f & (end-begin-1) )
				m_locMerger.go_right();
			else
			{
				m_locMerger.go_right();
				m_locMerger.set_dry();
				m_locMerger.go_up();
				m_locMerger.go_left();
			}
		}

		m_itFirstStream = begin;
		m_MergeTree.make_root(m_locMerger);
		LazyMerge(m_locMerger.left_buffer().head(), m_locMerger.left_buffer().tail(), m_locMerger.right_buffer().head(), m_locMerger.right_buffer().tail(), outbegin, outend, 0, LazyMergeFillLeft(), LazyMergeFillRight());
		if( m_locMerger.left_buffer().tail()==m_locMerger.left_buffer().head() && m_locMerger.right_buffer().tail()==m_locMerger.right_buffer().head() )
			m_locMerger.set_dry();
		return outbegin;
	}

protected:
	typedef typename MergeTree::MergerLoc MergerLoc;

	void reset_input()
	{
		for( InStream* p=begin; p<end; p++ )
			p->second = p->first;
	}

	MergeTree m_MergeTree;
	InStream* begin;
	InStream* end;

	// HACK: Not cross platform thread safe
	static MergerLoc /*__declspec(thread)*/ m_locMerger;
	static InStream* /*__declspec(thread)*/ m_itFirstStream;

	class LazyMergeFill
	{
	protected:
		template<class It>
		static void postmove(It& begin, It& end)
		{
			(void)begin; /* assert( m_locMerger.buffer().head() == m_locMerger.buffer().tail() ); */
			m_locMerger.buffer().head() = m_locMerger.buffer().begin();
			m_locMerger.buffer().tail() = m_locMerger.buffer().begin();
			ni();
			assert( begin == m_locMerger.buffer().head() );
//			begin = m_locMerger.buffer().head(); // Done automatically??
			end = m_locMerger.buffer().tail();
		}
		/*inline*/ static void ni()
		{
			if( m_locMerger.is_leaf() )
			{
				LazyMerge(m_itFirstStream[m_locMerger.bfs_index()].first, m_itFirstStream[m_locMerger.bfs_index()].second, m_itFirstStream[m_locMerger.bfs_index()+1].first, m_itFirstStream[m_locMerger.bfs_index()+1].second, m_locMerger.buffer().tail(), m_locMerger.buffer().end());
				if( m_itFirstStream[m_locMerger.bfs_index()].first==m_itFirstStream[m_locMerger.bfs_index()].second && m_itFirstStream[m_locMerger.bfs_index()+1].first==m_itFirstStream[m_locMerger.bfs_index()+1].second )
					m_locMerger.set_dry();
			}
			else
			{
				LazyMerge(m_locMerger.left_buffer().head(), m_locMerger.left_buffer().tail(), m_locMerger.right_buffer().head(), m_locMerger.right_buffer().tail(), m_locMerger.buffer().tail(), m_locMerger.buffer().end(), 0, LazyMergeFillLeft(), LazyMergeFillRight());
				if( m_locMerger.left_buffer().tail()==m_locMerger.left_buffer().head() && m_locMerger.right_buffer().tail()==m_locMerger.right_buffer().head() )
					m_locMerger.set_dry();
			}
		}
	};
	class LazyMergeFillLeft : public LazyMergeFill
	{
	public:
		template<class It>
		inline bool operator()(It& begin, It& end)
		{
			(void)begin; /* assert( begin == end ); */
			assert( !m_locMerger.dry() );
			m_locMerger.go_left();
			if( !m_locMerger.dry() )
				postmove(begin, end);
			m_locMerger.go_up();
			return true;
		}
	};
	class LazyMergeFillRight : public LazyMergeFill
	{
	public:
		template<class It>
		inline bool operator()(It& begin, It& end)
		{
			(void)begin; /* assert( begin == end ); */
			assert( !m_locMerger.dry() );
			m_locMerger.go_right();
			if( !m_locMerger.dry() )
				postmove(begin, end);
			m_locMerger.go_up();
			return true;
		}
	};

	void assert_valid()
	{
		m_MergeTree.make_root(m_locMerger);
		m_locMerger.go_left();
		_assert_valid();
		m_locMerger.go_up();
		m_locMerger.go_right();
		_assert_valid();
		m_locMerger.go_up();
	}
	void _assert_valid()
	{
//		ptrdiff_t buffsize = m_locMerger.buffer().end()-m_locMerger.buffer().begin();
//		T* tmp = new T[buffsize];
//		T* ref = new T[buffsize];
		m_locMerger.assert_valid();
		for( T* /*r=ref, *t=tmp, **/p=m_locMerger.buffer().begin(); p<m_locMerger.buffer().end(); p++/*, t++, r++*/ )
		{
//			*t = *p;
			/**r =*/ *p = 42;//rand();
		}
		if( !m_locMerger.is_leaf() )
		{
			m_locMerger.go_left();
			_assert_valid();
			m_locMerger.go_up();
			m_locMerger.go_right();
			_assert_valid();
			m_locMerger.go_up();
		}
		m_locMerger.assert_valid();
		for( T* /*r=ref, *t=tmp, **/p=m_locMerger.buffer().begin(); p<m_locMerger.buffer().end(); p++/*, t++, r++*/ )
		{
			assert( /**r*/ 42 == *p );
//			*p = *t;
		}
//		delete[] tmp;
//		delete[] ref;
	}
};

template<class T, class InStream, class MergeTree>
typename MergeTree::MergerLoc RecLazyMerger<T, InStream, MergeTree>::m_locMerger;
template<class T, class InStream, class MergeTree>
InStream* RecLazyMerger<T, InStream, MergeTree>::m_itFirstStream;


#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER
#ifdef __GNUC__
#undef inline
#endif // __GNUC__

#include "LazyMerger.timpl.h"

#endif // __LAZYMERGER_H_INCLUDED_
