/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __MERGETREE_H_INCLUDED_
#define __MERGETREE_H_INCLUDED_

#include <cmath>

#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER
#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__

extern unsigned long TreeOps;

template<class T, class It>
class EmptyBuffers
{
public:
	EmptyBuffers(It b, It e) : b(b), e(e) { }
	template<class Merger>
	inline void operator()(Merger* pm)
	{
		if( b == e )
			assert( pm->getHead() == pm->getTail() ); // This should at least be the case, when using Inplace Mergesort
		T *p;
		for( p = pm->getHead(); b!=e && p && p!=pm->getTail(); b++, p++ )
			*b = *p;
		if( p && b == e )
			assert( p == pm->getTail() ); // This should at least be the case, when using Inplace Mergesort
	}
	inline It begin() const
		{ return b; }
private:
	It b, e;
};

template<class T, class Merger>
class BinMergeTree
{
public:
	explicit inline BinMergeTree(size_t height) : m_pBase(new char[size(height)]), m_height(height)
	{
		Layout(height, m_pBase, NULL, 0, NULL, NULL);
//		LockMem(m_pBase, (char*)m_pBase+size(height));
		((Merger*)m_pBase)->set_root();
	}
	virtual ~BinMergeTree()
	{
//		UnlockMem(m_pBase, (char*)m_pBase+size(m_height));
		delete[] (char*)m_pBase;
//		free(m_pBase);
	}

	// Returns a pointer to the root Merger
	inline Merger* root()
		{ return (Merger*)m_pBase; }

	// Enumerates the mergers and calls f on each of them. Fun must have a void operator()(Merger*) member or (*f)(Merger*) must be void
	template<class Fun>
	inline Fun EnumMergers(Fun f)
		{ return EnumMergers(f, m_height, m_pBase); }


	static inline size_t out_buffer_size(size_t height)	
	{ 
		size_t s = 1u<<height;
#ifdef _CUBE_BUFFERS
		return (size_t)floor(pow((double)s, 3)+.5);
#else // _CUBE_BUFFERS
//		return s*s*s;
		return (size_t)floor(24*pow((double)s, 2.5)+.5);
#endif // _CUBE_BUFFERS
	}
	static inline size_t split(size_t height)
		{ return height/2; }
//		{ (void)height; return 1; }
//		{ return height-1; }
private:
	static size_t size(size_t height);
	static void Layout(size_t height, void* base, void* bottom_base, size_t bottom_size, T* buffer_begin, T* buffer_end)
	{
		if( height > 1 )
		{
			size_t top_height = split(height);
			size_t bottom_height = height-top_height;
			size_t mid_buffer_size = out_buffer_size(bottom_height)*sizeof(T);
			size_t top_tree_size = size(split(height));
			size_t bottom_tree_size = size(height-split(height));
			char* b = (char*)base;
			char* bb = (char*)bottom_base;

			Layout(top_height, b, b+top_tree_size+mid_buffer_size, bottom_tree_size+mid_buffer_size, buffer_begin, buffer_end);

			b += top_tree_size;
			for( size_t i=0; i<(1u<<top_height); i++, b+=bottom_tree_size+mid_buffer_size, bb+=(1u<<bottom_height)*bottom_size )
				Layout(bottom_height, b+mid_buffer_size, bb, bottom_size, (T*)b, (T*)(b+mid_buffer_size));
		}
		else
			new(base) Merger((Merger*)bottom_base, (Merger*)((char*)bottom_base+bottom_size), buffer_begin, buffer_end);
	}
	template<class Fun>
	static Fun EnumMergers(Fun f, size_t height, void* base)
	{
		if( height > 1 )
		{
			size_t top_height = split(height);
			size_t bottom_height = height-top_height;
			size_t mid_buffer_size = out_buffer_size(bottom_height)*sizeof(T);
			size_t top_tree_size = size(split(height));
			size_t bottom_tree_size = size(height-split(height));
			char* b = (char*)base + top_tree_size;
			for( size_t i=0; i<(1u<<top_height); i++, b+=bottom_tree_size+mid_buffer_size )
				f = EnumMergers(f, bottom_height, b+mid_buffer_size);

			return EnumMergers(f, top_height, (char*)base);
		}
		else
		{
			f((Merger*)base);
			return f;
		}
	}

	void* m_pBase;
	size_t m_height;
};

#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER
#ifdef __GNUC__
#undef inline
#endif // __GNUC__

#include "MergeTree.timpl.h"

#endif // __MERGETREE_H_INCLUDED_
