/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __SORT_H_INCLUDED_
#define __SORT_H_INCLUDED_

#include <memory>
#include <iterator>
#include <algorithm>
#include <functional>
#include "BasicMerger.h"
#include "funnel/stack_allocator.h"
#include "int64.h"

#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__
#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER

#include "Sort.aux.h"

namespace iosort
{

template<class RanIt>
void inplace_mergesort(RanIt begin, RanIt end)
{
	if( end-begin < base::RECURSELIMIT )
		base::__inplace_base_sort(begin, end);
	else
	{
		inplace_mergesort(begin, begin+(end-begin)/2);
		inplace_mergesort(begin+(end-begin)/2, end);
		std::inplace_merge(begin, begin+(end-begin)/2, end);
	}
}

template<class T, class Diff, class It, class OutIt>
void _mergesort(It begin, It end, OutIt dest)
{
	if( end-begin < base::RECURSELIMIT )
		base::__base_sort(begin, end, dest);
	else
	{
		Diff large_half = (end-begin+1)/2;
		Diff small_half = (end-begin)/2;
		assert( small_half+large_half == end-begin );
		std::pair<T*, size_t> tmp = std::get_temporary_buffer<T>((ptrdiff_t)large_half); // TODO: Support 64bit on 32bit arch
		assert( tmp.second == (size_t)large_half );

		_mergesort<T,Diff,It,T*>(begin+small_half, end, tmp.first);
		_mergesort<T,Diff,It,It>(begin, begin+small_half, begin+large_half);

		std::copy(tmp.first, tmp.first+large_half, begin);

		std::merge(begin, begin+large_half, begin+large_half, end, dest);
		std::return_temporary_buffer(tmp.first);
	}
}

template<class It, class T, class Diff, class OutIt>
inline void _mergesort(It begin, It end, T, Diff, OutIt dest)
{
	_mergesort<T, Diff, It, OutIt>(begin, end, dest);
}

template<class It, class OutIt>
inline void mergesort(It begin, It end, OutIt dest)
{
	_mergesort(begin, end, *begin, end-begin, dest);
}





template<class It>
class input_stream
{
public:
	typedef It iterator;
	typedef It pointer;

	inline input_stream() { }
	inline input_stream(const It& b, const It& e) : b(b), e(e) { }

	inline It begin()
		{ return b; }
	inline It end()
		{ return e; }

	inline It& head()
		{ return b; }
	inline const It& tail() const
		{ return e; }

private:
	It b, e;
};

template<class Splitter, class Diff>
inline Diff _run_size(Diff d)
{
	size_t order = Splitter::prefered_order_output(d);
	if( order < 2 )
		order = 2;
	Diff run_size = d/order;
	size_t rem = static_cast<size_t>(d % order);
	run_size += rem/order;
	size_t big_runs = rem % order;
	if( big_runs )
		run_size++;
	return run_size;
}

template<class Merger, class Splitter, class It, class ItDiff, class T>
inline typename Merger::allocator::size_type _tmp_space_needed(ItDiff d)
{
	if( d > (ItDiff)Splitter::switch_to_cache_aware() )
	{
		typedef typename Merger::stream Stream;
		size_t order = Splitter::prefered_order_output(d);
		if( order < 3 )
			return 0;
		ItDiff run_size = d/order;
		size_t rem = static_cast<size_t>(d % order);
		run_size += rem/order;
		size_t big_runs = rem % order;
		if( run_size > static_cast<ItDiff>(Splitter::switch_to_cache_aware()) )
		{
			if( big_runs )
				run_size++;
			ItDiff rs = run_size;
			typename Merger::allocator::size_type size = 0; 
			int rec_calls = 0;
			for( d=rs; d>(ItDiff)Splitter::switch_to_cache_aware(); d=_run_size<Splitter>(d), ++rec_calls )
				size += Merger::size_of(Splitter::prefered_order_output(d)) + d*sizeof(T);
			typename Merger::allocator::size_type root_merger_size = Merger::size_of(order);
			if( size < root_merger_size )
				size = root_merger_size;
			return size+rec_calls*sizeof(Merger);
		}
		return Merger::size_of(order);
	}
	return 0;
}
template<class Merger, class Splitter, class It, class ItDiff, class T>
inline typename Merger::allocator::size_type _tmp_space_needed(ItDiff d, T)
{
	return _tmp_space_needed<Merger,Splitter,It,ItDiff,T>(d);
}
template<class Merger, class Splitter, class It>
inline typename Merger::allocator::size_type tmp_space_needed(It begin, It end)
{
	return _tmp_space_needed<Merger,Splitter,It>(end-begin, *begin);
}


template<class Merger, class Splitter, class It, class OutIt, class T, class ItDiff>
void _merge_sort(It begin, It end, OutIt dest, Merger* m, typename Merger::allocator alloc)
{
	typedef typename Merger::stream Stream;
	size_t order = Splitter::prefered_order_output(end-begin);
	if( order < 2 )
		order = 2;
	ItDiff run_size = (end-begin)/order;
	size_t rem = (size_t)((end-begin) % order);
	run_size += rem/order;
	size_t big_runs = rem % order;

	m->reset();
	It last = begin;
	size_t i;
	if( run_size > static_cast<ItDiff>(Splitter::switch_to_cache_aware()) )
	{
		if( big_runs )
			run_size++;
		T* tmp = alloc.allocate((size_t)(run_size));
		if( big_runs )
		{
			_merge_sort<Merger,Splitter,It,T*,T,ItDiff>(last, last+run_size, tmp, m+1, alloc);
			for( i=1, last+=run_size; i!=big_runs; i++, last+=run_size )
				_merge_sort<Merger,Splitter,It,It,T,ItDiff>(last, last+run_size, last-run_size, m+1, alloc);
			run_size--;
			for( ; i!=order; i++, last+=run_size )
				_merge_sort<Merger,Splitter,It,It,T,ItDiff>(last, last+run_size, last-run_size-1, m+1, alloc);
			run_size++;
		}
		else
		{
			_merge_sort<Merger,Splitter,It,T*,T,ItDiff>(last, last+run_size, tmp, m+1, alloc);
			for( i=1, last+=run_size; i!=order; i++, last+=run_size )
				_merge_sort<Merger,Splitter,It,It,T,ItDiff>(last, last+run_size, last-run_size, m+1, alloc);
		}
		assert( last == end );
		std::copy(tmp, tmp+run_size, last-run_size);
		alloc.deallocate(tmp,static_cast<size_t>(run_size));
		if( big_runs )
		{
			for( i=0, last=begin; i!=big_runs-1; i++, last+=run_size )
				m->add_stream(Stream(last, last+run_size));
			run_size--;
			for( ; i!=order-1; i++, last+=run_size )
				m->add_stream(Stream(last, last+run_size));
			run_size++;
			m->add_stream(Stream(last, last+run_size));
			last+=run_size;
		}
		else
			for( i=0, last=begin; i!=order; i++, last+=run_size )
				m->add_stream(Stream(last, last+run_size));
		assert( last == end );
#ifdef UNLIMITED_MERGER_OUTPUT
		(*m)(dest);
#else // UNLIMITED_MERGER_OUTPUT
#ifdef NDEBUG
		(*m)(dest, dest+(end-begin));
#else // NDEBUG
		OutIt e = (*m)(dest, dest+(end-begin));
		assert( e == dest+(end-begin) );
#endif // NDEBUG
#endif // UNLIMITED_MERGER_OUTPUT
	}
	else
	{
		for( i=0; i!=order-big_runs; i++, last+=run_size )
		{
			base::__inplace_base_sort(last, last+run_size);
			m->add_stream(Stream(last, last+run_size));
		}
		run_size++;
		for( ; i!=order; i++, last+=run_size )
		{
			base::__inplace_base_sort(last, last+run_size);
			m->add_stream(Stream(last, last+run_size));
		}
#ifdef UNLIMITED_MERGER_OUTPUT
#ifdef NDEBUG
		(*m)(dest);
#else // NDEBUG
		OutIt e = (*m)(dest);
		assert( e == dest+(end-begin) );
#endif // NDEBUG
#else // UNLIMITED_MERGER_OUTPUT
#ifdef NDEBUG
		(*m)(dest, dest+(end-begin));
#else // NDEBUG
		OutIt e = (*m)(dest, dest+(end-begin));
		assert( e == dest+(end-begin) );
#endif // NDEBUG
#endif // UNLIMITED_MERGER_OUTPUT
	}
}

template<class Merger, class Splitter, class It, class OutIt, class T, class ItDiff>
void _merge_sort(It begin, It end, OutIt dest, typename Merger::allocator alloc)
{
	typedef typename Merger::stream Stream;
	size_t order = Splitter::prefered_order_output(end-begin);
	if( order < 2 )
		order = 2;
	ItDiff run_size = (end-begin)/order;
	size_t rem = static_cast<size_t>((end-begin) % order);
	run_size += rem/order;

	size_t big_runs = rem % order;
	It last = begin;
	size_t i;
	if( run_size > static_cast<ItDiff>(Splitter::switch_to_cache_aware()) )
	{
		if( big_runs )
			run_size++;
		typedef typename Merger::allocator Allocator;
		int rec_calls = 0;
		for( ItDiff d=run_size; d>static_cast<ItDiff>(Splitter::switch_to_cache_aware()); d=_run_size<Splitter>(d), ++rec_calls );
		typename Allocator::template rebind<Merger>::other malloc(alloc);
		Merger *begin_cache = malloc.allocate(rec_calls), *end_cache = begin_cache;
		for( ItDiff d=run_size; d>static_cast<ItDiff>(Splitter::switch_to_cache_aware()); d=_run_size<Splitter>(d), ++end_cache )
			new(end_cache) Merger(Splitter::prefered_order_output(d), alloc);

		T* tmp = alloc.allocate(static_cast<typename Merger::allocator::size_type>(run_size));
		if( big_runs )
		{
			_merge_sort<Merger,Splitter,It,T*,T,ItDiff>(last, last+run_size, tmp, begin_cache, alloc);
			for( i=1, last+=run_size; i!=big_runs; i++, last+=run_size )
				_merge_sort<Merger,Splitter,It,It,T,ItDiff>(last, last+run_size, last-run_size, begin_cache, alloc);
			run_size--;
			for( ; i!=order; i++, last+=run_size )
				_merge_sort<Merger,Splitter,It,It,T,ItDiff>(last, last+run_size, last-run_size-1, begin_cache, alloc);
			run_size++;
		}
		else
		{
			_merge_sort<Merger,Splitter,It,T*,T,ItDiff>(last, last+run_size, tmp, begin_cache, alloc);
			for( i=1, last+=run_size; i!=order; i++, last+=run_size )
				_merge_sort<Merger,Splitter,It,It,T,ItDiff>(last, last+run_size, last-run_size, begin_cache, alloc);
		}
		assert( last == end );
		std::copy(tmp, tmp+run_size, last-run_size);
		alloc.deallocate(tmp,static_cast<typename Merger::allocator::size_type>(run_size));
		for( Merger *pm=end_cache-1; pm>=begin_cache; --pm )
			malloc.destroy(pm);
		malloc.deallocate(begin_cache, rec_calls);

		Merger merger(order,alloc);
		if( big_runs )
		{
			for( i=0, last=begin; i!=big_runs-1; i++, last+=run_size )
				merger.add_stream(Stream(last, last+run_size));
			run_size--;
			for( ; i!=order-1; i++, last+=run_size )
				merger.add_stream(Stream(last, last+run_size));
			run_size++;
			merger.add_stream(Stream(last, last+run_size));
			last+=run_size;
		}
		else
			for( i=0, last=begin; i!=order; i++, last+=run_size )
				merger.add_stream(Stream(last, last+run_size));
		assert( last == end );
#ifdef UNLIMITED_MERGER_OUTPUT
		merger(dest);
#else // UNLIMITED_MERGER_OUTPUT
#ifdef NDEBUG
		merger(dest, dest+(end-begin));
#else // NDEBUG
		OutIt e = merger(dest, dest+(end-begin));
		assert( e == dest+(end-begin) );
#endif // NDEBUG
#endif // UNLIMITED_MERGER_OUTPUT
	}
	else
	{
		Merger merger(order, alloc);
		for( i=0; i!=order-big_runs; i++, last+=run_size )
		{
			base::__inplace_base_sort(last, last+run_size);
			merger.add_stream(Stream(last, last+run_size));
		}
		run_size++;
		for( ; i!=order; i++, last+=run_size )
		{
			base::__inplace_base_sort(last, last+run_size);
			merger.add_stream(Stream(last, last+run_size));
		}
#ifdef UNLIMITED_MERGER_OUTPUT
		merger(dest);
#else // UNLIMITED_MERGER_OUTPUT
#ifdef NDEBUG
		merger(dest, dest+(end-begin));
#else // NDEBUG
		OutIt e = merger(dest, dest+(end-begin));
		assert( e == dest+(end-begin) );
#endif // NDEBUG
#endif // UNLIMITED_MERGER_OUTPUT
	}
}

template<class Merger, class Splitter, class It, class OutIt, class T, class Diff>
inline void _merge_sort(It begin, It end, OutIt dest, T, Diff d)
{
	if( d > static_cast<Diff>(Splitter::switch_to_cache_aware()) )
	{
		stack_allocator<T> alloc(tmp_space_needed<Merger,Splitter>(begin,end));
		_merge_sort<Merger,Splitter,It,OutIt,T,Diff>(begin, end, dest, alloc);
	}
	else
		base::__base_sort(begin, end, dest);
}
template<class Merger, class Splitter, class It, class OutIt, class T, class Diff>
inline void _merge_sort_a(It begin, It end, OutIt dest, T, Diff d, typename Merger::allocator& alloc)
{
	if( d > static_cast<Diff>(Splitter::switch_to_cache_aware()) )
		_merge_sort<Merger,Splitter,It,OutIt,T,Diff>(begin, end, dest, alloc);
	else
		base::__base_sort(begin, end, dest);
}
template<class Merger, class Splitter, class It, class OutIt, class Allocator>
inline void _merge_sort_pt(It begin, It end, OutIt dest)
{
	Allocator alloc;
	_merge_sort<Merger,Splitter,It,OutIt,typename std::iterator_traits<It>::value_type,typename std::iterator_traits<It>::difference_type>(begin, end, dest, alloc);
}
template<class Allocator>
class allocator_factory
{
public:
	inline Allocator operator()(typename Allocator::size_type size)
		{ return Allocator(); }
};
template<class T>
class allocator_factory<stack_allocator<T> >
{
public:
	inline stack_allocator<T> operator()(typename stack_allocator<T>::size_type size)
		{ return stack_allocator<T>(size); }
};
template<class Merger, class Splitter, class It, class OutIt>
inline void merge_sort(It begin, It end, OutIt dest)
{
	if( end-begin > static_cast<typename std::iterator_traits<It>::difference_type>(Splitter::switch_to_cache_aware()) )
	{
		iosort::allocator_factory<typename Merger::allocator> af;
		_merge_sort<Merger,Splitter,It,OutIt,typename std::iterator_traits<It>::value_type,typename std::iterator_traits<It>::difference_type>(begin, end, dest, af(tmp_space_needed<Merger,Splitter>(begin,end)));
	}
	else
		base::__base_sort(begin, end, dest);
}
template<class Merger, class Splitter, class It, class OutIt>
inline void merge_sort(It begin, It end, OutIt dest, typename Merger::allocator& alloc)
{
	_merge_sort_a<Merger,Splitter>(begin, end, dest, *begin, end-begin, alloc);
}





template<class It>
class LimitedInplaceRefill
{
public:
	inline LimitedInplaceRefill() { }
	inline LimitedInplaceRefill(std::pair<It,It>* ss, It n, It l) : next(n), last(l), snapshot(ss) { }
	inline It get_next() const
		{ return next; }
	template<class Merger>
	inline bool operator()(Merger* pm, It b, It e)
	{
		// Type check: Merger::Stream::iterator == It
		if( e-b >= last-next )
		{
			if( next == last )
				return false;
			shoot(pm,b,e);
			e = b+(last-next); // only copy last-next elements
		}
		while( b != e )
			*b = *next, ++b, ++next;
		return false;
	}
private:
	template<class Merger>
	inline void shoot(Merger* pm, It b, It e)
	{
		for( typename Merger::stream_iterator sit=pm->begin(); sit!=pm->end(); ++sit, ++snapshot )
		{
			if( sit->begin() == e )
			{
				snapshot->first = b+(last-next);
				for( ++sit, ++snapshot; sit!=pm->end(); ++sit, ++snapshot )
				{
//					assert( sit->begin() != e );
					snapshot->first = sit->begin();
				}
				return;
			}
			else
				snapshot->first = sit->begin();
		}
		assert( false );
	}
	It next;
	It last;
	std::pair<It,It>* snapshot;
};

template<class It, class T>
class GapIterator
{
public:
	template<class Merger>
	inline static std::pair<GapIterator,GapIterator> beginend(Merger *pm, std::pair<typename Merger::stream::pointer,typename Merger::stream::pointer>* ss)
	{
		// Type check: Merger::Stream::iterator == It
		std::pair<GapIterator,GapIterator> ret;
		ret.first.snapshot = ret.second.snapshot = ss;

		size_t i=0, j=0;
		for( typename Merger::stream_iterator sit=pm->begin(); sit!=pm->end(); ++sit, ++j )
			if( ss[j].first != sit->begin() )
			{
				assert( i <= j );
				ss[i].first = ss[j].first;
				ss[i].second = sit->begin();
				assert(!( ss[i].second < ss[i].first ));
				++i;
			}
		ret.first.stream = 0;
		ret.first.current = ss[0].first;
		ret.second.stream = i;
		ret.second.current = ss[i].first;

		return ret;
	}
	// Uncomment second condition if you ever compare an iterator to a non-end iterator
	inline bool operator==(const GapIterator& it) const
		{ return stream == it.stream /*&& current==it.current*/; }
	inline bool operator!=(const GapIterator& it) const
		{ return stream != it.stream /*|| current!=it.current*/; }
	inline GapIterator operator++()
	{
		if( ++current == snapshot[stream].second )
			current = snapshot[++stream].first;
		return *this;
	}
	inline GapIterator operator++(int)
	{
		GapIterator gi = *this;
		if( ++current == snapshot[stream].second )
			current = snapshot[++stream].first;
		return gi;
	}
	inline T& operator*()
		{ return *current; }
private:
	It current;
	std::pair<It,It>* snapshot;
	size_t stream;
};

template<class It>
class InplaceRefill
{
public:
	inline InplaceRefill() { }
	inline InplaceRefill(It n) : next(n) { }
	inline It get_next() const
		{ return next; }
	template<class Merger>
	inline bool operator()(const Merger*, It b, It e)
	{
		while( b != e )
			*b = *next, ++b, ++next;
		return false;
	}
private:
	It next;
};

template<class Merger, class Splitter, class It, class ItDiff, class T>
inline typename Merger::allocator::size_type _inplace_tmp_space_needed(ItDiff d, T)
{
	if( d < (ItDiff)Splitter::switch_to_cache_aware() )
		return 0;
	else
	{
		typename Merger::allocator::size_type size = 0;
		ItDiff limit = _run_size<Splitter>(d);
		if( limit < (ItDiff)Splitter::switch_to_cache_aware() )
			limit = (ItDiff)Splitter::switch_to_cache_aware();

		ItDiff subsize = (d/2)/(ItDiff)Splitter::prefered_order_output((d/2));
		size += sizeof(T)*subsize;
		size += _tmp_space_needed<Merger,Splitter,It,ItDiff,T>(subsize);

		if( d > limit )
		{
			size_t order = static_cast<size_t>((d+subsize-1)/subsize);
			typename Merger::allocator::size_type msize = sizeof(std::pair<It,It>)*order;
			msize += Merger::size_of(order);
			if( msize > size )
				size = msize;
			typename Merger::allocator::size_type tsize = sizeof(T)*subsize*Merger::min_order(); 
			if( tsize > size )
				size = tsize;
		}
		if( limit > (ItDiff)Splitter::switch_to_cache_aware() )
			if( sizeof(T)*limit > size )
				size = sizeof(T)*limit;
		return size;
	}
}
template<class Merger, class Splitter, class It>
inline typename Merger::allocator::size_type inplace_tmp_space_needed(It begin, It end)
{
	return _inplace_tmp_space_needed<Merger,Splitter,It>(end-begin, *begin);
}

template<class T, class It, class ItDiff>
inline It ip_partition(It first, It last, ItDiff d)
{
	It unsorted = last;
	T pivot = *base::__find_rank<It, 32>(first, last, 14);
	for( ;; )
	{
		while( pivot < *first )
			first++;
		last--;
		while( *last < pivot )
			last--;
		for( ; (ItDiff)(unsorted-last)>d; unsorted-=d )
			base::__inplace_base_sort(unsorted-d, unsorted);
		if( !(first<last) )
		{
			base::__inplace_base_sort(first, unsorted);
			return first;
		}
		std::iter_swap(first, last);
		first++;
	}
}

template<class T, class Merger, class Splitter, class It, class ItDiff>
inline It ip_partition(It first, It last, ItDiff d, typename Merger::allocator& alloc)
{
	typedef typename Merger::allocator Alloc;/*MSCW*/
	typename Alloc::template rebind<T>::other talloc(alloc);
	T* tmp = talloc.allocate(d);
	It unsorted = last;
	bool washere = false;
	T pivot = *base::__find_rank<It, 32>(first, last, 14);
	for( ;; )
	{
		while( pivot < *first )
			first++;
		last--;
		while( *last < pivot )
			last--;
		for( ; (ItDiff)(unsorted-last) > d;  )
		{
			if( washere )
				merge_sort<Merger,Splitter>(unsorted-d, unsorted, unsorted, alloc);
			else
				merge_sort<Merger,Splitter>(unsorted-d, unsorted, tmp, alloc), washere = true;
			unsorted -= d;
		}
		if( !(first<last) )
		{
			if( washere )
				std::copy(tmp, tmp+d, unsorted);
			merge_sort<Merger,Splitter>(first, unsorted, tmp, alloc);
			std::copy(tmp, tmp+(unsorted-first), first);
			talloc.deallocate(tmp,d);
			return first;
		}
		std::iter_swap(first, last);
		first++;
	}
}

template<class Merger, class It, class ItDiff>
inline void set_input_streams(Merger& m, It end, It part, size_t order, ItDiff subsize)
{
	typedef typename Merger::stream Stream;
	size_t s = order-1;
	It p = end;
	for( ; subsize<p-part; s--, p-=subsize )
	{
		assert( s );
		m.add_stream(Stream(p-subsize,p));
	}
	if( p != part )
	{
		assert( s == 0 );
		assert( p-part <= subsize );
		m.add_stream(Stream(part,p));
	}
	else
		assert( s < 0 );
}

template<class Merger, class Splitter, class It, class T, class ItDiff>
void _inplace_merge_sort(It begin, It end, ItDiff limit, typename Merger::allocator alloc)
{
	typedef typename Merger::allocator Alloc; /*MSCW*/
	typename Alloc::template rebind<T>::other talloc(alloc);
	typename Alloc::template rebind<std::pair<It,It> >::other ssalloc(alloc);

	ItDiff half = (end-begin)/2;
	ItDiff subsize = half/(ItDiff)Splitter::prefered_order_output(half);
	std::pair<It,It>* snapshot = NULL;
	size_t ssorder = 0;
	while( end-begin > limit )
	{
		It part;
		if( subsize > (ItDiff)Splitter::switch_to_cache_aware() )
			part = ip_partition<T, Merger, Splitter>(begin, end, subsize, alloc);
		else
			part = ip_partition<T>(begin, end, subsize);
		It newbegin;
		if( part-begin < end-part )
			newbegin = begin+(part-begin);
		else
			newbegin = begin+(end-part);
		if( end-part > (ItDiff)Splitter::switch_to_cache_aware() )
		{
			size_t order = (size_t)(((end-part)+subsize-1)/subsize);
			if( order > static_cast<size_t>(Merger::min_order()) )
				if( part-begin < end-part )
				{
					if( snapshot == NULL )
						ssorder = (size_t)(((end-begin)+subsize-1)/subsize), snapshot = ssalloc.allocate(ssorder);
					typename Merger::template rebind<LimitedInplaceRefill<It> >::other merger(order,alloc);
					set_input_streams(merger, end, part, order, subsize);
					merger.set_refiller(LimitedInplaceRefill<It>(snapshot, begin, newbegin));
					merger(begin, newbegin);
					std::pair<GapIterator<It,T>,GapIterator<It,T> > be = GapIterator<It,T>::beginend(&merger, snapshot);
					GapIterator<It,T> gi = merger.empty(be.first, be.second);
					/*assert( gi == be.second );*/ (void)gi;
				}
				else
				{
					typename Merger::template rebind<InplaceRefill<It> >::other merger(order,alloc);
					set_input_streams(merger, end, part, order, subsize);
					merger.set_refiller(InplaceRefill<It>(begin));
					merger(begin, newbegin);
				}
			else
			{
				assert( end-part <= static_cast<ItDiff>(subsize*Merger::min_order()) );
				for( It b=begin, e=end-1; b!=newbegin; b++, e-- )
					std::iter_swap(b, e);
				T* tmp = talloc.allocate(end-part);
				merge_sort<Merger,Splitter>(begin, begin+(end-part), tmp);
				std::copy(tmp, tmp+(end-part), begin);
				talloc.deallocate(tmp,end-part);
			}
		}
		else
		{
			for( It b=begin, e=end-1; b!=newbegin; b++, e-- )
				std::iter_swap(b, e);
			base::__inplace_base_sort(begin, begin+(end-part));
		}
		begin = newbegin;
		half = (end-begin)/2;
		subsize = half/(ItDiff)Splitter::prefered_order_output(half);
	}
	if( snapshot )
		ssalloc.deallocate(snapshot, ssorder);
	if( end-begin > (ItDiff)Splitter::switch_to_cache_aware() )
	{
		T* tmp = talloc.allocate(end-begin);
		merge_sort<Merger,Splitter>(begin, end, tmp, alloc);
		std::copy(tmp, tmp+(end-begin), begin);
		talloc.deallocate(tmp, end-begin);
	}
	else
		base::__inplace_base_sort(begin, end);
}


template<class Merger, class Splitter, class It, class T, class Diff>
inline void _inplace_merge_sort(It begin, It end, T, Diff d)
{
	if( d < (Diff)Splitter::switch_to_cache_aware() )
		base::__inplace_base_sort(begin, end);
	else
	{
		typedef typename Merger::allocator Alloc; /*MSCW*/
		Alloc alloc(inplace_tmp_space_needed<Merger,Splitter>(begin,end));
		Diff run_size = _run_size<Splitter>(d);
		if( run_size < (Diff)Splitter::switch_to_cache_aware() )
			run_size = (Diff)Splitter::switch_to_cache_aware();
		_inplace_merge_sort<Merger,Splitter,It,T,Diff>(begin, end, run_size, alloc);
	}
}

template<class Merger, class Splitter, class It>
inline void inplace_merge_sort(It begin, It end)
{
	_inplace_merge_sort<Merger,Splitter>(begin, end, *begin, end-begin);
}


template<class Merger, class Splitter, class It, class OutIt>
inline void bu_merge_sort(It begin, It end, OutIt dest)
{
	typedef typename std::iterator_traits<It>::difference_type Diff;

	Diff N = end-begin;
	Diff n = Splitter::switch_to_cache_aware();
	It next=begin;
	for( ; next+n<end; next+=n )
		base::__base_sort(next, next+n);


	while( N > Splitter::prefered_order(N) )
	{

	}
}









template<class Merger, class It, class OutIt, class T, class ItDiff>
void _MergeSort(It begin, It end, OutIt dest)
{
	typedef typename Merger::stream Stream;
	static typename Merger::cache mcache;
	size_t order = Merger::prefered_order(end-begin);
	if( order < 2 )
		order = 2;
	ItDiff run_size = (end-begin)/order;
	size_t rem = (size_t)((end-begin) % order);
	run_size += rem/order;

	if( order > 2 )
	{
		Merger *pM;
		size_t big_runs = rem % order;
		It last = begin;
		size_t i;
		if( run_size+1 < (ItDiff)Merger::switch_to_cache_aware() )
		{
			pM = mcache.get_merger(order);
			for( i=0; i!=order-big_runs; i++, last+=run_size )
			{
				base::__inplace_base_sort(last, last+run_size);
				pM->add_stream(Stream(last, last+run_size));
			}
			run_size++;
			for( ; i!=order; i++, last+=run_size )
			{
				base::__inplace_base_sort(last, last+run_size);
				pM->add_stream(Stream(last, last+run_size));
			}
		}
		else
		{
			if( big_runs )
				run_size++;
			T* tmp = new T[(size_t)(run_size)];
			if( tmp == NULL /*|| run_size >> (8*sizeof(void*))*/ )
			{
				std::cerr << "Failure to allocate temporary array in MergeSort()" << std::endl;
				exit(EXIT_FAILURE);
			}
			if( big_runs )
			{
				_MergeSort<Merger,It,T*,T,ItDiff>(last, last+run_size, tmp);
				for( i=1, last+=run_size; i!=big_runs; i++, last+=run_size )
					_MergeSort<Merger,It,It,T,ItDiff>(last, last+run_size, last-run_size);
				run_size--;
				for( ; i!=order; i++, last+=run_size )
					_MergeSort<Merger,It,It,T,ItDiff>(last, last+run_size, last-run_size-1);
				run_size++;
			}
			else
			{
				_MergeSort<Merger,It,T*,T,ItDiff>(last, last+run_size, tmp);
				for( i=1, last+=run_size; i!=order; i++, last+=run_size )
					_MergeSort<Merger,It,It,T,ItDiff>(last, last+run_size, last-run_size);
			}
			assert( last == end );
			std::copy(tmp, tmp+run_size, last-run_size);
			delete[] tmp;
			pM = mcache.get_merger(order);

			if( big_runs )
			{
				for( i=0, last=begin; i!=big_runs-1; i++, last+=run_size )
					pM->add_stream(Stream(last, last+run_size));
				run_size--;
				for( ; i!=order-1; i++, last+=run_size )
					pM->add_stream(Stream(last, last+run_size));
				run_size++;
				pM->add_stream(Stream(last, last+run_size));
				last+=run_size;
			}
			else
				for( i=0, last=begin; i!=order; i++, last+=run_size )
					pM->add_stream(Stream(last, last+run_size));
			assert( last == end );
		}
#ifdef NDEBUG
		(*pM)(dest, dest+(end-begin));
#else // NDEBUG
		OutIt e = (*pM)(dest, dest+(end-begin));
		assert( e == dest+(end-begin) );
#endif // NDEBUG
	}
	else
	{
		It mid(begin+run_size);
		base::__inplace_base_sort(begin, mid);
		base::__inplace_base_sort(begin+run_size, end);
		LazyMerge(begin, mid, mid, end, dest, dest+(end-begin));
	}
}

template<class Merger, class It, class OutIt, class T, class Diff>
inline void _MergeSort(It begin, It end, OutIt dest, T, Diff d)
{
	if( d < (Diff)Merger::switch_to_cache_aware() )
		base::__base_sort(begin, end, dest);
	else
		_MergeSort<Merger,It,OutIt,T,Diff>(begin, end, dest);
}
template<class Merger, class It, class OutIt>
inline void MergeSort(It begin, It end, OutIt dest)
{
	_MergeSort<Merger>(begin, end, dest, *begin, end-begin);
}

template<class T, class Merger, class It, class ItDiff>
inline It __ip_partition(It first, It last, ItDiff d, T* tmp)
{
	assert(!( last < first ));
	It unsorted = last;
	bool washere = false;
	T pivot = *base::__find_rank<It, 32>(first, last, 14);
	for( ;; )
	{
		while( pivot < *first )
			first++;
		last--;
		while( *last < pivot )
			last--;
		while( (ItDiff)(unsorted-last) > d )
		{
			if( washere )
				MergeSort<Merger>(unsorted-d, unsorted, unsorted);
			else
				MergeSort<Merger>(unsorted-d, unsorted, tmp), washere = true;
			unsorted -= d;
		}
		if( !(first<last) )
		{
			if( washere )
				std::copy(tmp, tmp+d, unsorted);
			return first;
		}
		std::iter_swap(first, last);
		first++;
	}
}

template<class RecMerger, class Merger, class It, class ItDiff, class T>
inline void _SetInputStreams(Merger& m, It end, It part, size_t order, ItDiff subsize, T* tmp)
{
	typedef typename Merger::stream Stream;
	size_t s = order-1;
	It p = end;
	for( ; subsize<p-part; s--, p-=subsize )
	{
		assert( s );
		m.add_stream(Stream(p-subsize,p));
	}
	if( p != part )
	{
		assert( s == 0 );
		assert( p-part <= subsize );
		MergeSort<RecMerger>(part, p, tmp);
		std::copy(tmp, tmp+(p-part), part);
		m.add_stream(Stream(part,p));
	}
	else
		assert( s < 0 );
}

template<class Merger, class It, class T, class ItDiff>
void _InplaceMergeSort(It begin, It end, ItDiff limit)
{
	assert(!( end < begin ));
	ItDiff half = (end-begin)/2;
	ItDiff subsize = half/(ItDiff)Merger::prefered_order(half);
	size_t tmpsize = (size_t)subsize+Merger::prefered_order(half)+1;  // What if... log(subsize)>32
	T* tmp = new T[tmpsize];
	std::pair<It,It>* snapshot = NULL;
	while( end-begin > limit )
	{
		It part = __ip_partition<T, Merger>(begin, end, subsize, tmp);
		It newbegin;
		if( part-begin < end-part )
			newbegin = begin+(part-begin);
		else
			newbegin = begin+(end-part);
		if( end-part < (ItDiff)Merger::switch_to_cache_aware() )
		{
			for( It b=begin, e=end-1; b!=newbegin; b++, e-- )
				std::iter_swap(b, e);
			base::__inplace_base_sort(begin, begin+(end-part));
		}
		else
		{
			size_t order = (size_t)(((end-part)+subsize-1)/subsize);
			if( order > 2 )
				if( part-begin < end-part )
				{
					if( snapshot == NULL )
						snapshot = new std::pair<It,It>[2*(size_t)(((end-begin)+subsize-1)/subsize)]; // HACK: Multiplying by two, because LazyMerger(k) might have 2*k input streams
					typename Merger::template rebind<LimitedInplaceRefill<It> >::other merger(1<<_clog(order));
					_SetInputStreams<Merger>(merger, end, part, order, subsize, tmp);
					merger.set_refiller(LimitedInplaceRefill<It>(snapshot, begin, newbegin));
					merger(begin, newbegin);
					std::pair<GapIterator<It,T>,GapIterator<It,T> > be = GapIterator<It,T>::beginend(&merger, snapshot);
					GapIterator<It,T> gi = merger.empty(be.first, be.second);
					/*assert( gi == be.second );*/ (void)gi;
				}
				else
				{
					typename Merger::template rebind<InplaceRefill<It> >::other merger(1<<_clog(order));
					_SetInputStreams<Merger>(merger, end, part, order, subsize, tmp);
					merger.set_refiller(InplaceRefill<It>(begin));
					merger(begin, newbegin);
				}
			else if( order == 2 )
			{
				for( It b=begin, e=end-1; b!=newbegin; b++, e-- )
					std::iter_swap(b, e);
				if( tmpsize < (size_t)(end-part) )
				{
					assert( (size_t)(end-part) < 2*tmpsize );
					tmpsize = (size_t)(end-part);
					delete[] tmp;
					tmp = new T[tmpsize];
				}
				MergeSort<Merger>(begin, begin+(end-part), tmp);
				std::copy(tmp, tmp+(end-part), begin);
			}
			else
			{
				assert( end-part <= subsize );
				for( It b=begin, e=end-1; b!=newbegin; b++, e-- )
					std::iter_swap(b, e);
				MergeSort<Merger>(begin, begin+(end-part), tmp);
				std::copy(tmp, tmp+(end-part), begin);
			}
		}
//		std::cerr << ' ' << std::setprecision(3) << (float)(part-begin)/(end-begin);
		begin = newbegin;
		half = (end-begin)/2;
		subsize = half/(ItDiff)Merger::prefered_order(half);
	}
	delete[] snapshot;
	delete[] tmp;
	if( end-begin > (ItDiff)Merger::switch_to_cache_aware() )
	{
		T* tmp = new T[(size_t)(end-begin)]; // TODO: What if... log(limit) > 32?
		MergeSort<Merger>(begin, end, tmp);
		std::copy(tmp, tmp+(end-begin), begin);
		delete[] tmp;
	}
	else
		base::__inplace_base_sort(begin, end);
}

template<class Merger, class It, class T, class Diff>
inline void _InplaceMergeSort(It begin, It end, T, Diff d)
{
	if( d < (Diff)Merger::switch_to_cache_aware() )
		base::__inplace_base_sort(begin, end);
	else
	{
		size_t order = Merger::prefered_order(d);
		if( order < 2 )
			order = 2;
		Diff run_size = d/order;
		size_t rem = (size_t)(d % order);
		run_size += rem/order;
		if( run_size < (Diff)Merger::switch_to_cache_aware() )
			run_size = (Diff)Merger::switch_to_cache_aware();
		_InplaceMergeSort<Merger,It,T,Diff>(begin, end, run_size);
	}
}

template<class Merger, class It>
inline void InplaceMergeSort(It begin, It end)
{
	_InplaceMergeSort<Merger>(begin, end, *begin, end-begin);
}



#define __SWAP(x,y) { tmp=dest[x]; dest[x]=dest[y]; dest[y]=tmp; }
#define __PUSH(b,e) { bStack[stackHeight] = b; eStack[stackHeight] = e; stackHeight++; }
#define __POP() { stackHeight--; b = bStack[stackHeight]; e = eStack[stackHeight]; }
template<class T, class ArrayClass>
void QuickSort(ArrayClass dest, _uint64_t n)
{
	static const int STACKSIZE = 128*sizeof(_uint64_t) - 2; // HACK: Stack could grow heigher...
	T tmp;
	_uint64_t b, e, s, p, mid;
#ifdef _DEBUG
	int maxStackHeight = 0;
#endif // _DEBUG
	int stackHeight = 0;
	_uint64_t bStack[STACKSIZE];
	_uint64_t eStack[STACKSIZE];
	__PUSH(0, n-1);
	while( stackHeight )
	{
		__POP()
		if( b < e )
		{
#ifdef _DEBUG
//			std::cout << std::endl;
//			for( int i=0; i<stackHeight; i++ )
//				std::cout << ' ';
//			std::cout << " sorting [" << b << ';' << e << ']';
#endif // _DEBUG
			s = b;
			p = e; e--;
			if( b < e )
			{
				if( e-b > 11 ) // Take time to find a good median
				{
					mid = (b+e)/2;
					if( dest[b] < dest[mid] )
						__SWAP(mid, b)
					if( dest[mid] < dest[e] )
						__SWAP(mid, e)
				}
				if( dest[e] < dest[b] )
					__SWAP(b, e)
				if( dest[e] < dest[p] )
					__SWAP(p, e)
				if( dest[p] < dest[b] )
					__SWAP(b, p)
				while( b < e )
				{
					assert( dest[b] <= dest[p] && dest[p] <= dest[e] );
					while( b < e && !(dest[p] < dest[b]) )
						b++;
					while( b < e && !(dest[e] < dest[p]) )
						e--;
					assert( b==e || dest[e] < dest[p] && dest[p] < dest[b] );
					if( b < e )
						__SWAP(b, e)
					assert( b==e || dest[b] < dest[p] && dest[p] < dest[e] );
				}
//				assert( b == e && (dest[b]<dest[p] && dest[p]<dest[e+1] || dest[b-1]<dest[p] && dest[p]<dest[e]) );
				__SWAP(b, p)
				if( b > s )
					__PUSH(s, b-1)
				if( p > e+1 )
					__PUSH(e+1, p)
				assert( stackHeight < STACKSIZE );
#ifdef _DEBUG
				if( maxStackHeight < stackHeight )
					maxStackHeight = stackHeight;
#endif // _DEBUG
			}
			else // Two elements
				if( dest[p] < dest[e] )
					__SWAP(p, e)
		}
	}
#ifdef _DEBUG
//	std::cout << std::endl << "{max stackheight: " << maxStackHeight << '}';
#endif // _DEBUG
}

#undef __SWAP
#undef __PUSH
#undef __POP

} // namespace iosort

#ifdef __GNUC__
#undef inline
#endif // __GNUC__
#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __SORT_H_INCLUDED_
