/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __LAYOUT_BASE_H_INCLUDED_
#define __LAYOUT_BASE_H_INCLUDED_

#include "funnel_base.h"
#include <memory> // for Allocator parameter default

#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__
#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER


namespace funnel
{

template<class Navigator, class Splitter, class T, class Allocator=std::allocator<T> >
class layout_base
{
public:
	typedef Navigator navigator;
	typedef Allocator allocator;
	typedef typename Navigator::node node;
protected:
	static inline typename Allocator::pointer new_buffer(size_t size, Allocator& alloc)
	{
		typename Allocator::pointer ret = alloc.allocate(size);
		construct_buffer(ret, ret+size, alloc, *ret);
		return ret;
	}
	static inline void delete_buffer(typename Allocator::pointer b, typename Allocator::pointer e, Allocator& alloc)
	{
		for( typename Allocator::pointer p=b; p<e; ++p )
			alloc.destroy(p);
		alloc.deallocate(b, e-b);
	}
	static void compute_buffer_sizes(size_t *b, size_t *e)
	{
		height_t sp = Splitter::split(static_cast<height_t>(e-b));
		if( e-b > 2 )
		{
			compute_buffer_sizes(b, b+sp);
			compute_buffer_sizes(b+sp, e);
		}
		b[sp] = Splitter::out_buffer_size(pow_of_order<Navigator::node::order>(static_cast<height_t>(e-b-sp)));
	}
	static typename Allocator::size_type size_of(order_t k, height_t height)
	{
		if( height > 1 )
		{
			height_t top_height = Splitter::split(height);
			height_t bottom_height = height-top_height;
			order_t bottom_tree_k = pow_of_order<Navigator::node::order>(bottom_height);
			order_t top_tree_k = (k+bottom_tree_k-1)/bottom_tree_k;
			size_t mid_buffer_size = Splitter::out_buffer_size(bottom_tree_k);

			typename Allocator::size_type size = size_of(top_tree_k, top_height);
			for( ; k>=bottom_tree_k; k=static_cast<order_t>(k-bottom_tree_k) )
				size += mid_buffer_size*sizeof(T) + size_of(bottom_tree_k, bottom_height);
			if( k )
				size += mid_buffer_size*sizeof(T) + size_of(k, bottom_height);
/*			while( k > 0 )
				for( basic_order_t n=0; k>0 && n!=Navigator::node::order; ++n, k=static_cast<order_t>(k-bottom_tree_k) )
					size += mid_buffer_size*sizeof(T) + size_of(k<bottom_tree_k? k : bottom_tree_k, bottom_height);
*/
			return size;
		}
		else
			return sizeof(typename Navigator::node);
	}

	static inline void construct_buffer(typename Allocator::pointer p, typename Allocator::pointer e, Allocator& alloc, T)
	{
		for( T dflt; p<e; ++p )
			alloc.construct(p,dflt);
		(void)alloc;
	}
};


template<class Navigator, class Splitter, class T, class Allocator=std::allocator<T> >
class mixed_layout_base : public layout_base<Navigator,Splitter,T,Allocator>
{
protected:
	using layout_base<Navigator,Splitter,T,Allocator>::construct_buffer;
	using layout_base<Navigator,Splitter,T,Allocator>::compute_buffer_sizes;
	static inline typename Allocator::pointer new_buffer(size_t size, Allocator& alloc)
	{
		typename Allocator::pointer ret = alloc.allocate(size);
		construct_buffer(ret, ret+size, alloc, *ret);
		return ret;
	}
	static inline void delete_buffer(typename Allocator::pointer b, typename Allocator::pointer e, Allocator& alloc)
	{
		for( typename Allocator::pointer p=b; p<e; ++p )
			alloc.destroy(p);
		alloc.deallocate(b, e-b);
	}
};


template<class Navigator, class Splitter, class T, class Allocator=std::allocator<T> >
class pooled_layout_base : public layout_base<Navigator,Splitter,T,Allocator>
{
public:
	static typename Allocator::size_type size_of(order_t k)
		{ return size_of(k, logc<Navigator::node::order>(k)); }

protected:
	using layout_base<Navigator,Splitter,T,Allocator>::compute_buffer_sizes;

	class pool_pointer
	{
	public:
		typedef typename Allocator::pointer pointer;
		pointer p;
	};

	static typename Allocator::size_type size_of(order_t k, height_t height)
		{ return buffer_space(k, height)*sizeof(T) + (pow_of_order<Navigator::node::order>(height)-1)*sizeof(typename Navigator::node)/(Navigator::node::order-1); }

	static inline typename Allocator::pointer new_buffer(size_t size, pool_pointer& alloc)
	{
		typename Allocator::pointer ret = alloc.p;
		alloc.p += size;
		return ret;
	}
	static inline void delete_buffer(typename Allocator::pointer, typename Allocator::pointer, pool_pointer&)
		{ }
	
	static inline pool_pointer new_pool(order_t k, Allocator& alloc)
	{
		pool_pointer pp;
		size_t bs = buffer_space(k, logc<Navigator::node::order>(k));
		pp.p = alloc.allocate(bs);
		T dflt;
		for( typename Allocator::pointer p=pp.p; p<pp.p+bs; ++p )
			alloc.construct(p,dflt);
		return pp;
	}
	static inline void delete_pool(typename Allocator::pointer begin, order_t k, Allocator& alloc)
	{
		typename Allocator::pointer end = begin+buffer_space(k,logc<Navigator::node::order>(k));
		for( typename Allocator::pointer p=begin; p<end; ++p )
			alloc.destroy(p);
		alloc.deallocate(begin, end-begin);
	}

	static typename Allocator::size_type buffer_space(order_t k, height_t height)
	{
		if( height > 1 )
		{
			height_t top_height = Splitter::split(height);
			height_t bottom_height = height-top_height;
			order_t bottom_tree_k = pow_of_order<Navigator::node::order>(bottom_height);
			order_t top_tree_k = (k+bottom_tree_k-1)/bottom_tree_k;
			size_t mid_buffer_size = Splitter::out_buffer_size(bottom_tree_k);
			typename Allocator::size_type size = size_of(top_tree_k, top_height);
			while( k > 0 )
				for( basic_order_t n=0; k>0 && n!=Navigator::node::order; ++n, k=static_cast<order_t>(k-bottom_tree_k) )
					size += mid_buffer_size + size_of(k<bottom_tree_k? k : bottom_tree_k, bottom_height);
			return size;
		}
		else
			return 0;
	}
};

} // namespace funnel


#ifdef __GNUC__
#undef inline
#endif // __GNUC__
#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __LAYOUT_BASE_H_INCLUDED_
