/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __LAYOUT_MVEB_H_INCLUDED_
#define __LAYOUT_MVEB_H_INCLUDED_

#include <iterator>

#include "funnel_base.h"
#include "funnel_structs.h"
#include "layout_base.h"

#include "stack_allocator.h"
#include "pointer_navigator.h"

#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__
#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER


namespace funnel
{

template<basic_order_t order, class Splitter, class T, class Node, height_t max_height>
class mveb_index_base
{
public:
	inline size_t t(height_t h) const
		{ return table[h].t; }
	inline order_t n(height_t h) const
		{ return table[h].n; }
	inline size_t b(height_t h) const
		{ return table[h].b; }
protected:
	explicit inline mveb_index_base(height_t tree_height)
	{
		if(!( tree_height < max_height ))
			throw std::invalid_argument("tree_height of mveb_index out of bounds");
		size_t St[max_height]; // Size of a merger tree, excluding the root's output buffer
		size_t Sb[max_height]; // Size of buffer
		InitS(St,tree_height);
		InitTBD(St, Sb, 0, tree_height);
		table[0].b = St[tree_height];
		for( height_t i=0; i<tree_height; i++ )
		{
			table[i].t += Sb[i];
			table[i].b += Sb[i];
		}
	}
	void InitS(size_t *St, height_t tree_height)
	{
		St[0] = static_cast<size_t>(-1);
		St[1] = sizeof(Node);
		for( height_t height=2; height<tree_height; height++ )
		{
			height_t top_height = Splitter::split(height);
			height_t bottom_height = height-top_height;
			St[height] = St[top_height]+(pow_of_order<order>(top_height))*(Splitter::out_buffer_size(pow_of_order<order>(bottom_height))*sizeof(T)+St[bottom_height]);
		}
	}
	void InitTBD(const size_t *St, size_t *Sb, height_t level1, height_t level2)
	{
		height_t sp = Splitter::split(level2-level1);
		if( level2-level1 > 2 )
		{
			InitTBD(St, Sb, level1, level1+sp);
			InitTBD(St, Sb, level1+sp, level2);
		}
		table[level1+sp].t = St[sp];
		table[level1+sp].n = pow_of_order<order>(sp);
		table[level1+sp].d = level1;
		table[level1+sp].b = St[level2-level1-sp];
		Sb[level1+sp] = Splitter::out_buffer_size(pow_of_order<order>(level2-level1-sp))*sizeof(T);
	}
	struct TE
	{
		size_t t;	// Size of a merger toptree
		order_t n;	// Number of mergers in toptree
		height_t d;	// Depth of root of toptree
		size_t b;   // Size of bottomtree
	} table[max_height];
};

template<basic_order_t order, class Splitter, class T, class Node, height_t max_height>
class mveb_index : public mveb_index_base<order,Splitter,T,Node,max_height>
{
	using mveb_index_base<order,Splitter,T,Node,max_height>::table;
public:
	explicit inline mveb_index(height_t tree_height) : mveb_index_base<order,Splitter,T,Node,max_height>(tree_height) { }
	inline Node *child(Node *root_path[], order_t bi, height_t level) const
		{ assert( level < max_height ); return reinterpret_cast<Node*>(reinterpret_cast<char*>(root_path[table[level].d])-table[level].t-(bi%table[level].n)*table[level].b); }
};

template<class Splitter, class T, class Node, height_t max_height>
class mveb_index<2,Splitter,T,Node,max_height> : public mveb_index_base<2,Splitter,T,Node,max_height>
{
	using mveb_index_base<2,Splitter,T,Node,max_height>::table;
public:
	explicit inline mveb_index(height_t tree_height) : mveb_index_base<2,Splitter,T,Node,max_height>(tree_height) { }
	inline Node *child(Node *root_path[], order_t bi, height_t level) const
		{ assert( level < max_height ); return reinterpret_cast<Node*>(reinterpret_cast<char*>(root_path[table[level].d])-table[level].t-(bi&(table[level].n-1))*table[level].b); }
};

	
template<class Node, class Splitter>
class mixed_veb_navigator;

template<class Navigator, class Splitter, class T, class Allocator=std::allocator<T> >
class mixed_veb_layout : public mixed_layout_base<Navigator,Splitter,T,Allocator>
{
	using mixed_layout_base<Navigator,Splitter,T,Allocator>::compute_buffer_sizes;
	using mixed_layout_base<Navigator,Splitter,T,Allocator>::new_buffer;
	using mixed_layout_base<Navigator,Splitter,T,Allocator>::delete_buffer;
public:
	typedef typename Navigator::node node;

	static node *do_layout(order_t k, Allocator& alloc)
	{
		Navigator::REQUIRE_ALLOCATOR(alloc);
		MIXED_VEB_LAYOUT_REQUIRES_NAVIGATOR(Navigator(NULL,k));
		typename Allocator::template rebind<T>::other buffer_alloc(alloc);
		typename Allocator::template rebind<node>::other node_alloc(alloc);
		return do_layout(k, logc<Navigator::node::order>(k), true, node_alloc, buffer_alloc);
	}

	static void destroy(node *root, order_t k, Allocator& alloc)
	{
		typename Allocator::template rebind<T>::other buffer_alloc(alloc);
		typename Allocator::template rebind<node>::other node_alloc(alloc);
		destroy(root, k, logc<Navigator::node::order>(k), node_alloc, buffer_alloc);
	}

	static typename Allocator::size_type size_of(order_t k)
		{ return layout_base<Navigator,Splitter,T,Allocator>::size_of(k, logc<Navigator::node::order>(k)); }

protected:
	template<class NAlloc, class BAlloc>
	static typename Navigator::node *do_layout(order_t k, height_t height, bool construct_leaves, NAlloc& nalloc, BAlloc& balloc)
	{
		if( height > 1 )
		{
			height_t top_height = Splitter::split(height);
			height_t bottom_height = height-top_height;
			order_t bottom_tree_k = pow_of_order<node::order>(bottom_height);
			order_t top_tree_k = (k+bottom_tree_k-1)/bottom_tree_k;
			size_t mid_buffer_size = Splitter::out_buffer_size(bottom_tree_k);

			node *top_root = do_layout(top_tree_k, top_height, false, nalloc, balloc);
			Navigator nav(top_root, top_tree_k);
			for( typename Navigator::level_iterator leaf = nav.begin_level(top_height-1); k>0; ++leaf )
			{
				node_info<node::order,typename BAlloc::pointer,typename NAlloc::pointer> info;
				for( info.n=0; k>0 && info.n!=node::order; ++info.n, k=static_cast<order_t>(k-bottom_tree_k) )
				{
					typename BAlloc::pointer b = new_buffer(mid_buffer_size, balloc);
                    info.buffers[info.n] = buffer<typename BAlloc::pointer>(b,b+mid_buffer_size);
					info.children[info.n] = do_layout(k<bottom_tree_k? k : bottom_tree_k, bottom_height, construct_leaves, nalloc, balloc);
				}
				node n(info);
				nalloc.construct(*leaf, n);
			}
/*			for( typename Navigator::level_iterator leaf=nav.begin_level(top_height-1); k; ++leaf )
			{
				node_info<node::order,typename BAlloc::pointer,typename NAlloc::pointer> info;
				for( info.n=0; k>=bottom_tree_k && info.n!=node::order; ++info.n, k=static_cast<order_t>(k-bottom_tree_k) )
				{
					typename BAlloc::pointer b = new_buffer(mid_buffer_size, balloc);
					info.buffers[info.n] = buffer<typename BAlloc::pointer>(b,b+mid_buffer_size);
					info.children[info.n] = do_layout(bottom_tree_k, bottom_height, construct_leaves, nalloc, balloc);
				}
				if( k && k < bottom_tree_k && info.n!=node::order )
				{
					typename BAlloc::pointer b = new_buffer(mid_buffer_size, balloc);
					info.buffers[info.n] = buffer<typename BAlloc::pointer>(b,b+mid_buffer_size);
					info.children[info.n] = do_layout(k, bottom_height, construct_leaves, nalloc, balloc);
					k = 0;
				}
				node n(info);
				nalloc.construct(*leaf, n);
			}
*/
			return top_root;
		}
		else
		{
			assert( k <= node::order );
			node *p = nalloc.allocate(1);
			if( construct_leaves )
			{
				node n(static_cast<basic_order_t>(k));
				nalloc.construct(p, n);
			}
			return p;
		}
	}
	template<class NAlloc, class BAlloc>
	static void destroy(typename Navigator::node *root, order_t k, height_t height, NAlloc& nalloc, BAlloc& balloc)
	{
		if( height > 1 )
		{
			height_t top_height = Splitter::split(height);
			height_t bottom_height = height-top_height;
			order_t bottom_tree_k = pow_of_order<node::order>(bottom_height);
			order_t top_tree_k = (k+bottom_tree_k-1)/bottom_tree_k;
			size_t mid_buffer_size = Splitter::out_buffer_size(bottom_tree_k);

			Navigator nav(root, k);
			typename Navigator::level_iterator leaf = nav.end_level(top_height-1);
			basic_order_t n = static_cast<basic_order_t>((top_tree_k+node::order-1)%node::order);
			order_t dk = ((k+bottom_tree_k-1) % bottom_tree_k)+1;
			while( k > 0 )
			{
				--leaf;
				for( ; k>0 && n>=0; --n, k=static_cast<order_t>(k-dk) )
				{
					destroy(leaf.child(n), dk, bottom_height, nalloc, balloc);
					mixed_layout_base<Navigator,Splitter,T,Allocator>::delete_buffer(leaf.input_begin(n), leaf.input_begin(n)+mid_buffer_size, balloc);
					dk = bottom_tree_k;
				}
				n = node::order-1;
				nalloc.destroy(*leaf);
			}
			destroy(root, top_tree_k, top_height, nalloc, balloc);
		}
		else
		{
			nalloc.destroy(root);
			nalloc.deallocate(root, 1);
		}
	}

	template<class Node>
	static inline void MIXED_VEB_LAYOUT_REQUIRES_NAVIGATOR(const mixed_veb_navigator<Node,Splitter>&) {}
	template<class Node>
	static inline void MIXED_VEB_LAYOUT_REQUIRES_NAVIGATOR(const pointer_navigator<Node,Splitter>&) {}
	template<class Node>
	static inline void MIXED_VEB_LAYOUT_REQUIRES_NAVIGATOR(const pointer_flip_navigator<Node,Splitter>&) {}
};


template<class Node, class Splitter>
class mveb_level_iterator;

template<class Node, class Splitter>
class mixed_veb_navigator
{
	friend class mveb_level_iterator<Node,Splitter>;
protected:
	inline basic_order_t goto_parent()
	{
		basic_order_t ch = static_cast<basic_order_t>(bfsi%Node::order);
		bfsi = static_cast<order_t>(bfsi/Node::order);
		assert( !is_root() ); 
		current = *--path;
		if( path > root_path )
			boutput = path[-1]->input(static_cast<basic_order_t>(bfsi%Node::order));
		return ch;
	}
	inline void goto_child(basic_order_t i)
	{
		boutput = current->input(i);
		bfsi = static_cast<order_t>(bfsi*Node::order+i);
		height_t level = static_cast<height_t>(path-root_path)+1;
		current = *++path = mvebi.child(root_path,bfsi,level);
		assert( path[-1]->child(i) == current );
	}

public:
	template<class Allocator>
	static inline void REQUIRE_ALLOCATOR(Allocator& al)
		{ MIXED_VEB_NAVIGATOR_REQUIRES_STACK_ALLOCATOR(al); }

	typedef Node node;
	typedef basic_order_t token;
	typedef Node* bookmark;
	typedef mveb_level_iterator<Node,Splitter> level_iterator;
	typedef buffer<typename Node::pointer> buffer;
	enum { order = Node::order };

	inline bool is_root() const
		{ return path == root_path; }

	inline bookmark mark()
		{ return current; }
	inline bool operator==(bookmark m) const
		{ return current == m; }
	inline bool operator!=(bookmark m) const
		{ return current != m; }

	inline buffer *input(basic_order_t i)
		{ return current->input(i); }
	inline buffer *output()
		{ return boutput; }

	inline mixed_veb_navigator(Node *root, order_t k) : mvebi(logc<Node::order>(k)), current(root), path(root_path), bfsi(1), boutput(NULL), k(k)
	{
		*path = root;
		height_t height = logc<Node::order>(k);
		if( pow_of_order<Node::order>(height) != k )
			throw std::invalid_argument("mixed_veb_navigator only supports balanced trees");
		leaf = root_path+height-1;
		order_t pi = 1;
		for( height_t l=1; l<height; ++l )
			pi *= order;
		left_bfsi = pi;
	}

	inline bool is_leaf() const
		{ return path == leaf; }

	inline token child(basic_order_t i) const
		{ return i; }
	inline token parent() const
		{ return order; }
	inline mixed_veb_navigator& operator+=(token t)
	{
		if( t == order ) // goto parent
		{
			basic_order_t ch = goto_parent();
			flip(ch);
		}
		else // goto child
		{
			assert( !is_leaf() ); 
			flip(t);
			goto_child(t);
		}
		return *this;
	}

	inline mixed_veb_navigator& next_dfs()
	{
		if( !is_root() )
		{
			basic_order_t ch = goto_parent();
			flip(ch);
			if( ch==order-1 )
				return *this;
			flip(ch+1);
			goto_child(ch+1);
		}
		while( !is_leaf() )
		{
			flip(0);
			goto_child(0);
		}
		return *this;
	}

	inline order_t leaf_index()
		{ assert( is_leaf() ); return bfsi-left_bfsi; }

	inline level_iterator begin_level(height_t level)
		{ return level_iterator(*root_path,k,level,false); }
	inline level_iterator end_level(height_t level)
		{ return level_iterator(*root_path,k,level,true); }

	template<class Functor>
	inline Functor enum_buffers(Functor f)
		{ return enum_buffers(f, root_path[0], 0, static_cast<height_t>(leaf-root_path+1), NULL, 0, &mvebi); }

protected:
	inline void flip(basic_order_t i)
	{
		goto_child(i);
		typename Node::pointer b = reinterpret_cast<typename Node::pointer>(current+1);
		goto_parent();
		*input(i) = buffer(b, input(i)->begin());
	}

	template<class Functor>
	static Functor enum_buffers(Functor f, Node *base, height_t h1, height_t h2, Node *bottom_base, size_t outbufs, mveb_index<Node::order,Splitter,typename std::iterator_traits<typename Node::pointer>::value_type,Node,MAX_MERGE_TREE_HEIGHT> *tables)
	{
		if( h2-h1 > 1 )
		{
			height_t top_height = Splitter::split(h2-h1);
			height_t h = h1+top_height;
			order_t top_tree_k = tables->n(h);
			order_t bottom_tree_k = pow_of_order<order>(h2-h);
			Node *bb = reinterpret_cast<Node*>(reinterpret_cast<char*>(base)-tables->t(h));
			f = enum_buffers(f, base, h1, h, bb, Splitter::out_buffer_size(bottom_tree_k), tables);
			if( bottom_base )
				for( base = bb; top_tree_k; base = reinterpret_cast<Node*>(reinterpret_cast<char*>(base)-tables->b(h)), --top_tree_k, bottom_base = reinterpret_cast<Node*>(reinterpret_cast<char*>(bottom_base)-bottom_tree_k*tables->b(h2)) )
					f = enum_buffers(f, base, h, h2, bottom_base, outbufs, tables);
			else
				for( base = bb; top_tree_k; base = reinterpret_cast<Node*>(reinterpret_cast<char*>(base)-tables->b(h)), --top_tree_k )
					f = enum_buffers(f, base, h, h2, NULL, 0, tables);
		}
		else if( bottom_base )
			for( basic_order_t i=0; i!=order; ++i, bottom_base = reinterpret_cast<Node*>(reinterpret_cast<char*>(bottom_base)-tables->b(h2)) )
			{
				typename buffer::pointer b = reinterpret_cast<typename buffer::pointer>(bottom_base+1);
				typename buffer::pointer e = b+outbufs;
				assert( b == base->input_begin(i) );
				assert( e == base->input_end(i) );
				f(base->input(i), b,e);
			}
		return f;
	}

private:
	mveb_index<Node::order,Splitter,typename std::iterator_traits<typename Node::pointer>::value_type,Node,MAX_MERGE_TREE_HEIGHT> mvebi;
	Node *current;
	Node **path;
	order_t bfsi;
	buffer *boutput;
	order_t left_bfsi;
	Node **leaf;
	order_t k;
	Node *root_path[MAX_MERGE_TREE_HEIGHT];

	template<class T>
	static inline void MIXED_VEB_NAVIGATOR_REQUIRES_STACK_ALLOCATOR(stack_allocator<T>&) { }
};

	template<class Node, class Splitter>
	class mveb_level_iterator
	{
		friend class mixed_veb_navigator<Node,Splitter>;
	public:
		inline mveb_level_iterator& operator++()
		{
			height_t i;
			for( i=level; !nav.is_root() && nav.bfsi%Node::order == Node::order-1; --i )
				nav.goto_parent();
			if( nav.is_root() )
				return *this;
			basic_order_t ch = static_cast<basic_order_t>(nav.bfsi%Node::order)+1;
			nav.goto_parent();
			nav.goto_child(ch);
			for( ; i<level; ++i )
				nav.goto_child(0);
			return *this;
		}
		inline mveb_level_iterator& operator--()
		{
			height_t i;
			if( nav.is_root() )
			{
				for( i=0; i<level; ++i )
					nav.goto_child(Node::order-1);
				return *this;
			}
			for( i=level; !nav.is_root() && nav.bfsi%Node::order == 0; --i )
				nav.goto_parent();
			if( nav.is_root() )
				return *this;
			basic_order_t ch = static_cast<basic_order_t>(nav.bfsi%Node::order)-1;
			nav.goto_parent();
			nav.goto_child(ch);
			for( ; i<level; ++i )
				nav.goto_child(Node::order-1);
			return *this;
		}
		inline bool operator==(const mveb_level_iterator& i)
			{ return nav == i.nav.mark(); }
		inline bool operator!=(const mveb_level_iterator& i)
			{ return nav != i.nav.mark(); }
		inline Node* operator*()
			{ return nav.root_path[level]; }
		inline Node* child(basic_order_t i)
		{
			nav.goto_child(i);
			Node *ret = nav.current;
			nav.goto_parent();
			return ret;
		}
		inline typename Node::pointer input_begin(basic_order_t i)
		{
			nav.goto_child(i);
			typename Node::pointer ret = reinterpret_cast<typename Node::pointer>(nav.current+1);
			nav.goto_parent();
			return ret;
		}
	protected:
		inline mveb_level_iterator(Node* root, order_t k, height_t level, bool end) : level(level), nav(root,k)
		{
			if( !end )
				for( height_t i=0; i<level; ++i )
					nav.goto_child(0);
		}

		height_t level;
		mixed_veb_navigator<Node,Splitter> nav;
	};

} // namespace funnel


#ifdef __GNUC__
#undef inline
#endif // __GNUC__
#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __LAYOUT_MVEB_H_INCLUDED_
