/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __LAYOUT_PBF_H_INCLUDED_
#define __LAYOUT_PBF_H_INCLUDED_

#include "funnel_base.h"
#include "funnel_structs.h"
#include "layout_base.h"

#include "stack_allocator.h"
#include "pointer_navigator.h"

#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__
#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER


namespace funnel
{

template<class Node, class Splitter>
class pooled_bf_navigator;

template<class Navigator, class Splitter, class T, class Allocator=std::allocator<T> >
class pooled_bf_layout : public pooled_layout_base<Navigator,Splitter,T,Allocator>
{
	using pooled_layout_base<Navigator,Splitter,T,Allocator>::compute_buffer_sizes;
	using pooled_layout_base<Navigator,Splitter,T,Allocator>::new_pool;
	using pooled_layout_base<Navigator,Splitter,T,Allocator>::delete_pool;
	using pooled_layout_base<Navigator,Splitter,T,Allocator>::new_buffer;
public:
	typedef typename Navigator::node node;

	static node *do_layout(order_t k, Allocator& alloc)
	{
		Navigator::REQUIRE_ALLOCATOR(alloc);
		POOLED_BF_LAYOUT_REQUIRES_NAVIGATOR(Navigator(NULL,k));
		height_t height = logc<Navigator::node::order>(k);
		size_t buffer_sizes[MAX_MERGE_TREE_HEIGHT];
		compute_buffer_sizes(buffer_sizes,buffer_sizes+height);
		typename Allocator::template rebind<T>::other buffer_alloc(alloc);
		typename Allocator::template rebind<node>::other node_alloc(alloc);
		typename pooled_layout_base<Navigator,Splitter,T,Allocator>::pool_pointer pp = new_pool(k,buffer_alloc);
		node *root = do_layout(k, height, buffer_sizes, node_alloc, pp);
		order_t r = pow_of_order<node::order>(height-1);
		Navigator nav(root, k);
		typename Navigator::level_iterator leaf = nav.begin_level(height-1);
		for( ; k>0; ++leaf, --r )
		{
			basic_order_t d;
			for( d=0; k>0 && d!=node::order; ++d, --k );
			node n(d);
			node_alloc.construct(*leaf, n);
		}
		node n(0);
		for( ; r; ++leaf, --r )
			node_alloc.construct(*leaf, n);
		return root;
	}

	static void destroy(node *root, order_t k, Allocator& alloc)
	{
		height_t height = logc<node::order>(k);
		size_t buffer_sizes[MAX_MERGE_TREE_HEIGHT];
		compute_buffer_sizes(buffer_sizes,buffer_sizes+height);
		typename Allocator::template rebind<node>::other node_alloc(alloc);
		order_t r = pow_of_order<node::order>(height-1);
		Navigator nav(root, static_cast<order_t>(r*node::order));
		typename Allocator::template rebind<T>::other::pointer pool = nav.begin_level(0).input_begin(0);
		typename Navigator::level_iterator leaf = nav.begin_level(height-1); --leaf;
		for( order_t i=0; i<r; ++i, --leaf )
			node_alloc.destroy(*leaf);
		destroy(root, height, buffer_sizes, node_alloc);

		typename Allocator::template rebind<T>::other buffer_alloc(alloc);
		delete_pool(pool, k, buffer_alloc);
	}

protected:
	template<class NAlloc, class BAlloc>
	static node *do_layout(order_t k, height_t height, size_t *buffer_sizes, NAlloc& nalloc, BAlloc& balloc)
	{
		if( height > 1 )
		{
			node *root = do_layout(static_cast<order_t>((k+node::order-1)/node::order), height-1, buffer_sizes, nalloc, balloc);
			order_t r = pow_of_order<node::order>(height-2);
			Navigator nav(root, static_cast<order_t>(node::order*r));
			typename Navigator::level_iterator leaf = nav.begin_level(height-2);
			for( ; k>0; ++leaf, --r )
			{
				node_info<node::order,typename BAlloc::pointer,typename NAlloc::pointer> info;
				for( info.n=0; k>0 && info.n!=node::order; ++info.n, k-=node::order )
				{
					typename BAlloc::pointer b = new_buffer(buffer_sizes[height-1],balloc);
                    info.buffers[info.n] = buffer<typename BAlloc::pointer>(b,b+buffer_sizes[height-1]);
					info.children[info.n] = nalloc.allocate(1);
				}
				for( ; info.n!=node::order; ++info.n )
					info.children[info.n] = nalloc.allocate(1);
				node n(info);
				nalloc.construct(*leaf, n);
			}
			for( ; r; ++leaf, --r )
			{
				node_info<node::order,typename BAlloc::pointer,typename NAlloc::pointer> info;
				for( info.n=0; info.n!=node::order; ++info.n )
					info.children[info.n] = nalloc.allocate(1);
				node n(info);
				nalloc.construct(*leaf, n);
			}
			return root;
		}
		else
			return nalloc.allocate(1);
	}
	template<class NAlloc>
	static void destroy(node *root, height_t height, size_t *buffer_sizes, NAlloc& nalloc)
	{
		if( height > 1 )
		{
			order_t k = pow_of_order<node::order>(height-2);
			Navigator nav(root, k);
			typename Navigator::level_iterator leaf = nav.end_level(height-2);
			for( --leaf; k; --k, --leaf )
			{
				for( basic_order_t n=node::order-1; n>=0; --n )
					nalloc.deallocate(leaf.child(n),1);
				nalloc.destroy(*leaf);
			}
			destroy(root, height-1, buffer_sizes, nalloc);
		}
		else
			nalloc.deallocate(root,1);
	}
	template<class Node>
	static inline void POOLED_BF_LAYOUT_REQUIRES_NAVIGATOR(const pooled_bf_navigator<Node,Splitter>&) {}
	template<class Node>
	static inline void POOLED_BF_LAYOUT_REQUIRES_NAVIGATOR(const pointer_navigator<Node,Splitter>&) {}
	template<class Node>
	static inline void POOLED_BF_LAYOUT_REQUIRES_NAVIGATOR(const pointer_flip_navigator<Node,Splitter>&) {}
};


template<class Node, class Splitter>
class pooled_bf_navigator
{
protected:
	class bf_level_iterator
	{
		friend class pooled_bf_navigator<Node,Splitter>;
	public:
		inline bf_level_iterator& operator++()
			{ ++index; return *this; }
		inline bf_level_iterator& operator--()
			{ --index; return *this; }
		inline bool operator==(const bf_level_iterator& i)
			{ return index == i.index; }
		inline bool operator!=(const bf_level_iterator& i)
			{ return index != i.index; }
		inline Node* operator*()
			{ return root-index; }
		inline Node* child(basic_order_t i)
			{ return root-(node::order*(index-1)+i+2); }
		inline typename Node::pointer input_begin(basic_order_t i)
			{ return (root-index)->input_begin(i); }
	protected:
		inline bf_level_iterator(Node* root, height_t level) : root(root)
		{
			bfs_index<order> bfs;
			for( height_t l=0; l<level; ++l )
				bfs.child(0);
			index = bfs;
		}
		Node *root;
		order_t index;
	};

	class bf_token
	{
		typedef typename Node::pointer pointer;
		enum { order = Node::order };
	public:
		bfs_index<order> bfsi;
		funnel::buffer<typename Node::pointer> *passing;
		pointer begin;
		bool operator==(const bf_token& rhs) const
			{ return bfsi == rhs.bfsi; }
	};
public:
	template<class Allocator>
	static inline void REQUIRE_ALLOCATOR(Allocator& al)
		{ POOLED_BF_NAVIGATOR_REQUIRES_STACK_ALLOCATOR(al); }

	enum { order = Node::order };
	typedef Node node;
	typedef buffer<typename Node::pointer> buffer;
	typedef bf_level_iterator level_iterator;
	typedef bfs_index<order> bookmark;
	typedef bf_token token;

	inline pooled_bf_navigator(Node *root, order_t k) : root(root+1)
	{
		bfs_index<order> bfs;
		for( height_t l=1; l<logc<order>(k); ++l )
			bfs.child(0);
		left_bfsi = bfs;
	}

	inline token child(basic_order_t i) const
	{
		token ret;
		ret.bfsi = bfsi;
		ret.bfsi.child(i);
		ret.passing = (root-bfsi)->input(i);
		ret.begin = (root-bfsi)->input_begin(i);
		return ret;
	}
	inline token parent() const
	{
		token ret;
		ret.bfsi = bfsi;
		ret.bfsi.parent();
		basic_order_t ch = bfsi.child_index();
		ret.passing = (root-ret.bfsi)->input(ch);
		ret.begin = (root-ret.bfsi)->input_begin(ch);
		return ret;
	}
	inline pooled_bf_navigator& operator+=(token t)
	{
		assert( (order_t)t.bfsi );
		bfsi = t.bfsi;
		bfs_index<order> parent = bfsi;
		parent.parent();
		boutput = (root-parent)->input(bfsi.child_index());
		*t.passing = buffer(t.begin, t.passing->begin());
		return *this;
	}

	inline bool is_root() const
		{ return (order_t)bfsi == 1; }
	inline bool is_leaf() const
		{ return (order_t)bfsi >= left_bfsi; }

	inline bookmark mark()
		{ return bfsi; }
	inline bool operator==(bookmark m) const
		{ return bfsi == m; }
	inline bool operator!=(bookmark m) const
		{ return bfsi != m; }

	inline buffer *input(basic_order_t i)
		{ return (root-bfsi)->input(i); }
	inline buffer *output()
		{ assert( !is_root() ); return boutput; }
	inline order_t leaf_index()
		{ assert( is_leaf() ); return bfsi-left_bfsi; }

	inline pooled_bf_navigator& next_dfs()
	{
		if( !is_root() )
		{
			basic_order_t ch = bfsi.child_index();
			bfsi.parent();
			flip(ch);
			if( ch==order-1 /*|| current->child(ch+1)==NULL*/ )
			{
				if( !is_root() )
				{
					bfs_index<order> p = bfsi;
					p.parent();
					boutput = (root-p)->input(bfsi.child_index());
				}
				return *this;
			}
			flip(ch+1);
			boutput = (root-bfsi)->input(ch+1);
			bfsi.child(ch+1);
		}
		while( !is_leaf() )
		{
			flip(0);
			boutput = (root-bfsi)->input(0);
			bfsi.child(0);
		}
		return *this;
	}

	inline bf_level_iterator begin_level(height_t level)
		{ return bf_level_iterator(root,level); }
	inline bf_level_iterator end_level(height_t level)
		{ return bf_level_iterator(root,level+1); }

	template<class Functor>
	Functor enum_buffers(Functor f)
		{ return enum_buffers(root-left_bfsi,root,f); }

protected:
	template<class Functor>
	static Functor enum_buffers(Node* b, Node* e, Functor f)
	{
		for( ; b!=e; ++b )
			for( basic_order_t i=0; i!=order; ++i )
				f(b->input(i), b->input_begin(i), b->input_end(i));
		return f;
	}
	inline void flip(basic_order_t i)
	{
		*(root-bfsi)->input(i) = buffer((root-bfsi)->input_begin(i), (root-bfsi)->input(i)->begin());
	}

	Node *root;
	bfs_index<order> bfsi;
	buffer *boutput;
	order_t left_bfsi;

private:
	template<class T>
	static inline void POOLED_BF_NAVIGATOR_REQUIRES_STACK_ALLOCATOR(stack_allocator<T>&) { }
};

} // namespace funnel


#ifdef __GNUC__
#undef inline
#endif // __GNUC__
#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __LAYOUT_PBF_H_INCLUDED_
