/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __LAYOUT_PDF_H_INCLUDED_
#define __LAYOUT_PDF_H_INCLUDED_

#include "funnel_base.h"
#include "funnel_structs.h"
#include "layout_base.h"

#include "stack_allocator.h"
#include "pointer_navigator.h"

#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__
#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER


namespace funnel
{

template<class Node, class Splitter>
class pooled_df_navigator;

template<class Navigator, class Splitter, class T, class Allocator=std::allocator<T> >
class pooled_df_layout : public pooled_layout_base<Navigator,Splitter,T,Allocator>
{
	using pooled_layout_base<Navigator,Splitter,T,Allocator>::compute_buffer_sizes;
	using pooled_layout_base<Navigator,Splitter,T,Allocator>::new_pool;
	using pooled_layout_base<Navigator,Splitter,T,Allocator>::delete_pool;
	using pooled_layout_base<Navigator,Splitter,T,Allocator>::new_buffer;
public:
	typedef typename Navigator::node node;

	static node *do_layout(order_t k, Allocator& alloc)
	{
		Navigator::REQUIRE_ALLOCATOR(alloc);
		POOLED_DF_LAYOUT_REQUIRES_NAVIGATOR(Navigator(NULL,k));
		height_t height = logc<Navigator::node::order>(k);
		size_t buffer_sizes[MAX_MERGE_TREE_HEIGHT];
		compute_buffer_sizes(buffer_sizes,buffer_sizes+height);
		typename Allocator::template rebind<T>::other buffer_alloc(alloc);
		typename Allocator::template rebind<node>::other node_alloc(alloc);
		typename pooled_layout_base<Navigator,Splitter,T,Allocator>::pool_pointer pp = new_pool(k,buffer_alloc);
		return do_layout(k, height, buffer_sizes, node_alloc, pp);
	}

	static void destroy(node *root, order_t k, Allocator& alloc)
	{
		height_t height = logc<Navigator::node::order>(k);
		size_t buffer_sizes[MAX_MERGE_TREE_HEIGHT];
		compute_buffer_sizes(buffer_sizes,buffer_sizes+height);
		Navigator nav(root, pow_of_order<node::order>(height));
		typename Allocator::template rebind<T>::other::pointer pool = nav.begin_level(0).input_begin(0);
		typename Allocator::template rebind<node>::other node_alloc(alloc);
		destroy(root, pow_of_order<Navigator::node::order>(height), buffer_sizes, node_alloc);
		typename Allocator::template rebind<T>::other buffer_alloc(alloc);
		delete_pool(pool, k, buffer_alloc);
	}

protected:
	template<class NAlloc, class BAlloc>
	static node *do_layout(order_t k, height_t height, size_t *buffer_sizes, NAlloc& nalloc, BAlloc& balloc)
	{
		if( height > 1 )
		{
			node *root = nalloc.allocate(1);
			order_t sk = pow_of_order<node::order>(height-1);
			node_info<node::order,typename BAlloc::pointer,typename NAlloc::pointer> info;
			typename BAlloc::pointer b = new_buffer(0,balloc);
			for( info.n=0; k>0 && info.n!=node::order; ++info.n, k=static_cast<order_t>(k-sk) )
			{
				b = new_buffer(buffer_sizes[1],balloc);
				info.buffers[info.n] = buffer<typename BAlloc::pointer>(b,b+buffer_sizes[1]);
				info.children[info.n] = do_layout(k<sk? k : sk, height-1, buffer_sizes+1, nalloc, balloc);
			}
			for( ; info.n!=node::order; ++info.n )
			{
				info.buffers[info.n] = buffer<typename BAlloc::pointer>(b,b);
				info.children[info.n] = do_layout(0, height-1, buffer_sizes+1, nalloc, balloc);
			}
			node n(info);
			nalloc.construct(root, n);
			return root;
		}
		else
		{
			assert( k <= node::order );
			node *p = nalloc.allocate(1);
			node n(static_cast<basic_order_t>(k));
			nalloc.construct(p, n);
			return p;
		}
	}
	template<class NAlloc>
	static void destroy(node *root, order_t k, size_t *buffer_sizes, NAlloc& nalloc)
	{
		if( k > node::order )
		{
			Navigator nav(root, k);
			typename Navigator::level_iterator leaf = nav.begin_level(0);
			for( basic_order_t n=node::order-1; n>=0; --n )
				destroy(leaf.child(n), static_cast<order_t>(k/node::order), buffer_sizes+1, nalloc);
		}
		nalloc.destroy(root);
		nalloc.deallocate(root,1);
	}
	template<class Node>
	static inline void POOLED_DF_LAYOUT_REQUIRES_NAVIGATOR(const pooled_df_navigator<Node,Splitter>&) {}
	template<class Node>
	static inline void POOLED_DF_LAYOUT_REQUIRES_NAVIGATOR(const pointer_navigator<Node,Splitter>&) {}
	template<class Node>
	static inline void POOLED_DF_LAYOUT_REQUIRES_NAVIGATOR(const pointer_flip_navigator<Node,Splitter>&) {}
};


template<class Node, class Splitter>
class pooled_df_navigator
{
protected:
	class df_level_iterator
	{
		friend class pooled_df_navigator<Node,Splitter>;
	public:
		inline df_level_iterator& operator++() // untested
		{
			height_t d = 0;
			for( order_t bfs=bfsi; (bfs+order-2) % order == order-1; bfs=(bfs-2)/order+1, d++ );
			dfsi += dc + d;
			++bfsi;
			return *this;
		}
		inline df_level_iterator& operator--() // untested
		{
			height_t d = 0;
			for( order_t bfs=bfsi; (bfs+order-2) % order == 0; bfs=(bfs-2)/order+1, d++ );
			dfsi -= dc + d;
			--bfsi;
			return *this;
		}
		inline bool operator==(const df_level_iterator& i)
			{ return dfsi == i.dfsi; }
		inline bool operator!=(const df_level_iterator& i)
			{ return dfsi != i.dfsi; }
		inline Node* operator*()
			{ return root-dfsi; }
		inline Node* child(basic_order_t i)
			{ return root-dfsi-1-i*dc; }
		inline typename Node::pointer input_begin(basic_order_t i)
			{ return (root-dfsi)->input_begin(i); }
	protected:
		inline df_level_iterator(Node* root, height_t level, order_t root_dc) : root(root), dfsi(level+1), dc(root_dc)
		{
			bfs_index<order> bfs;
			for( height_t l=0; l<level; ++l )
			{
				bfs.child(0);
				dc /= order;
			}
			bfsi = bfs;
		}
		Node *root;
		order_t bfsi;
		order_t dfsi;
		order_t dc;
	};

/*	class df_token
	{
		typedef typename Node::pointer pointer;
		enum { order = Node::order };
	public:
		bfs_index<order> bfsi;
		order_t dfsi;
		order_t dc;
		buffer<pointer> *passing, *out;
		pointer begin;

		bool operator==(const df_token& rhs) const
			{ return bfsi == rhs.bfsi; }
	};*/
public:
	template<class Allocator>
	static inline void REQUIRE_ALLOCATOR(Allocator& al)
		{ POOLED_DF_NAVIGATOR_REQUIRES_STACK_ALLOCATOR(al); }

	enum { order = Node::order };
	typedef Node node;
	typedef buffer<typename Node::pointer> buffer;
	typedef df_level_iterator level_iterator;
	typedef bfs_index<order> bookmark;
//	typedef df_token token;
	typedef basic_order_t token;

	inline pooled_df_navigator(Node *root, order_t k) : root(root+1), dfsi(1)
	{
		height_t h = logc<order>(k);
		root_dc = dc = (pow_of_order<order>(h-1)-1)/(order-1);
		bfs_index<order> bfs;
		for( height_t l=1; l<h; ++l )
			bfs.child(0);
		left_bfsi = bfs;
	}

	inline token child(basic_order_t i) const
	{
		return i;
/*		token ret;
		ret.bfsi = bfsi;
		ret.bfsi.child(i);
		ret.dfsi = dfsi+1+i*dc;
		ret.dc = static_cast<order_t>(dc/order);
		assert( (root-dfsi)->child(i) == root-ret.dfsi );
		assert( (root-dfsi)->child(0) == dc*i + (root-dfsi)->child(i) );
		ret.out = ret.passing = (root-dfsi)->input(i);
		ret.begin = (root-dfsi)->input_begin(i);
		return ret;
*/	}
	inline token parent() const
	{
		return order;
/*		token ret;
		ret.bfsi = bfsi;
		ret.bfsi.parent();
		ret.dc = static_cast<order_t>(dc*order+1);
		basic_order_t ch = bfsi.child_index();
		ret.dfsi = dfsi-1-ch*ret.dc;
		if( bfsi != 1 )
		{
			assert( (root-ret.dfsi)->child(ch) == root-dfsi );
			ret.passing = (root-ret.dfsi)->input(ch);
			ret.begin = (root-ret.dfsi)->input_begin(ch);
			if( ret.bfsi != 1 )
			{
				ch = ret.bfsi.child_index();
				order_t p = static_cast<order_t>(ret.dfsi-1-ch*(ret.dc*order+1));
				assert( (root-p)->child(ch) == root-ret.dfsi );
				ret.out = (root-p)->input(ch);
			}
		}
		return ret;
*/	}
	inline pooled_df_navigator& operator+=(token t)
	{
		if( t == order )
		{
			assert( (order_t)bfsi > 1 );
			basic_order_t ch = bfsi.child_index();
			bfsi.parent();
			dc = static_cast<order_t>(dc*order+1);
			assert( (root-(dfsi-1-ch*dc))->child(ch) == root-dfsi );
			dfsi = dfsi-1-ch*dc;
			flip(ch);
			if( bfsi != 1 )
			{
				ch = bfsi.child_index();
				order_t p = static_cast<order_t>(dfsi-1-ch*(dc*order+1));
				assert( (root-p)->child(ch) == root-dfsi );
				boutput = (root-p)->input(ch);
			}
		}
		else
		{
			assert( (root-dfsi)->child(t) == root-(dfsi+1+t*dc) );
			assert( (root-dfsi)->child(0) == t*dc + (root-dfsi)->child(t) );
			boutput = (root-dfsi)->input(t);
			flip(t);
			bfsi.child(t);
			dfsi = dfsi+1+t*dc;
			dc = static_cast<order_t>(dc/order);
		}
/*		dc = t.dc;
		dfsi = t.dfsi;
		bfsi = t.bfsi;
		boutput = t.out;
		*t.passing = buffer(t.begin, t.passing->begin());
*/		return *this;
	}

	inline bool is_root() const
		{ return (order_t)bfsi == 1; }
	inline bool is_leaf() const
		{ return (order_t)bfsi >= left_bfsi; }

	inline bookmark mark()
		{ return bfsi; }
	inline bool operator==(bookmark m) const
		{ return bfsi == m; }
	inline bool operator!=(bookmark m) const
		{ return bfsi != m; }

	inline buffer *input(basic_order_t i)
		{ return (root-dfsi)->input(i); }
	inline buffer *output()
		{ assert( !is_root() ); return boutput; }
	inline order_t leaf_index()
		{ assert( is_leaf() ); return bfsi-left_bfsi; }

	inline pooled_df_navigator& next_dfs()
	{
		if( !is_root() )
		{
			dc = static_cast<order_t>(dc*order+1);
			basic_order_t ch = bfsi.child_index();
			dfsi = dfsi-1-ch*dc;
			bfsi.parent();
			flip(ch);
			if( ch==order-1 /*|| current->child(ch+1)==NULL*/ )
			{
				if( !is_root() )
				{
					basic_order_t ch = bfsi.child_index();
					order_t p = static_cast<order_t>(dfsi-1-ch*(dc*order+1));
					boutput = (root-p)->input(ch);
				}
				return *this;
			}
			flip(ch+1);
			boutput = (root-dfsi)->input(ch+1);
			bfsi.child(ch+1);
			dfsi = dfsi+1+(ch+1)*dc;
			dc = static_cast<order_t>(dc/order);
		}
		while( !is_leaf() )
		{
			flip(0);
			boutput = (root-dfsi)->input(0);
			bfsi.child(0);
			dfsi = dfsi+1;
			dc = static_cast<order_t>(dc/order);
		}
		return *this;
	}

	inline df_level_iterator begin_level(height_t level)
		{ return df_level_iterator(root,level,root_dc); }
	inline df_level_iterator end_level(height_t level)
		{ assert( false ); return df_level_iterator(root,level+1,root_dc); }

	template<class Functor>
	inline Functor enum_buffers(Functor f)
		{ return enum_buffers(f, root, 1, root_dc); }

protected:
	template<class Functor>
	static Functor enum_buffers(Functor f, Node* root, order_t dfsi, order_t dc)
	{
		for( basic_order_t i=0; i!=order; ++i )
			if( dc )
			{
				f((root-dfsi)->input(i), (root-dfsi)->input_begin(i), (root-dfsi)->input_end(i));
				assert( (root-dfsi)->child(i) == root-(dfsi+1+i*dc) );
				f = enum_buffers(f, root, static_cast<order_t>(dfsi+1+i*dc), static_cast<order_t>(dc/order));
			}
		return f;
	}
	inline void flip(basic_order_t i)
	{
		*(root-dfsi)->input(i) = buffer((root-dfsi)->input_begin(i), (root-dfsi)->input(i)->begin());
	}

	Node *root;
	bfs_index<order> bfsi;
	buffer *boutput;
	order_t left_bfsi;
	order_t dfsi;
	order_t dc, root_dc;

private:
	template<class T>
	static inline void POOLED_DF_NAVIGATOR_REQUIRES_STACK_ALLOCATOR(stack_allocator<T>&) { }
};

} // namespace funnel


#ifdef __GNUC__
#undef inline
#endif // __GNUC__
#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __LAYOUT_PDF_H_INCLUDED_
