/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __LAYOUT_PVEB_H_INCLUDED_
#define __LAYOUT_PVEB_H_INCLUDED_

#include "funnel_base.h"
#include "funnel_structs.h"
#include "layout_base.h"

#include "stack_allocator.h"
#include "pointer_navigator.h"

#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__
#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER


namespace funnel
{

template<basic_order_t order, class Splitter, class Node, height_t max_height>
class veb_index
{
public:
	explicit inline veb_index(height_t tree_height)
	{
		if(!( tree_height < max_height ))
			throw std::invalid_argument("tree_height of veb_index out of bounds");
		order_t St[max_height]; // Size of a merger tree, excluding the root's output buffer
		InitS(St,tree_height);
		InitTBD(St, 0, tree_height);
		table[0].b = St[tree_height];
	}
	inline size_t t(height_t h) const
		{ return static_cast<size_t>(table[h].t); }
	inline order_t n(height_t h) const
		{ return table[h].n; }
	inline size_t b(height_t h) const
		{ return static_cast<size_t>(table[h].b); }
	inline Node *child(Node *root_path[], order_t bi, height_t level) const
		{ assert( level < max_height ); return root_path[table[level].d]-table[level].t-(bi%table[level].n)*table[level].b; }
protected:
	inline void InitS(order_t *St, height_t tree_height)
	{
		St[0] = static_cast<order_t>(-1);
		St[1] = 1;
		for( height_t height=2; height<tree_height; height++ )
		{
			height_t top_height = Splitter::split(height);
			height_t bottom_height = height-top_height;
			St[height] = St[top_height]+pow_of_order<order>(top_height)*St[bottom_height];
		}
	}
	void InitTBD(const order_t *St, height_t level1, height_t level2)
	{
		height_t sp = Splitter::split(level2-level1);
		if( level2-level1 > 2 )
		{
			InitTBD(St, level1, level1+sp);
			InitTBD(St, level1+sp, level2);
		}
		table[level1+sp].t = St[sp];
		table[level1+sp].n = pow_of_order<order>(sp);
		table[level1+sp].d = level1;
		table[level1+sp].b = St[level2-level1-sp];
	}
	struct TE
	{
		order_t t;	// Size of a merger toptree
		order_t n;	// Number of mergers in toptree
		height_t d;	// Depth of root of toptree
		order_t b;   // Size of bottomtree
	} table[max_height];
};
template<class Splitter, class Node, height_t max_height>
class veb_index<2,Splitter,Node,max_height>
{
public:
	explicit inline veb_index(height_t tree_height)
	{
		if(!( tree_height < max_height ))
			throw std::invalid_argument("tree_height of veb_index out of bounds");
		order_t St[max_height]; // Size of a merger tree, excluding the root's output buffer
		InitS(St,tree_height);
		InitTBD(St, 0, tree_height);
		table[0].b = St[tree_height];
	}
	inline size_t t(height_t h) const
		{ return static_cast<size_t>(table[h].t); }
	inline order_t n(height_t h) const
		{ return table[h].t+1; }
	inline size_t b(height_t h) const
		{ return static_cast<size_t>(table[h].b); }
	inline Node *child(Node *root_path[], order_t bi, height_t level) const
		{ assert( level < max_height ); return root_path[table[level].d]-table[level].t-(bi&table[level].t)*table[level].b; }
protected:
	inline void InitS(order_t *St, height_t tree_height)
	{
		St[0] = static_cast<order_t>(-1);
		St[1] = 1;
		for( height_t height=2; height<tree_height; height++ )
		{
			height_t top_height = Splitter::split(height);
			height_t bottom_height = height-top_height;
			St[height] = St[top_height]+pow_of_order<2>(top_height)*St[bottom_height];
		}
	}
	void InitTBD(const order_t *St, height_t level1, height_t level2)
	{
		height_t sp = Splitter::split(level2-level1);
		if( level2-level1 > 2 )
		{
			InitTBD(St, level1, level1+sp);
			InitTBD(St, level1+sp, level2);
		}
		table[level1+sp].t = St[sp];
		table[level1+sp].d = level1;
		table[level1+sp].b = St[level2-level1-sp];
	}
	struct TE
	{
		order_t t;	// Size of a merger toptree / Number of mergers in toptree
		height_t d;	// Depth of root of toptree
		order_t b;   // Size of bottomtree
	} table[max_height];
};

template<class Node, class Splitter>
class pooled_veb_navigator;

template<class Navigator, class Splitter, class T, class Allocator=std::allocator<T> >
class pooled_veb_layout : public pooled_layout_base<Navigator,Splitter,T,Allocator>
{
	using pooled_layout_base<Navigator,Splitter,T,Allocator>::compute_buffer_sizes;
	using pooled_layout_base<Navigator,Splitter,T,Allocator>::new_pool;
	using pooled_layout_base<Navigator,Splitter,T,Allocator>::delete_pool;
	using pooled_layout_base<Navigator,Splitter,T,Allocator>::new_buffer;
public:
	typedef typename Navigator::node node;

	static node *do_layout(order_t k, Allocator& alloc)
	{
		Navigator::REQUIRE_ALLOCATOR(alloc);
		POOLED_VEB_LAYOUT_REQUIRES_NAVIGATOR(Navigator(NULL,k));
		typename Allocator::template rebind<T>::other buffer_alloc(alloc);
		typename Allocator::template rebind<node>::other node_alloc(alloc);
		typename pooled_layout_base<Navigator,Splitter,T,Allocator>::pool_pointer pp = new_pool(k,buffer_alloc);
		return do_layout(k, logc<Navigator::node::order>(k), true, node_alloc, pp);
	}

	static void destroy(node *root, order_t k, Allocator& alloc)
	{
		height_t height = logc<Navigator::node::order>(k);
		Navigator nav(root, k);
		typename Allocator::template rebind<T>::other::pointer pool = nav.begin_level(0).input_begin(0);
		typename Allocator::template rebind<node>::other node_alloc(alloc);
		destroyh(root, height, node_alloc);
		typename Allocator::template rebind<T>::other buffer_alloc(alloc);
		delete_pool(pool, k, buffer_alloc);
	}

protected:
	template<class NAlloc, class BAlloc>
	static node *do_layout(order_t k, height_t height, bool construct_leaves, NAlloc& nalloc, BAlloc& balloc)
	{
		if( height > 1 )
		{
			height_t top_height = Splitter::split(height);
			height_t bottom_height = height-top_height;
			order_t bottom_tree_k = pow_of_order<node::order>(bottom_height);
			order_t top_tree_k = (k+bottom_tree_k-1)/bottom_tree_k;
			size_t mid_buffer_size = Splitter::out_buffer_size(bottom_tree_k);
			order_t r = pow_of_order<node::order>(top_height-1);

			node *top_root = do_layout(top_tree_k, top_height, false, nalloc, balloc);
			Navigator nav(top_root, static_cast<order_t>(node::order*r));
			typename Navigator::level_iterator leaf = nav.begin_level(top_height-1);
			for( ; k>0; ++leaf, --r )
			{
				node_info<node::order,typename BAlloc::pointer,typename NAlloc::pointer> info;
				for( info.n=0; k>0 && info.n!=node::order; ++info.n, k=static_cast<order_t>(k-bottom_tree_k) )
				{
					typename BAlloc::pointer b = new_buffer(mid_buffer_size,balloc);
                    info.buffers[info.n] = buffer<typename BAlloc::pointer>(b,b+mid_buffer_size);
					info.children[info.n] = do_layout(k<bottom_tree_k? k : bottom_tree_k, bottom_height, construct_leaves, nalloc, balloc);
				}
				for( ; info.n!=node::order; ++info.n )
					info.children[info.n] = do_layout(0, bottom_height, construct_leaves, nalloc, balloc);
				node n(info);
				nalloc.construct(*leaf, n);
			}
			for( ; r; ++leaf, --r )
			{
				node_info<node::order,typename BAlloc::pointer,typename NAlloc::pointer> info;
				for( info.n=0; info.n!=node::order; ++info.n )
					info.children[info.n] = do_layout(0, bottom_height, construct_leaves, nalloc, balloc);
				node n(info);
				nalloc.construct(*leaf, n);
			}

			return top_root;
		}
		else
		{
			assert( k <= node::order );
			node *p = nalloc.allocate(1);
			if( construct_leaves )
			{
				node n(static_cast<basic_order_t>(k));
				nalloc.construct(p, n);
			}
			return p;
		}
	}
	template<class NAlloc>
	static void destroyh(node *root, height_t height, NAlloc& nalloc)
	{
		if( height > 1 )
		{
			height_t top_height = Splitter::split(height);
			height_t bottom_height = height-top_height;
			order_t top_tree_k = pow_of_order<node::order>(top_height-1);

			Navigator nav(root, pow_of_order<node::order>(height));
			typename Navigator::level_iterator leaf = nav.end_level(top_height-1);
			for( --leaf; top_tree_k; --top_tree_k, --leaf )
				for( basic_order_t n=node::order-1; n>=0; --n )
					destroyh(leaf.child(n), bottom_height, nalloc);
			destroyh(root, top_height, nalloc);
		}
		else
		{
			nalloc.destroy(root);
			nalloc.deallocate(root, 1);
		}
	}
	template<class Node>
	static inline void POOLED_VEB_LAYOUT_REQUIRES_NAVIGATOR(const pooled_veb_navigator<Node,Splitter>&) {}
	template<class Node>
	static inline void POOLED_VEB_LAYOUT_REQUIRES_NAVIGATOR(const pointer_navigator<Node,Splitter>&) {}
	template<class Node>
	static inline void POOLED_VEB_LAYOUT_REQUIRES_NAVIGATOR(const pointer_flip_navigator<Node,Splitter>&) {}
};


template<class Node, class Splitter>
class pveb_level_iterator;


template<class Node, class Splitter>
class pooled_veb_navigator
{
	friend class pveb_level_iterator<Node,Splitter>;
protected:
	inline basic_order_t goto_parent()
	{
		basic_order_t ch = static_cast<basic_order_t>(bfsi%Node::order);
		bfsi = static_cast<order_t>(bfsi/Node::order);
		assert( !is_root() ); 
		current = *--path;
		if( path > root_path )
			boutput = path[-1]->input(static_cast<basic_order_t>(bfsi%Node::order));
		return ch;
	}
	inline void goto_child(basic_order_t i)
	{
		boutput = current->input(i);
		bfsi = static_cast<order_t>(bfsi*Node::order+i);
		height_t level = static_cast<height_t>(path-root_path)+1;
		current = *++path = vebi.child(root_path,bfsi,level);
		assert( path[-1]->child(i) == current );
	}

public:
	template<class Allocator>
	static inline void REQUIRE_ALLOCATOR(Allocator& al)
		{ POOLED_VEB_NAVIGATOR_REQUIRES_STACK_ALLOCATOR(al); }

	enum { order = Node::order };
	typedef Node node;
	typedef buffer<typename Node::pointer> buffer;
	typedef pveb_level_iterator<Node,Splitter> level_iterator;
	typedef Node* bookmark;
	typedef basic_order_t token;

	inline pooled_veb_navigator(Node *root, order_t k) : vebi(logc<Node::order>(k)), current(root), path(root_path), bfsi(1), boutput(NULL), k(k)
	{
		*path = root;
		height_t height = logc<Node::order>(k);
		leaf = root_path+height-1;
		order_t pi = 1;
		for( height_t l=1; l<height; ++l )
			pi *= order;
		left_bfsi = pi;
	}

	inline bool is_root() const
		{ return path == root_path; }

	inline bookmark mark()
		{ return current; }
	inline bool operator==(bookmark m) const
		{ return current == m; }
	inline bool operator!=(bookmark m) const
		{ return current != m; }

	inline buffer *input(basic_order_t i)
		{ return current->input(i); }
	inline buffer *output()
		{ return boutput; }

	inline bool is_leaf() const
		{ return path == leaf; }

	inline token child(basic_order_t i) const
		{ return i; }
	inline token parent() const
		{ return order; }
	inline pooled_veb_navigator& operator+=(token t)
	{
		if( t == order ) // goto parent
		{
			basic_order_t ch = goto_parent();
			flip(ch);
		}
		else // goto child
		{
			assert( !is_leaf() ); 
			flip(t);
			goto_child(t);
		}
		return *this;
	}

	inline pooled_veb_navigator& next_dfs()
	{
		if( !is_root() )
		{
			basic_order_t ch = goto_parent();
			flip(ch);
			if( ch==order-1 )
				return *this;
			flip(ch+1);
			goto_child(ch+1);
		}
		while( !is_leaf() )
		{
			flip(0);
			goto_child(0);
		}
		return *this;
	}

	inline order_t leaf_index()
		{ assert( is_leaf() ); return bfsi-left_bfsi; }

	inline level_iterator begin_level(height_t level)
		{ return level_iterator(*root_path,k,level,false); }
	inline level_iterator end_level(height_t level)
		{ return level_iterator(*root_path,k,level,true); }

	template<class Functor>
	inline Functor enum_buffers(Functor f)
		{ return enum_buffers(f, root_path[0], 0, static_cast<height_t>(leaf-root_path+1), static_cast<height_t>(leaf-root_path+1), &vebi); }

protected:
	inline void flip(basic_order_t i)
	{
		*current->input(i) = buffer(current->input_begin(i), current->input(i)->begin());
	}

	template<class Functor>
	static Functor enum_buffers(Functor f, Node* base, height_t h1, height_t h2, height_t h0, veb_index<Node::order,Splitter,Node,MAX_MERGE_TREE_HEIGHT> *tables)
	{
		if( h2-h1 > 1 )
		{
			height_t top_height = Splitter::split(h2-h1);
			height_t h = h1+top_height;
			order_t top_tree_k = tables->n(h);
			f = enum_buffers(f, base, h1, h, h0, tables);
			for( base-=tables->t(h); top_tree_k; base-=tables->b(h), --top_tree_k )
				f = enum_buffers(f, base, h, h2, h0, tables);
		}
		else if( h2 < h0 )
			for( basic_order_t i=0; i!=order; ++i )
				f(base->input(i), base->input_begin(i),base->input_end(i)); 
		return f;
	}

private:
	veb_index<Node::order,Splitter,Node,MAX_MERGE_TREE_HEIGHT> vebi;
	Node *current;
	Node **path;
	order_t bfsi;
	buffer *boutput;
	order_t left_bfsi;
	Node **leaf;
	order_t k;
	Node *root_path[MAX_MERGE_TREE_HEIGHT];

	template<class T>
	static inline void POOLED_VEB_NAVIGATOR_REQUIRES_STACK_ALLOCATOR(stack_allocator<T>&) { }
};

	template<class Node, class Splitter>
	class pveb_level_iterator
	{
		friend class pooled_veb_navigator<Node,Splitter>;
	public:
		inline pveb_level_iterator& operator++()
		{
			height_t i;
			for( i=level; !nav.is_root() && nav.bfsi%Node::order == Node::order-1; --i )
				nav.goto_parent();
			if( nav.is_root() )
				return *this;
			basic_order_t ch = static_cast<basic_order_t>(nav.bfsi%Node::order)+1;
			nav.goto_parent();
			nav.goto_child(ch);
			for( ; i<level; ++i )
				nav.goto_child(0);
			return *this;
		}
		inline pveb_level_iterator& operator--()
		{
			height_t i;
			if( nav.is_root() )
			{
				for( i=0; i<level; ++i )
					nav.goto_child(Node::order-1);
				return *this;
			}
			for( i=level; !nav.is_root() && nav.bfsi%Node::order == 0; --i )
				nav.goto_parent();
			if( nav.is_root() )
				return *this;
			basic_order_t ch = static_cast<basic_order_t>(nav.bfsi%Node::order)-1;
			nav.goto_parent();
			nav.goto_child(ch);
			for( ; i<level; ++i )
				nav.goto_child(Node::order-1);
			return *this;
		}
		inline bool operator==(const pveb_level_iterator& i)
			{ return nav == i.nav.mark(); }
		inline bool operator!=(const pveb_level_iterator& i)
			{ return nav != i.nav.mark(); }
		inline Node* operator*()
			{ return nav.root_path[level]; }
		inline Node* child(basic_order_t i)
		{
			nav.goto_child(i);
			Node *ret = nav.current;
			nav.goto_parent();
			return ret;
		}
		inline typename Node::pointer input_begin(basic_order_t i)
			{ return nav.current->input_begin(i); }
	protected:
		inline pveb_level_iterator(Node* root, order_t k, height_t level, bool end) : level(level), nav(root,k)
		{
			if( !end )
				for( height_t i=0; i<level; ++i )
					nav.goto_child(0);
		}

		height_t level;
		pooled_veb_navigator<Node,Splitter> nav;
	};

} // namespace funnel


#ifdef __GNUC__
#undef inline
#endif // __GNUC__
#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __LAYOUT_PVEB_H_INCLUDED_
