/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __MERGE_TREE_H_INCLUDED_
#define __MERGE_TREE_H_INCLUDED_

#include "funnel_structs.h" // for convenience
#include "funnel_base.h"
#include "basic_merger.h"
#include <stdexcept>
#include <vector> // for cache datastructure
#include "../int64.h"

#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER
#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__


namespace funnel
{

#ifdef _CUBE_BUFFERS
template<order_t order>
class splitter
{
public:
	static inline order_t prefered_order_output(_uint64_t N)
		{ return (order_t)floor(pow((double)N, 1.0/3)+.5); }
	static inline order_t prefered_order_input(_uint64_t N)
		{ return (order_t)floor(pow((double)N, 1.0/3)+.5); }
	static inline uint64_t switch_to_cache_aware()
		{ return static_cast<funnel::order_t>(floor(pow((order+0.5f), 3)+.5))+1; }
	static inline height_t split(height_t h)
		{ return h/2; }
	static inline size_t out_buffer_size(order_t k)
		{ return static_cast<size_t>(k)*k*k; }
};
#else // _CUBE_BUFFERS
template<order_t order>
class splitter
{
public:
	static inline order_t prefered_order_output(_uint64_t N)
		{ return (order_t)floor(pow((double)N/32, 1.0/2.0)+.5); }
	static inline order_t prefered_order_input(_uint64_t N)
		{ return (order_t)floor(pow((double)N/32, 1.0/2.0)+.5); }
	static inline uint64_t switch_to_cache_aware()
		{ return static_cast<funnel::order_t>(floor(pow(32*(order+0.5f), 2)+.5))+1; }
	static inline height_t split(height_t h)
		{ return h/2; }
	static inline size_t out_buffer_size(order_t k)
		{ return static_cast<size_t>(floor(32*pow((double)k, 2)+.5)); }
};
#endif // _CUBE_BUFFERS

template<class It> class nop_refill { public: template<class Merger> inline bool operator()(const Merger*, It, It)
	{ return false; } };

template<class T, class InStream, class Layout, class BasicMerger, class Splitter, class Refiller>
class merge_tree_base
{
protected:
	typedef typename Layout::navigator::token token;
	typedef typename Layout::navigator::bookmark bookmark;
	typedef typename Layout::navigator::buffer buffer;
	typedef typename Layout::node node;
public:
	enum { order = node::order };
	typedef Splitter splitter;
	typedef typename Layout::allocator allocator;
	typedef InStream stream;
	typedef InStream* stream_iterator;
	typedef T value_type;

protected:
	typedef typename allocator::template rebind<InStream>::other sallocator;

	inline merge_tree_base(size_t k) : k(static_cast<order_t>(k)), m_bCold(true)
		{ construct(static_cast<order_t>(k)); }
	inline merge_tree_base(size_t k, allocator& alloc) : k(static_cast<order_t>(k)), m_bCold(true), alloc(alloc), salloc(alloc)
		{ construct(static_cast<order_t>(k)); }
	virtual ~merge_tree_base()
	{
		for( InStream *s=input_streams; s!=input_streams+k; ++s )
			salloc.destroy(s);
		salloc.deallocate(input_streams, k);
		Layout::destroy(m_pRoot, k, alloc);
	}
public:
	static inline order_t min_order()
		{ return order; }
	static inline typename allocator::size_type size_of(size_t k)
		{ return Layout::size_of(static_cast<order_t>(k))+k*sizeof(InStream); }

	inline void add_stream(const InStream& stream)
		{ *last_stream++ = stream; }
	inline stream_iterator begin()
		{ return input_streams; }
	inline stream_iterator end()
		{ return last_stream; }
	inline void reset()
	{
		if( !m_bCold )
		{
			InStream dflt;
			for( InStream *s=input_streams; s!=last_stream; ++s )
			{
				salloc.destroy(s);
				salloc.construct(s,dflt);
			}
			buffer_reseter br;
			typename Layout::navigator nav(m_pRoot,k);
			nav.enum_buffers(br);
			m_bCold = true;
		}
		last_stream = input_streams;
	}

	inline void set_refiller(const Refiller& r)
		{ m_Refiller = r; }
	inline const Refiller& get_refiller() const
		{ return m_Refiller; }

	template<class It>
	It empty(It begin, It end)
	{
		empty_buffers<It> eb(begin,end);
		typename Layout::navigator nav(m_pRoot,k);
		nav.enum_buffers(eb);
		return eb.begin();
	}

protected:
	class buffer_reseter
	{
	public:
		inline void operator()(typename Layout::navigator::buffer* buf, typename Layout::navigator::buffer::pointer, typename Layout::navigator::buffer::pointer end)
			{ *buf = typename Layout::navigator::buffer(end, end); }
	};
	template<class It>
	class empty_buffers
	{
	public:
		empty_buffers(It b, It e) : b(b), e(e) { }
		inline void operator()(typename Layout::navigator::buffer* buf, typename Layout::navigator::buffer::pointer, typename Layout::navigator::buffer::pointer)
		{
			if( b == e )
				assert( buf->begin() == buf->end() ); // This should at least be the case, when using Inplace Mergesort
			typename Layout::navigator::buffer::pointer p;
			for( p = buf->begin(); b!=e && p && p!=buf->end(); b++, p++ )
				*b = *p;
			if( p && b == e )
				assert( p == buf->end() ); // This should at least be the case, when using Inplace Mergesort
		}
		inline It begin() const
			{ return b; }
	private:
		It b, e;
	};

	void construct(order_t k)
	{
		if(!( order < k ))
			throw std::invalid_argument("Merge tree too small");
		m_pRoot = Layout::do_layout(k, alloc);
		try
		{
			last_stream = input_streams = salloc.allocate(k);
		}
		catch( ... )
		{
			Layout::destroy(m_pRoot, k, alloc);
			throw;
		}
		InStream *s = input_streams;
		try
		{
			InStream dflt;
			for( ; s!=input_streams+k; ++s )
				salloc.construct(s,dflt);
		}
		catch( ... )
		{
			for( --s; s>=input_streams; --s )
				salloc.destroy(s);
			salloc.deallocate(input_streams, k);
			Layout::destroy(m_pRoot, k, alloc);
			throw;
		}
	}

	order_t k;
	bool m_bCold;
	typename Layout::node *m_pRoot;
	InStream *input_streams, *last_stream;
	allocator alloc;
	sallocator salloc;
	Refiller m_Refiller;
};

template<class T, class InStream, class Layout, class BasicMerger = comp_merger<Layout::navigator::order, typename Layout::navigator::buffer, typename Layout::navigator::token>, class Splitter = splitter<Layout::navigator::order>, class Refiller = nop_refill<typename InStream::pointer> >
class merge_tree : public merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>
{
	using merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>::m_pRoot;
	using merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>::k;
	using merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>::input_streams;
	using merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>::last_stream;
	using merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>::m_bCold;
	using merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>::m_Refiller;
protected:
	typedef typename Layout::navigator::buffer buffer;
	typedef typename Layout::navigator::token token;
	typedef typename Layout::navigator::bookmark bookmark;
	typedef typename Layout::node node;
public:
	enum { order = node::order };
	typedef typename Layout::allocator allocator;
	typedef InStream stream;
	typedef InStream* stream_iterator;
	typedef T value_type;
protected:
	typedef typename allocator::template rebind<InStream>::other sallocator;
	typedef typename BasicMerger::template rebind<order,buffer,token>::other node_merger;
	typedef typename BasicMerger::template rebind<order,InStream,char>::other leaf_merger;

public:
	template<class NewRefiller>
	struct rebind { typedef merge_tree<T, InStream, Layout, BasicMerger, Splitter, NewRefiller> other; };

	explicit inline merge_tree(size_t k) : merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>(static_cast<order_t>(k)) { }
	inline merge_tree(size_t k, allocator& alloc) : merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>(static_cast<order_t>(k),alloc) { }
	template<class It>
	inline It operator()(It begin, It end)
	{
		if( m_bCold )
			if( warmup() )
				return begin;
		m_bCold = false;
		return fill_root(begin, end);
	}
protected:
	bool warmup()
	{
		typename Layout::navigator nav(m_pRoot,k);
		for( nav.next_dfs(); !nav.is_root(); nav.next_dfs() )
		{
			if( nav.is_leaf() )
			{
				leaf_merger merger;
				InStream is[order];
				InStream *bs = input_streams+order*nav.leaf_index();
				if( bs < last_stream )
				{
					basic_order_t active = static_cast<basic_order_t>(order < last_stream-bs? order : last_stream-bs);
					for( basic_order_t i=0; i!=active; ++i )
					{
						is[i] = bs[i];
						merger.add_stream(is+i,0);
					}
					typename buffer::pointer b = nav.output()->begin(), e = nav.output()->end();
					merger(b,e,0);
					*nav.output() = buffer(b,e);
					bool abort = false;
					for( basic_order_t i=0; i!=active; ++i )
					{
						InStream s = bs[i];
						bs[i] = is[i];
						abort = abort || m_Refiller(this, s.begin(), bs[i].begin());
					}
					if( abort )
						return true;
				}
			}
			else
			{
				assert( !nav.is_leaf() && !nav.is_root() );
				bookmark m = nav.mark();
				for( ;; )
				{
					node_merger merger;
					for( basic_order_t i=0; i<order; ++i )
						merger.add_stream(nav.input(i),nav.child(i));
					typename buffer::pointer b = nav.output()->begin(), e = nav.output()->end();
					token t = merger(b,e,nav.parent());
					*nav.output() = buffer(b,e);
					if( t == nav.parent() )
						break;
					nav += t;
					while( nav != m )
						if( nav.is_leaf() )
						{
							leaf_merger merger;
							InStream is[order];
							InStream *bs = input_streams+order*nav.leaf_index();
							if( bs < last_stream )
							{
								basic_order_t active = static_cast<basic_order_t>(order < last_stream-bs? order : last_stream-bs);
								for( basic_order_t i=0; i!=active; ++i )
								{
									is[i] = bs[i];
									merger.add_stream(is+i,0);
								}
								typename buffer::pointer b = nav.output()->begin(), e = nav.output()->end();
								merger(b,e,0);
								*nav.output() = buffer(b,e);
								bool abort = false;
								for( basic_order_t i=0; i!=active; ++i )
								{
									typename InStream::pointer b = bs[i].begin();
									bs[i] = InStream(is[i].begin(),bs[i].end());
									abort = abort || m_Refiller(this, b, is[i].begin());
								}
								if( abort )
									return true;
							}
							nav += nav.parent();
						}
						else
						{
							node_merger merger;
							for( basic_order_t i=0; i<order; ++i )
								merger.add_stream(nav.input(i),nav.child(i));
							typename buffer::pointer b = nav.output()->begin(), e = nav.output()->end();
							token t = merger(b,e,nav.parent());
							*nav.output() = buffer(b,e);
							nav += t;
						}
				}
			}
		}
		return false;

	}
	template<class It>
	It fill_root(It begin, It end)
	{
		typename Layout::navigator nav(m_pRoot,k);

		for( ;; )
		{
			node_merger merger;
			for( basic_order_t i=0; i<order; ++i )
				merger.add_stream(nav.input(i),nav.child(i));
			token t = merger(begin, end, nav.parent());
			if( t == nav.parent() )
				return begin;
			nav += t;
			while( !nav.is_root() )
			{
				if( nav.is_leaf() )
				{
					leaf_merger merger;
					InStream is[order];
					InStream *bs = input_streams+order*nav.leaf_index();
					assert( bs < last_stream );
					{
						basic_order_t active = static_cast<basic_order_t>(order < last_stream-bs? order : last_stream-bs);
						for( basic_order_t i=0; i!=active; ++i )
						{
							is[i] = bs[i];
							merger.add_stream(is+i,0);
						}
						typename buffer::pointer b = nav.output()->begin(), e = nav.output()->end();
						merger(b,e,0);
						*nav.output() = buffer(b,e);
						bool abort = false;
						for( basic_order_t i=0; i!=active; ++i )
						{
							typename InStream::pointer b = bs[i].begin();
							bs[i] = InStream(is[i].begin(),bs[i].end());
							abort = abort || m_Refiller(this, b, is[i].begin());
						}
						if( abort )
							return begin;
					}
					nav += nav.parent();
				}
				if( !nav.is_root() )
				{
					node_merger merger;
					for( basic_order_t i=0; i<order; ++i )
						merger.add_stream(nav.input(i),nav.child(i));
					typename buffer::pointer b = nav.output()->begin(), e = nav.output()->end();
					token t = merger(b,e,nav.parent());
					*nav.output() = buffer(b,e);
					nav += t;
				}
			}
		}
	}
};

template<class T, class InStream, class Layout, class BasicMerger = comp_merger<Layout::navigator::order, typename Layout::navigator::buffer, typename Layout::navigator::token>, class Splitter = splitter<Layout::navigator::order>, class Refiller = nop_refill<typename InStream::pointer> >
class rec_merge_tree : public merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>
{
	using merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>::m_pRoot;
	using merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>::k;
	using merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>::input_streams;
	using merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>::last_stream;
	using merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>::m_bCold;
	using merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>::m_Refiller;
protected:
	typedef typename Layout::navigator::buffer buffer;
	typedef typename Layout::navigator::token token;
	typedef typename Layout::navigator::bookmark bookmark;
	typedef typename Layout::node node;
public:
	enum { order = node::order };
//	typedef splitter<order> splitter;
	typedef typename Layout::allocator allocator;
	typedef InStream stream;
	typedef InStream* stream_iterator;
	typedef T value_type;
protected:
	typedef typename allocator::template rebind<InStream>::other sallocator;
	typedef typename BasicMerger::template rebind<order,buffer,token>::other node_merger;
	typedef typename BasicMerger::template rebind<order,InStream,char>::other leaf_merger;

public:
	template<class NewRefiller>
	struct rebind { typedef rec_merge_tree<T, InStream, Layout, BasicMerger, Splitter, NewRefiller> other; };

	explicit inline rec_merge_tree(size_t k) : merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>(k)
	{
		bfs_index<order> bfs;
		for( height_t l=1; l<logc<order>(static_cast<order_t>(k)); ++l )
			bfs.child(0);
		leaf_index = bfs;
	}
	inline rec_merge_tree(size_t k, allocator& alloc) : merge_tree_base<T,InStream,Layout,BasicMerger,Splitter,Refiller>(k,alloc)
	{
		bfs_index<order> bfs;
		for( height_t l=1; l<logc<order>(static_cast<order_t>(k)); ++l )
			bfs.child(0);
		leaf_index = bfs;
	}
	template<class It>
	inline It operator()(It begin, It end)
	{
		if( m_bCold )
			for( basic_order_t i=0; i<order; ++i )
				if( m_pRoot->child(i) )
				{
					bfs_index<order> c; c.child(i);
					flip(m_pRoot,i);
					warmup(m_pRoot->child(i),m_pRoot->input(i),c);
					flip(m_pRoot,i);
				}
		m_bCold = false;
		return fill_root(begin, end);
	}
	template<class OutIt>
	inline OutIt operator()(OutIt begin)
	{
		if( m_bCold )
			for( basic_order_t i=0; i<order; ++i )
				if( m_pRoot->child(i) )
				{
					bfs_index<order> c; c.child(i);
					flip(m_pRoot,i);
					warmup(m_pRoot->child(i),m_pRoot->input(i),c);
					flip(m_pRoot,i);
				}
		m_bCold = false;
		return fill_root(begin);
	}
protected:
	static inline void flip(typename Layout::node *n, basic_order_t t)
		{ *n->input(t) = buffer(n->input_begin(t), n->input(t)->begin()); }
	void fill_leaf(buffer *output, bfs_index<order> index)
	{
		leaf_merger merger;
		InStream is[order];
		InStream *bs = input_streams+order*(index-leaf_index);
		assert( bs < last_stream );
		basic_order_t active = static_cast<basic_order_t>(order < last_stream-bs? order : last_stream-bs);
		for( basic_order_t i=0; i!=active; ++i )
		{
			is[i] = bs[i];
			merger.add_stream(is+i,0);
		}
		typename buffer::pointer b = output->begin(), e = output->end();
		merger(b,e,0);
		*output = buffer(b,e);
		for( basic_order_t i=0; i!=active; ++i )
		{
			typename InStream::pointer b = bs[i].begin();
			bs[i] = InStream(is[i].begin(),bs[i].end());
			m_Refiller(this, b, is[i].begin());
		}
	}
	void fill(typename Layout::node *n, buffer *output, bfs_index<order> index)
	{
		for( ;; )
		{
			node_merger merger;
			for( basic_order_t i=0; i<order; ++i )
				merger.add_stream(n->input(i),i);
			typename buffer::pointer b = output->begin(), e = output->end();
			basic_order_t t = merger(b,e,order);
			*output = buffer(b,e);
			if( t == order )
				return;
			bfs_index<order> c = index; c.child(t);
			flip(n,t);
			if( n->child(t)->child(0) == NULL )
				fill_leaf(n->input(t),c);
			else
				fill(n->child(t),n->input(t),c);
			flip(n,t);
		}
	}
	template<class It>
	inline It fill_root(It begin, It end)
	{
		for( ;; )
		{
			node_merger merger;
			for( basic_order_t i=0; i<order; ++i )
				merger.add_stream(m_pRoot->input(i),i);
			basic_order_t t = merger(begin, end, order);
			if( t == order )
				return begin;
			bfs_index<order> c; c.child(t);
			flip(m_pRoot,t);
			if( m_pRoot->child(t)->child(0) == NULL )
				fill_leaf(m_pRoot->input(t),c);
			else
				fill(m_pRoot->child(t),m_pRoot->input(t),c);
			flip(m_pRoot,t);
		}
	}
	template<class OutIt>
	inline OutIt fill_root(OutIt begin)
	{
		for( ;; )
		{
			node_merger merger;
			for( basic_order_t i=0; i<order; ++i )
				merger.add_stream(m_pRoot->input(i),i);
			basic_order_t t = merger(begin, order);
			if( t == order )
				return begin;
			bfs_index<order> c; c.child(t);
			flip(m_pRoot,t);
			if( m_pRoot->child(t)->child(0) == NULL )
				fill_leaf(m_pRoot->input(t),c);
			else
				fill(m_pRoot->child(t),m_pRoot->input(t),c);
			flip(m_pRoot,t);
		}
	}
	void warmup(typename Layout::node *n, buffer *output, bfs_index<order> index)
	{
		if( n->child(0) )
		{
			for( basic_order_t i=0; i<order; ++i )
				if( n->child(i) )
				{
					bfs_index<order> c = index; c.child(i);
					flip(n,i);
					warmup(n->child(i), n->input(i), c);
					flip(n,i);
				}
			fill(n,output,index);
		}
		else if( input_streams+order*(index-leaf_index) < last_stream )
				fill_leaf(output,index);
	}


	order_t leaf_index;
};

} // namespace funnel


#ifdef __GNUC__
#undef inline
#endif // __GNUC__
#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __MERGE_TREE_H_INCLUDED_
