/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __POINTER_NAVIGATOR_H_INCLUDED_
#define __POINTER_NAVIGATOR_H_INCLUDED_

#include "funnel_base.h"

#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__
#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER


namespace funnel
{

template<class Node, class Splitter>
class navigator_base
{
protected:
	inline navigator_base(Node *root, order_t k)
	{
		path = root_path;
		current = *path = root;
		boutput = NULL;
		height_t h = logc<order>(k);
		bfs_index<order> bfs;
		for( height_t l=1; l<h; ++l )
			bfs.child(0);
		left_bfsi = bfs;
	}

	inline basic_order_t goto_parent()
	{
		assert( !is_root() ); 
		basic_order_t ch = bfsi.child_index();
		current = *--path;
		bfsi.parent();
		if( path > root_path )
			boutput = path[-1]->input(bfsi.child_index());
		return ch;
	}
public:
	template<class Allocator>
	static inline void REQUIRE_ALLOCATOR(Allocator&) { }

	typedef Node node;
	typedef basic_order_t token;
	typedef Node* bookmark;
	typedef buffer<typename Node::pointer> buffer;
	enum { order = Node::order };

	inline bool is_root() const
		{ return path == root_path; }

	inline bookmark mark()
		{ return current; }
	inline bool operator==(bookmark m) const
		{ return current == m; }
	inline bool operator!=(bookmark m) const
		{ return current != m; }

	inline buffer *input(basic_order_t i)
		{ return current->input(i); }
	inline buffer *output()
		{ return boutput; }

protected:
	Node *current;
	Node **path;
	Node *root_path[MAX_MERGE_TREE_HEIGHT];
	bfs_index<order> bfsi;
	buffer *boutput;
	order_t left_bfsi;
};

template<class Node, class Splitter>
class pointer_navigator_base : public navigator_base<Node, Splitter>
{
protected:
	using navigator_base<Node,Splitter>::order;
	using navigator_base<Node,Splitter>::root_path;
	using navigator_base<Node,Splitter>::current;
	using navigator_base<Node,Splitter>::path;
	using navigator_base<Node,Splitter>::boutput;
	using navigator_base<Node,Splitter>::bfsi;
	using navigator_base<Node,Splitter>::left_bfsi;

public:
	class level_iterator_base
	{
	public:
		inline level_iterator_base& operator++()
		{
			order_t i; basic_order_t c;
			for( i=level; i && root_path[i-1]->child(Node::order-1) == root_path[i]; --i );
			if( i == 0 )
			{
				root_path[level] = root_path[0];
				return *this;
			}
			assert( root_path[i-1]->child(Node::order-1) != root_path[i] );
			for( c=0; c<Node::order-1; ++c )
				if( root_path[i-1]->child(c) == root_path[i] )
					break;
			root_path[i] = root_path[i-1]->child(c+1);
			if( root_path[i] == NULL )
			{
				root_path[level] = root_path[0];
				return *this;
			}
			for( ; i<level; ++i )
				root_path[i+1] = root_path[i]->child(0);
			return *this;
		}
		inline level_iterator_base& operator--()
		{
			order_t i; basic_order_t c;
			if( root_path[level] == root_path[0] )
			{
				for( i=0; i<level; ++i )
				{
					for( c=Node::order-1; c>=0; --c )
						if( root_path[i]->child(c) )
							break;
					root_path[i+1] = root_path[i]->child(c);
				}
				return *this;
			}
			for( i=level; i && root_path[i-1]->child(0) == root_path[i]; --i );
			if( i == 0 )
			{
				root_path[level] = root_path[0];
				return *this;
			}
			assert( root_path[i-1]->child(0) != root_path[i] );
			for( c=1; c<Node::order; ++c )
				if( root_path[i-1]->child(c) == root_path[i] )
					break;
			root_path[i] = root_path[i-1]->child(c-1);
			for( ; i<level; ++i )
				root_path[i+1] = root_path[i]->child(Node::order-1);
			return *this;
		}
		inline bool operator==(const level_iterator_base& i)
			{ return root_path[level] == i.root_path[i.level]; }
		inline bool operator!=(const level_iterator_base& i)
			{ return root_path[level] != i.root_path[i.level]; }
		inline Node* operator*()
			{ return root_path[level]; }
		inline Node* child(basic_order_t i)
			{ return root_path[level]->child(i); }
	protected:
		inline level_iterator_base(Node* root, height_t level, bool end) : level(level)
		{
			root_path[0] = root;
			if( end )
				root_path[level] = root;
			else
				for( height_t i=0; i<level; i++ )
					root_path[i+1] = root_path[i]->child(0);
		}

		height_t level;
		Node *root_path[MAX_MERGE_TREE_HEIGHT];
	};

	inline void goto_child(basic_order_t i)
	{
		boutput = current->input(i);
		current = *++path = current->child(i);
		bfsi.child(i);
	}
public:
	using navigator_base<Node,Splitter>::input;
	using navigator_base<Node,Splitter>::is_root;
	typedef typename navigator_base<Node,Splitter>::buffer buffer;
	typedef basic_order_t token;

	inline pointer_navigator_base(Node *root, order_t k) : navigator_base<Node,Splitter>(root,k) { }

	inline bool is_leaf() const
		{ return current->child(0) == NULL; }

	inline basic_order_t leaf_index()
		{ assert( left_bfsi <= bfsi ); return static_cast<basic_order_t>(bfsi-left_bfsi); }
};

template<class Node, class Splitter>
class pointer_flip_navigator : public pointer_navigator_base<Node,Splitter>
{
	using pointer_navigator_base<Node,Splitter>::root_path;
	using pointer_navigator_base<Node,Splitter>::current;
	using pointer_navigator_base<Node,Splitter>::goto_child;
	using pointer_navigator_base<Node,Splitter>::goto_parent;
protected:
	class flip_level_iterator : public pointer_navigator_base<Node,Splitter>::level_iterator_base
	{
#ifdef _MSC_VER
		using level_iterator_base::level;
		using level_iterator_base::root_path;
#else //  _MSC_VER
		using pointer_navigator_base<Node,Splitter>::level_iterator_base::level;
		using pointer_navigator_base<Node,Splitter>::level_iterator_base::root_path;
#endif //  _MSC_VER
		friend class pointer_flip_navigator<Node,Splitter>;
	public:
		inline typename Node::pointer input_begin(basic_order_t i)
			{ return root_path[level]->input_begin(i); }
	protected:
		inline flip_level_iterator(Node* root, height_t level, bool end) : pointer_navigator_base<Node,Splitter>::level_iterator_base(root,level,end) { }
	};
public:
	using pointer_navigator_base<Node,Splitter>::order;
	using pointer_navigator_base<Node,Splitter>::is_root;
	using pointer_navigator_base<Node,Splitter>::is_leaf;
	typedef flip_level_iterator level_iterator;
	typedef basic_order_t token;

	inline pointer_flip_navigator(Node *root, order_t k) : pointer_navigator_base<Node,Splitter>(root,k) { }

	inline token child(basic_order_t i) const
		{ return i; }
	inline token parent() const
		{ return order; }
	inline pointer_flip_navigator& operator+=(token t)
	{
		if( t == order ) // goto parent
		{
			basic_order_t ch = goto_parent();
			flip(ch);
		}
		else // goto child
		{
			assert( !is_leaf() ); 
			flip(t);
			goto_child(t);
		}
		return *this;
	}

	inline pointer_flip_navigator& next_dfs()
	{
		if( !is_root() )
		{
			basic_order_t ch = goto_parent();
			flip(ch);
			if( ch==order-1 || current->child(ch+1)==NULL )
				return *this;
			flip(ch+1);
			goto_child(ch+1);
		}
		while( !is_leaf() )
		{
			flip(0);
			goto_child(0);
		}
		return *this;
	}

	inline flip_level_iterator begin_level(height_t level)
		{ return flip_level_iterator(*root_path,level,false); }
	inline level_iterator end_level(height_t level)
		{ return flip_level_iterator(*root_path,level,true); }

	template<class Functor>
	Functor enum_buffers(Functor f)
		{ return enum_buffers(*root_path,f); }

protected:
	template<class Functor>
	static Functor enum_buffers(Node* n, Functor f)
	{
		for( basic_order_t i=0; i<order; ++i )
			if( n->child(i) )
			{
				f(n->input(i), n->input_begin(i), n->input_end(i));
				f = enum_buffers(n->child(i), f);
			}
		return f;
	}

	inline void flip(basic_order_t i)
		{ current->flip_buffer(i); }
};


template<class Node, class Splitter>
class pointer_navigator : public pointer_navigator_base<Node,Splitter>
{
	using pointer_navigator_base<Node,Splitter>::root_path;
	using pointer_navigator_base<Node,Splitter>::current;
	using pointer_navigator_base<Node,Splitter>::goto_child;
	using pointer_navigator_base<Node,Splitter>::goto_parent;
protected:
	class literator : public pointer_navigator_base<Node,Splitter>::level_iterator_base
	{
#ifdef _MSC_VER
		using level_iterator_base::level;
		using level_iterator_base::root_path;
#else //  _MSC_VER
		using pointer_navigator_base<Node,Splitter>::level_iterator_base::level;
		using pointer_navigator_base<Node,Splitter>::level_iterator_base::root_path;
#endif //  _MSC_VER
		friend class pointer_navigator<Node,Splitter>;
	public:
		inline typename Node::pointer input_begin(basic_order_t i)
		{
			typename Node::pointer p = reinterpret_cast<typename Node::pointer>(root_path[level]->child(i)+1);
			assert( p == root_path[level]->input_begin(i) );
			return p;
		}
	protected:
		inline literator(Node* root, height_t level, bool end) : pointer_navigator_base<Node,Splitter>::level_iterator_base(root,level,end) { }
	};
public:
	using pointer_navigator_base<Node,Splitter>::order;
	using pointer_navigator_base<Node,Splitter>::input;
	using pointer_navigator_base<Node,Splitter>::is_root;
	using pointer_navigator_base<Node,Splitter>::is_leaf;
	typedef typename pointer_navigator_base<Node,Splitter>::buffer buffer;
	typedef basic_order_t token;
	typedef literator level_iterator;

	inline pointer_navigator(Node *root, order_t k) : pointer_navigator_base<Node,Splitter>(root,k)
	{
		height_t height = logc<Node::order>(k);
		compute_buffer_sizes(buffer_sizes,buffer_sizes+height);
		buffer_size = buffer_sizes+1;
	}

	inline token child(basic_order_t i) const
		{ return i; }
	inline token parent() const
		{ return order; }
	inline pointer_navigator& operator+=(token t)
	{
		if( t == order ) // goto parent
		{
			basic_order_t ch = goto_parent();
			buffer_size--;
			flip(ch);
		}
		else // goto child
		{
			assert( !is_leaf() );
			flip(t);
			buffer_size++;
			goto_child(t);
		}
		return *this;
	}

	inline pointer_navigator& next_dfs()
	{
		if( !is_root() )
		{
			basic_order_t ch = goto_parent();
			buffer_size--;
			flip(ch);
			if( ch==order-1 || current->child(ch+1)==NULL )
				return *this;
			flip(ch+1);
			buffer_size++;
			goto_child(ch+1);
		}
		while( !is_leaf() )
		{
			flip(0);
			buffer_size++;
			goto_child(0);
		}
		return *this;
	}

	inline literator begin_level(height_t level)
		{ return literator(*root_path,level,false); }
	inline literator end_level(height_t level)
		{ return literator(*root_path,level,true); }

	template<class Functor>
	Functor enum_buffers(Functor f)
		{ return enum_buffers(*root_path,buffer_sizes+1,f); }

protected:
	template<class Functor>
	static Functor enum_buffers(Node* n, size_t *bs, Functor f)
	{
		for( basic_order_t i=0; i<order; ++i )
			if( n->child(i) )
			{
				typename Node::pointer b = reinterpret_cast<typename Node::pointer>(n->child(i)+1);
				assert( n->input_begin(i) == b );
				assert( n->input_end(i) == b+*bs );
				f(n->input(i), b, b+*bs);
				f = enum_buffers(n->child(i), bs+1, f);
			}
		return f;
	}

	inline void flip(basic_order_t i)
	{
		typename Node::pointer b = reinterpret_cast<typename Node::pointer>(current->child(i)+1);
		assert( b == current->input_begin(i) );
		assert( b+*buffer_size == current->input_end(i) );
		*input(i) = buffer(b, input(i)->begin());
	}

private:
	static void compute_buffer_sizes(size_t *b, size_t *e)
	{
		assert( b <= e );
		height_t sp = Splitter::split(static_cast<height_t>(e-b));
		if( e-b > 2 )
		{
			compute_buffer_sizes(b, b+sp);
			compute_buffer_sizes(b+sp, e);
		}
		b[sp] = Splitter::out_buffer_size(pow_of_order<order>(static_cast<height_t>(e-b-sp)));
	}

	size_t buffer_sizes[MAX_MERGE_TREE_HEIGHT];
	size_t *buffer_size;
};

} // namespace funnel


#ifdef __GNUC__
#undef inline
#endif // __GNUC__
#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER

#endif // __POINTER_NAVIGATOR_H_INCLUDED_
