/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __STACK_ALLOCATOR_H_INCLUDED_
#define __STACK_ALLOCATOR_H_INCLUDED_

#include <exception>
#include <stdexcept>
#include <new> // for bad_alloc

#ifdef _MSC_VER
#define inline __forceinline
#endif // _MSC_VER
#ifdef __GNUC__
#define inline inline __attribute__ ((always_inline))
#endif // __GNUC__


class stack_allocator_head
{
protected:
	class stack_bad_alloc : public std::bad_alloc
	{
	public:
		stack_bad_alloc(const char *what) throw() : m_szWhat(what) { }
		virtual const char *what() const throw()
			{ return m_szWhat; }
	private:
		const char *m_szWhat;
	};
	mutable struct head
	{
		int ref_count;
		char *next;
		void *last;
#ifndef NDEBUG
		void *first;
#endif // NDEBUG
	} *m_pHead;
	static const int allignment = sizeof(long);

	stack_allocator_head() throw() : m_pHead(NULL) { }
	stack_allocator_head(head* pHead) throw() : m_pHead(pHead) { }
};

template<class T>
class stack_allocator : public stack_allocator_head
{
	template<class U>
	friend class stack_allocator;
public:
	typedef size_t size_type;
	typedef ptrdiff_t difference_type;
	typedef T* pointer;
	typedef const T* const_pointer;
	typedef T& reference;
	typedef const T& const_reference;
	typedef T value_type;

	template<class U>
	struct rebind { typedef stack_allocator<U> other; };

	explicit stack_allocator(size_type max_size)
	{
		if( max_size )
		{
			size_type s = max_size+((allignment-(max_size&(allignment-1)))&(allignment-1));
			m_pHead = reinterpret_cast<head*>(new char[s+sizeof(head)]);
			if( m_pHead == NULL )
				throw stack_bad_alloc("stack_allocator: chould not allocate pool");
			m_pHead->next = reinterpret_cast<char*>(m_pHead)+s+sizeof(head);
			m_pHead->last = m_pHead->next-max_size;
			m_pHead->ref_count = 1;
#ifndef NDEBUG
			m_pHead->first = m_pHead->next;
#endif // NDEBUG
		}
		else
			m_pHead = NULL;
	}
	stack_allocator(const stack_allocator<T>& a) throw() : stack_allocator_head(a.m_pHead)
	{
		if( m_pHead )
			m_pHead->ref_count++;
	}
	template<class U>
	stack_allocator(const stack_allocator<U>& a) throw() : stack_allocator_head(a.m_pHead)
	{
		if( m_pHead )
			m_pHead->ref_count++;
	}
	~stack_allocator() throw()
	{
		if( m_pHead )
		{
			assert( m_pHead->ref_count );
			m_pHead->ref_count--;
			if( m_pHead->ref_count == 0 )
			{
				assert( m_pHead->next == m_pHead->first );
				delete[] reinterpret_cast<char*>(m_pHead);
			}
		}
	}

	static inline pointer address(reference x)
		{ return &x; }
	static inline const_pointer address(const_reference x)
		{ return &x; }
	inline pointer allocate(size_type n, typename stack_allocator<T>::const_pointer hint = 0)
	{
		(void)hint;
		if( m_pHead )
		{
			m_pHead->next -= sizeof(T)*n;
			if( m_pHead->next < m_pHead->last )
			{
				m_pHead->next += sizeof(T)*n;
				throw stack_bad_alloc("stack_allocator: ran out of pool space");
			}
			return reinterpret_cast<T*>(m_pHead->next);
		}
		else
			throw stack_bad_alloc("stack_allocator: allocating from deallocated allocator (don't dynamically allocate allocators)");
	}
	inline void deallocate(pointer p, size_type n) throw()
	{
		if( p && m_pHead )
		{
			assert( m_pHead->next == reinterpret_cast<char*>(p) );
			m_pHead->next += sizeof(T)*n;
			assert(!( m_pHead->next > m_pHead->first ));
		}
	}
	inline size_type max_size() const throw()
	{
		if( m_pHead )
			return (m_pHead->next-reinterpret_cast<char*>(m_pHead->last))/sizeof(T);
		else
			return 0;
	}
	inline void construct(pointer p, const T& val)
		{ assert( m_pHead->last <= p && p < m_pHead->first ); new(p) T(val); }
	inline void destroy(pointer p)
		{ assert( m_pHead->last <= p && p < m_pHead->first ); p->~T(); (void)p; }

protected:
	stack_allocator() throw() { }
};


#ifdef _MSC_VER
#undef inline
#endif // _MSC_VER
#ifdef __GNUC__
#undef inline
#endif // __GNUC__

#endif // __STACK_ALLOCATOR_H_INCLUDED_
