/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __SORTGEN_H_INCLUDED_
#define __SORTGEN_H_INCLUDED_

#include <ctime>
#include <cstdlib>
#include <cmath>
#include "../Bench/int64.h"
#include "../Bench/aarray.h"
#include "../Bench/elements.h"
#include "random.h"

#include <functional>
#include <iterator>
#include <vector>
#include <algorithm>


template<class It, class Pred=std::less<typename std::iterator_traits<It>::value_type> >
class killer : public std::binary_function<It,It,bool>
{
	typedef typename std::iterator_traits<It>::value_type T;
	typedef typename std::iterator_traits<It>::difference_type Diff;
public:
	typedef typename std::vector<Diff>::iterator iterator;
	class predicate
	{
	public:
		predicate(killer *k) : host(k) { }
		bool operator()(const Diff& lhs, const Diff& rhs)
		{
			if( host->is_gas(lhs) && host->is_gas(rhs) )
			{
				if( lhs == host->candidate )
				{
					host->freeze(lhs);
					return true;
				}
				else
				{
					host->freeze(rhs);
					return false;
				}
			}
			else
				if( host->is_gas(lhs) )
				{
					host->candidate = lhs;
					return false;
				}
				else if( host->is_gas(rhs) )
				{
					host->candidate = rhs;
					return true;
				}
				else
					return host->compare(lhs,rhs);
		}
	private:
		killer *host;
	};

	killer(It b, It e, Pred p=Pred()) : its(e-b), solid(), first(b), last(e), candidate(e-b), comp(p)
	{
		T gas = T();
		Diff d = 0;
		for( typename std::vector<Diff>::iterator s=its.begin(); b!=e; ++d, ++s, ++b )
			*s = d, *b = gas;
	}
	~killer()
	{
		for( It i=first; i!=last; ++i ) // Find the one, that possibly didn't get fixed
			if( *i == T() )
			{
				factory<T>::increment(solid);
				*i = solid;
			}
	}

	inline iterator begin()
		{ return its.begin(); }
	inline iterator end()
		{ return its.end(); }
	inline predicate pred()
		{ return predicate(this); }

private:
	inline void freeze(Diff i)
	{
		factory<T>::increment(solid);
		*(first+i) = solid;
	}
	inline bool is_gas(Diff i)
		{ return *(first+i) == T(); }
	inline bool compare(Diff lhs, Diff rhs)
		{ return comp(*(first+lhs),*(first+rhs)); }

	typename std::vector<Diff> its;
	typename std::iterator_traits<It>::value_type solid;
	It first, last;
	Diff candidate;
	Pred comp;
};

template<class T, class It>
void mmerge(It b, It e, size_t k)
{
	random_source rs;
	size_t run_size = (b-e)/k;
	size_t rem = (b-e)%k;
	run_size += rem/k;
	size_t big_runs = rem%k;
	if( big_runs )
		++run_size;
	for( It i=b; b!=e; ++b )
	{
		rs >> *b;
		if( static_cast<size_t>(i-b) == run_size )
		{
			std::sort(b,i);
			if( --big_runs == 0 )
				--run_size;
			b = i;
		}
	}
}

template<class T, class It>
void distr(It b, It e, size_t k)
{
	random_source rs;
	factory<T> f;
	for( double d=0.0, dd=static_cast<double>(e-b)/k; b!=e; d+=dd )
		for( size_t j=k; j && b!=e; ++b, --j )
		{
			double r;
			rs >> r;
			*b = f.from_scalar(r*dd+d);
		}
}

template<class T, class It>
void disti(It b, It e, size_t k)
{
	random_source rs;
	T *elts = new T[k];
	for( T *p=elts; p!=elts+k; ++p )
		rs >> *p;
	for( ; b!=e; ++b )
	{
		size_t r;
		rs >> r;
		*b = elts[r%k];
	}
}

template<class T>
void sortgen(const char *filename, uint64_t size, char d)
{
	auto_array::map_array<T> A(filename, auto_array::MAP_CREATE | auto_array::MAP_WRITE | auto_array::MAP_STREAMING, size);
	if( d == 'k' )
	{
		killer<typename auto_array::map_array<T>::iterator> k(A.begin(), A.end());
		std::sort(k.begin(), k.end(), k.pred());
	}
	else
	{
		random_source rs;
		switch( d )
		{
		case 'u':
			for( uint64_t i=0; i!=size; ++i )
				rs >> A[i];
			break;
		case 'm':
			mmerge<T>(A.begin(), A.end(),static_cast<size_t>(log(static_cast<float>(A.end()-A.begin()))));
			break;
		case 'M':
			mmerge<T>(A.begin(), A.end(),static_cast<size_t>(sqrt(static_cast<float>(A.end()-A.begin()))));
			break;
		case 'd':
			distr<T>(A.begin(), A.end(),static_cast<size_t>(log(static_cast<float>(A.end()-A.begin()))));
			break;
		case 'D':
			distr<T>(A.begin(), A.end(),static_cast<size_t>(sqrt(static_cast<float>(A.end()-A.begin()))));
			break;
		case 'e':
			disti<T>(A.begin(), A.end(),static_cast<size_t>(log(static_cast<float>(A.end()-A.begin()))));
			break;
		case 'R':
			disti<T>(A.begin(), A.end(),static_cast<size_t>(sqrt(static_cast<float>(A.end()-A.begin()))));
			break;
		default:
			throw std::invalid_argument("Unknown distribution");
		}
	}
}

#endif // __SORTGEN_H_INCLUDED_
