/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __STOPWATCH_H_INCLUDED_
#define __STOPWATCH_H_INCLUDED_

#ifdef WIN32
#include "../ansiwin.h"
#else // WIN32
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <unistd.h>
#endif // WIN32
#include <time.h>
#include <iostream.h>
#include <iomanip.h>
#include "int64.h"


class sys_names
{
public:
	static const char *compiler();
	static const char *os();
	static const char *release();
	static const char *host();
	static const char *buildtime()
		{ return __DATE__ " " __TIME__; }

private:
#ifdef WIN32
	static class utsn
	{
	public:
		utsn();
		char version[6];
	} name;
#else // WIN32
	static class utsn
	{
	public:
		utsn() { uname(&name); }
		struct utsname name;
	} name;
#endif // WIN32
};

class local_time
{
	friend ostream& operator<<(ostream& s, local_time t);
public:
	local_time()
	{
		time_t aclock;
		time(&aclock);
		ltime = localtime(&aclock);
	};
private:
	struct tm *ltime;
};

class stopwatch
{
public:
	class xt
	{
		friend ostream& operator<<(ostream&,xt);
	public:
		xt(double t, int i, char ch) : time(t), indent(i), ichar(ch) {}
	private:
		double time;
		int indent;
		char ichar;
	};

	stopwatch();

#ifdef WIN32
	inline void start()
		{ QueryPerformanceCounter(&begin); }
	inline void stop()
	{
		LARGE_INTEGER end;
		QueryPerformanceCounter(&end);
		if( is_running() )
			do_stop(end);
	}
	bool is_running() const
		{ return begin.u.LowPart != 0 || begin.u.HighPart != 0; }
#else // WIN32
	inline void start()
		{ gettimeofday(&begin, NULL); }
	inline void stop()
	{
		struct timeval end;
		gettimeofday(&end, NULL);
		if( is_running() )
			do_stop(end);
	}
	bool is_running() const
		{ return begin.tv_sec != 0 || begin.tv_usec != 0; }
#endif // WIN32

	inline void reset()
	{
		m_time = 0.0;
		if( is_running() )
			start();
	}
	double time() const
		{ return m_time; }
	xt xmltag(int i=0, char ch='\t') const
		{ return xt(m_time,i,ch); }

private:
#ifdef WIN32
	void do_stop(LARGE_INTEGER& end);
	LARGE_INTEGER perf_freq, begin;
	int perf_freq_mag;
#else // WIN32
	void do_stop(struct timeval& end);
	struct timeval begin;
#endif // WIN32
	long overhead_tick;
	double m_time;
};

class pf_counter
{
public:
	class xt
	{
		friend ostream& operator<<(ostream&,xt);
	public:
		xt(unsigned long c, int i, char ch) : count(c), indent(i), ichar(ch) {}
	private:
		unsigned long count;
		int indent;
		char ichar;
	};

	pf_counter();
#ifdef WIN32
	~pf_counter();
#endif // WIN32

	void start();
	void stop();
	void reset();
	inline bool is_running() const
		{ return begin != 0; }
	inline unsigned long count() const
		{ return m_count; }
	inline xt xmltag(int i=0, char ch='\t') const
		{ return xt(m_count,i,ch); }

private:
	unsigned long begin;
	unsigned long m_count;
#ifdef WIN32
	HANDLE hJob;
#endif // WIN32
};

#ifdef PAPI

#include <papiStdEventDefs.h>
#include <papi.h>

class miss_counter
{
public:
	class xt
	{
		friend ostream& operator<<(ostream&,xt);
	public:
		xt(int64_t l, int64_t t, int i=0, char ch='\t') : l2(l), tlb(t), indent(i), ichar(ch) {}
	private:
		int64_t l2;
		int64_t tlb;
		int indent;
		char ichar;
	};

	miss_counter();
	~miss_counter();

	void start();
	void stop();
	void reset();

	inline bool is_running() const
		{ return true; }
	inline uint64_t l2() const
	{
		long_long counters[2];
		if( PAPI_read(EventSet, counters) != PAPI_OK )
			throw exception();//"PAPI_stop");
		return counters[0];
	}
	inline uint64_t tlb() const
	{
		long_long counters[2];
		if( PAPI_read(EventSet, counters) != PAPI_OK )
			throw exception();//"PAPI_stop");
		return counters[1];
	}
	inline xt xmltag(int i=0, char ch='\t') const
	{
		long_long counters[2];
		if( PAPI_read(EventSet, counters) != PAPI_OK )
			throw exception();//"PAPI_stop");
		return xt(counters[0], counters[1], i, ch);
	}

private:
	static class initializer
	{
	public:
		initializer();
	} init;

	int EventSet;
};

#else // PAPI

class miss_counter
{
public:
	class xt
	{
		friend ostream& operator<<(ostream&,xt);
	};

	miss_counter() {};
	void start() {};
	void stop() {};
	void restart() {};

	inline bool is_running() const
		{ return false; }
	inline uint64_t tlb() const
		{ return 0; }
	inline uint64_t l2() const
		{ return 0; }
	inline xt xmltag(int i=0, char ch='\t') const
		{ return xt(); }

private:
	static class initializer
	{
	public:
		initializer();
	} init;
};

#endif // PAPI


#endif // __STOPWATCH_H_INCLUDED_
