/*
 * Copyright (C) 2003 Kristoffer Vinther.
 * For contact information, please visit
 * http://www.brics.dk/~kv/thesis/.
 * If you use this code in commercial applications, please notify me.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef _MSC_VER
#pragma warning(disable: 4267)
#pragma inline_depth(18)
#endif // _MSC_VER

#ifdef WIN32
#ifdef _DEBUG
//#include <afxwin.h> // Enable memoryleak detection
#endif // _DEBUG
#endif // WIN32

#define MAPPINGSTAT

#ifdef _DEBUG
#define _CUBE_BUFFERS
#endif // _DEBUG

#include "FArray.h"
#include "cmd.h"
#include "Benchmark.h"

#include "Sort/Sort.h"
#include "Sort/PointerBMT.h"
#include "Sort/ImplicitBMT.h"
#include "Sort/LazyMerger.h"
#include "Sort/HeapMerger.h"

#include "Sort/funnel/stack_allocator.h"
#include "Sort/funnel/merge_tree.h"
#include "Sort/funnel/merge_tree_types.h"

using namespace auto_array;
using namespace iosort;

template<class T>
class CArrayAllocator
{
public:
	typedef T* array_type;
	typedef T* it_type;
	template<class ItDiff>
	static T* allocate(ItDiff size)
	{
		array = new T[size];
		return array;
	}
	static inline void deallocate()
		{ delete[] array; }
private:
	static T* array;
};
template<class Array>
class AutoArrayAllocator
{
public:
	typedef Array array_type;
	typedef typename Array::iterator it_type;
	template<class ItDiff>
	static it_type allocate(ItDiff size)
	{
		Array n(size);
		array = n;
		return array.begin();
	}
	static inline void deallocate() { }
private:
	static Array array;
};
template<class Array>
class MapArrayAllocator
{
public:
	typedef Array array_type;
	typedef typename Array::iterator it_type;
//	typedef typename Array::stream_iterator it_type;
	template<class ItDiff>
	static it_type allocate(ItDiff size)
	{
		array = array_type(DST_FILE, MAP_CREATE | MAP_WRITE | MAP_STREAMING, size);
		return array.begin();
//		return array.begin_stream();
	}
	static inline void deallocate() { }
private:
	static array_type array;
};

template<class T>
T* CArrayAllocator<T>::array = NULL;
template<class Array>
Array AutoArrayAllocator<Array>::array;
template<class Array>
Array MapArrayAllocator<Array>::array;


template<class T, class It>
T check_sort(It begin, It end)
{
#ifdef WIN32 // We only check the output on the development platform, not on benchmarking platforms
	if( begin == end )
		return T();
	T mysum = *begin;
	T last = *begin;
	for( It p=begin+1; p!=end; p++ )
	{
		if( *p < last )
		{
			std::cerr << "Elements " << (int)(p-begin-1) << " and " << (int)(p-begin) << " out of order: " << last << '<' << *p << std::endl;
			return *begin;
		}
		mysum += *p;
		last = *p;
	}
	return mysum;
#else // WIN32
	return *begin;
#endif // WIN32
}

template<class T, class It>
void check_sort(It begin, It end, T chksum)
{
#ifdef WIN32 // We only check the output on the development platform, not on benchmarking platforms
	T mysum = check_sort<T, It>(begin,end);
	if( mysum != chksum )
		std::cerr << "Checksum mismatch of " << (mysum<chksum? chksum-mysum : mysum-chksum) << std::endl;
#endif // Win32
}

using namespace cmd_switch;

template<class T, class Alloc, class It>
void RunAlg(algorithm alg, It begin, It end)
{
	typename Alloc::it_type dest;
	double time = 0.0;
	T chksum = RandArray::InitArray<It, T>(begin, end);
	BeginPFCounter();
#ifdef PAPI
	StartMisses();
#endif // PAPI
	switch( alg )
    {
	case q:
		std::cout << "\t\t<algorithm>qsort</algorithm>" << std::endl;
		StartTimer();
		gen_qsort(begin, end, *begin);
		time = StopTimer();
		check_sort(begin, end, chksum);
		break;
	case s:
		std::cout << "\t\t<algorithm>std::sort</algorithm>" << std::endl;
		StartTimer();
		std::sort(begin, end);
		time = StopTimer();
		check_sort(begin, end, chksum);
		break;
	case h:
		StartTimer();
		std::cout << "\t\t<algorithm>std::sort_heap</algorithm>" << std::endl;
		std::make_heap(begin, end);
		std::sort_heap(begin, end);
		time = StopTimer();
		check_sort(begin, end, chksum);
		break;
	case mg:
		std::cout << "\t\t<algorithm>mergesort</algorithm>" << std::endl;
		dest = Alloc::allocate(end-begin);
		StartTimer();
		mergesort(begin, end, dest);
		time = StopTimer();
		check_sort(dest, dest+(end-begin), chksum);
		break;
	case img:
		std::cout << "\t\t<algorithm>imergesort</algorithm>" << std::endl;
		StartTimer();
		inplace_mergesort(begin, end);
		time = StopTimer();
		check_sort(begin, end, chksum);
		break;
	case Mg:
		std::cout << "\t\t<algorithm>CacheMergesort</algorithm>" << std::endl;
		dest = Alloc::allocate(end-begin);
		StartTimer();
		MergeSort<HeapMerger<T,    240000,    128, InputStream<It> > >(begin, end, dest);
		time = StopTimer();
		check_sort(dest, dest+(end-begin), chksum);
		break;
	case MMg:
		std::cout << "\t\t<algorithm>MemMergesort</algorithm>" << std::endl;
		dest = Alloc::allocate(end-begin);
		StartTimer();
		MergeSort<HeapMerger<T, 214748364, 131072, InputStream<It> > >(begin, end, dest);
		time = StopTimer();
		check_sort(dest, dest+(end-begin), chksum);
		break;
	case MIg:
		std::cout << "\t\t<algorithm>InplaceCacheMergesort</algorithm>" << std::endl;
		StartTimer();
//		InplaceMergeSort<HeapMerger<T,    240000,    128, InputStream<It> > >(begin, end);
		time = StopTimer();
		check_sort(begin, end, chksum);
		break;
	case MMIg:
		std::cout << "\t\t<algorithm>InplaceMemMergesort</algorithm>" << std::endl;
		StartTimer();
//		InplaceMergeSort<HeapMerger<T, 214748364, 131072, InputStream<It> > >(begin, end);
		inplace_merge_sort<HeapMerger<T, 214748364, 131072, InputStream<It>, stack_allocator<T> >,mem_splitter<T,InputStream<It> > >(begin, end);
		time = StopTimer();
		check_sort(begin, end, chksum);
		break;
	case Q:
		std::cout << "\t\t<algorithm>QuickSort</algorithm>" << std::endl;
		StartTimer();
//		QuickSort<T>(begin, end-begin);
		time = StopTimer();
//		check_sort(begin, end, chksum);
		break;
	case FE:
		std::cout << "\t\t<algorithm>FunnelSort[EL]</algorithm>" << std::endl;
		dest = Alloc::allocate(end-begin);
		StartTimer();
		MergeSort<LazyMerger<PBBinMergeTree<T>, InputStream<It> > >(begin, end, dest);
		time = StopTimer();
		check_sort(dest, dest+(end-begin), chksum);
		break;
	case FI:
		std::cout << "\t\t<algorithm>FunnelSort[IL]</algorithm>" << std::endl;
		dest = Alloc::allocate(end-begin);
		StartTimer();
//		MergeSort<LazyMerger<ImplBinMergeTree<T>, InputStream<It> > >(begin, end, dest);
		merge_sort<typename funnel::merge_tree_types<T,4,funnel::buffer<It> >::rec_stack_mveb, funnel::splitter<4> >(begin, end, dest);
		time = StopTimer();
		check_sort(dest, dest+(end-begin), chksum);
		break;
	case FEI:
		std::cout << "\t\t<algorithm>FunnelSort[EIL]</algorithm>" << std::endl;
		StartTimer();
		InplaceMergeSort<LazyMerger<PBBinMergeTree<T>, InputStream<It> > >(begin, end);
		time = StopTimer();
		check_sort(begin, end, chksum);
		break;
	case FII:
		std::cout << "\t\t<algorithm>FunnelSort[IIL]</algorithm>" << std::endl;
		StartTimer();
		inplace_merge_sort<typename funnel::merge_tree_types<T,4,funnel::buffer<It> >::rec_stack_mveb, funnel::splitter<4> >(begin, end);
		time = StopTimer();
		check_sort(begin, end, chksum);
		break;
	}
	unsigned long pf = EndPFCounter();
#ifdef PAPI
	std::pair<_int64_t,_int64_t> msss = StopMisses();
	std::cout	<< "\t\t<l2misses>" << msss.first << "</l2misses>" << std::endl
				<< "\t\t<tlbmisses>" << msss.second << "</tlbmisses>" << std::endl;
#endif // PAPI
	Alloc::deallocate();
	std::cout	<< "\t\t<time unit=\"s\">" << time << "</time>" << std::endl 
				<< "\t\t<pagefaults>" << pf << "</pagefaults>" << std::endl;
}

template<class T>
void Run(algorithm alg, array_type arr, _uint64_t size64, const char *file)
{
	size_t size = (size_t)size64;
	std::cout << "\t\t<size>" << size64 << "</size>" << std::endl;
	std::cout << "\t\t<elementsize>" << sizeof(T) << "</elementsize>" << std::endl;
	switch( arr )
	{
	case c:
		{
			std::cout << "\t\t<array>c</array>" << std::endl;
			T *A = new T[size];
			RunAlg<T, CArrayAllocator<T> >(alg, A, A+size);
		}
		break;
	case d:
		{
			std::cout << "\t\t<array>DynArray</array>" << std::endl;
			DynArray<T> A = DynArray<T>(size);
//			RunAlg<T, AutoArrayAllocator<DynArray<T> > >(alg, A.begin(), A.end(size));
		}
		break;
	case m:
		{
			std::cout << "\t\t<array>MapArray</array>" << std::endl;
			MapArray<T> A = MapArray<T>(file, MAP_CREATE | MAP_READ | MAP_WRITE, size);
			RunAlg<T, MapArrayAllocator<MapArray<T> > >(alg, A.begin(), A.end(size));
		}
		break;
	case M:
		{
			std::cout << "\t\t<array>MultiMapArray</array>" << std::endl;
			MultiMapArray<T> A = MultiMapArray<T>(file, MAP_CREATE | MAP_READ | MAP_WRITE, size64);
//			RunAlg<T, MapArrayAllocator<MultiMapArray<T> > >(alg, A.begin(), A.end(size));
			std::cout << "\t\t<remaps>" << MultiMapArray<T>::PageFaults << "</remaps>" << std::endl;
		}
		break;
	case S:
		{
			std::cout << "\t\t<array>SeqMapArray</array>" << std::endl;
			SeqMapArray<T> A = SeqMapArray<T>(file, MAP_CREATE | MAP_READ | MAP_WRITE, size64);
//			RunAlg<T, MapArrayAllocator<SeqMapArray<T> > >(alg, A.begin(), A.end(size));
		}
		break;
	}
}


class Element
{
	friend std::ostream& operator<<(std::ostream& os, const Element& e);
public:
	inline Element& operator=(unsigned long int k)
		{ key = k; return *this; }
	inline bool operator<(const Element& rhs) const
		{ return key<rhs.key; }
	inline bool operator>(const Element& rhs) const
		{ return key>rhs.key; }

	inline Element operator+=(const Element& rhs)
		{ key+=rhs.key; return *this; }
	inline bool operator!=(const Element& rhs) const
		{ return key!=rhs.key; }
	inline bool operator==(const Element& rhs) const
		{ return key==rhs.key; }
	inline int operator-(const Element& rhs) const
		{ return (int)(rhs.key-key); }
private:
	unsigned long int key;
	void* ptr;
};
std::ostream& operator<<(std::ostream& os, const Element& e)
{
	return os << '(' << e.key << ',' << e.ptr << ')';
}

int main(int argc, char *argv[])
{
	algorithm alg;
	array_type arr;
	_uint64_t size64;

	try
	{
		parseargs(argc, argv, alg, arr, size64);
		std::cout << "\t<run buildtime=\"" << buildtime() << "\" compiler=\"" << compilername() << "\" commandline=\"";
		std::cout << argv[0];
		for( int i=1; i<argc; i++ )
			std::cout << ' ' << argv[i];
		std::cout << "\" execstatus=$>" << std::endl;
		if( sizeof(void*)<8 && size64>0xFFFFFFFF && arr!=M && arr!=S )
		{
			std::cerr << "Error: Size not supported by this type of array. Use -aM or -aS." << std::endl;
			return 1;
		}
		if( alg == q && (arr==M || arr==S) )
		{
			std::cerr << "Error: qsort() doesn't support this type of array." << std::endl;
			return 1;
		}

//		Counting<long int>::reset();
//		Run<Counting<long int> >(alg, arr, size64, SRC_FILE);
		Run<Element>(alg, arr, size64, SRC_FILE);
//		std::cout	<< "\t\t<comparisons>" << Counting<long int>::getComps() << "</comparisons>" << std::endl 
//				<< "\t\t<moves>" << Counting<long int>::getComps() << "</moves>" << std::endl;

		std::cout << "\t</run>" << std::endl;
	}
	catch( std::exception& ex )
	{
		std::cerr << "Caught std::exception: " << ex.what() << std::endl;
		return 1;
	}
/*	catch( ... )
	{
		std::cerr << "Caught exception in main()" << std::endl;
		return 1;
	}
*/	return 0;
}









using namespace funnel;
using namespace iosort::base;



//#ifdef NDEBUG
template<class T>
class SortedOutputR : public std::iterator<std::random_access_iterator_tag, T, int>
{
public:
	SortedOutputR() : cnt(0) { }
	SortedOutputR(size_t count) : cnt(count) { }
	inline SortedOutputR& operator++()
	{
		if( cnt )
		{
			if( prev > current )
				std::cerr << "Elements " << (int)cnt << " and " << (int)(cnt+1) << " out of order: " << prev << '>' << current << std::endl;
			prev = current;
			chksum += current;
		}
		else
		{
			prev = current;
			chksum = current;
		}
		++cnt;
		return *this;
	}
	inline T& operator*()
		{ return current; }
	inline const T& operator*() const
		{ return current; }
	inline bool operator==(const SortedOutputR& rhs) const
		{ return cnt == rhs.cnt; }
	inline bool operator!=(const SortedOutputR& rhs) const
		{ return cnt != rhs.cnt; }
	inline int operator-(const SortedOutputR& rhs) const
		{ return cnt - rhs.cnt; }
	inline T checksum() const
		{ return chksum; }
	inline size_t count() const
		{ return cnt; }
protected:
	T current;
	T prev;
	T chksum;
	size_t cnt;
};

template<class Merger>
void check_heap_merger(funnel::order_t k, const char* v, bool balanced = false)
{
	std::cout << "\t<run buildtime=\"" << buildtime() << "\" compiler=\"" << compilername() << "\" execstatus=$>" << std::endl;
	std::cout << "\t\t<size>" << k << "</size>" << std::endl;
	std::cout << "\t\t<algorithm>" << v << "</algorithm>" << std::endl;
	typedef typename Merger::value_type T;
	typedef typename Merger::stream stream;
	try
	{
		std::cout << "\t\t<time unit=\"s\">" << std::flush; 
		funnel::order_t bk = k;
		if( balanced )
			bk = funnel::pow_of_order<2>(funnel::logc<2>(k));
		std::allocator<T> alloc;
		Merger merge_tree(bk,alloc);
		T *input = new T[k*k*k];
		T chksum = RandArray::InitArray<T*,T>(input,input+k*k*k);
		for( T *p=input; p+k*k<=input+k*k*k; p+=k*k)
			std::sort(p, p+k*k);
		StartTimer();
		for( int i=0; i<20000000/(k*k*k); ++i )
		{
			for( T *p=input; p+k*k<=input+k*k*k; p+=k*k)
				merge_tree.add_stream(stream(p,p+k*k));
			merge_tree(SortedOutputR<T>(), SortedOutputR<T>(k*k*k));
			merge_tree.reset();
		}
		double t = StopTimer();
		delete[] input;
		std::cout << t << "</time>" << std::endl;
	}
	catch( std::exception& ex )
	{
		std::cerr << "Caught std::exception: " << ex.what() << std::endl;
	}
	std::cout << "\t</run>" << std::endl;
}

template<class Merger>
void check_stack_merger(funnel::order_t k, const char* v, bool balanced = false)
{
	std::cout << "\t<run buildtime=\"" << buildtime() << "\" compiler=\"" << compilername() << "\" execstatus=$>" << std::endl;
	std::cout << "\t\t<size>" << k << "</size>" << std::endl;
	std::cout << "\t\t<algorithm>" << v << "</algorithm>" << std::endl;
	typedef typename Merger::value_type T;
	typedef typename Merger::stream stream;
	try
	{
		std::cout << "\t\t<time unit=\"s\">" << std::flush; 
		funnel::order_t bk = k;
		if( balanced )
			bk = funnel::pow_of_order<4>(funnel::logc<4>(k));
		stack_allocator<T> alloc(Merger::size_of(bk));
		Merger merge_tree(bk,alloc);
		T *input = new T[k*k*k];
		T chksum = RandArray::InitArray<T*,T>(input,input+k*k*k);
		for( T *p=input; p+k*k<=input+k*k*k; p+=k*k)
			std::sort(p, p+k*k);
		StartTimer();
		for( int i=0; i<20000000/(k*k*k); ++i )
		{
			for( T *p=input; p+k*k<=input+k*k*k; p+=k*k)
				merge_tree.add_stream(stream(p,p+k*k));
			merge_tree(SortedOutputR<T>(), SortedOutputR<T>(k*k*k));
			merge_tree.reset();
		}
		double t = StopTimer();
		delete[] input;
		std::cout << t << "</time>" << std::endl;
	}
	catch( std::exception& ex )
	{
		std::cerr << "Caught std::exception: " << ex.what() << std::endl;
	}
	std::cout << "\t</run>" << std::endl;
}

template<int order>
void benchMergers(funnel::order_t k)
{
	check_heap_merger <typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::rec_heap_mveb>(k,"rec_heap_mveb");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::rec_stack_mveb>(k,"rec_stack_mveb");
	check_heap_merger <typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::rec_heap_veb>(k,"rec_heap_veb");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::rec_stack_veb>(k,"rec_stack_veb");
	check_heap_merger <typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::rec_heap_mdf>(k,"rec_heap_mdf");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::rec_stack_mdf>(k,"rec_stack_mdf");
	check_heap_merger <typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::rec_heap_df>(k,"rec_heap_df");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::rec_stack_df>(k,"rec_stack_df");
	check_heap_merger <typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::rec_heap_mbf>(k,"rec_heap_mbf");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::rec_stack_mbf>(k,"rec_stack_mbf");
	check_heap_merger <typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::rec_heap_bf>(k,"rec_heap_bf");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::rec_stack_bf>(k,"rec_stack_bf");
	check_heap_merger <typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::pb_heap_mveb>(k,"pb_heap_mveb");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::pb_stack_mveb>(k,"pb_stack_mveb");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::impl_mveb>(k,"impl_mveb",true);
	check_heap_merger <typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::pb_heap_veb>(k,"pb_heap_veb");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::pb_stack_veb>(k,"pb_stack_veb");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::impl_veb>(k,"impl_veb");
	check_heap_merger <typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::pb_heap_mdf>(k,"pb_heap_mdf");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::pb_stack_mdf>(k,"pb_stack_mdf");
	check_heap_merger <typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::pb_heap_df>(k,"pb_heap_df");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::pb_stack_df>(k,"pb_stack_df");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::impl_df>(k,"impl_df");
	check_heap_merger <typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::pb_heap_mbf>(k,"pb_heap_mbf");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::pb_stack_mbf>(k,"pb_stack_mbf");
	check_heap_merger <typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::pb_heap_bf>(k,"pb_heap_bf");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element> >::pb_stack_bf>(k,"pb_stack_bf");
	check_stack_merger<typename merge_tree_types<Element,order,buffer<Element*>,two_merger<2,buffer<Element*>,Element>	>::impl_bf>(k,"impl_bf");
	check_heap_merger <LazyMerger<PBBinMergeTree<Element>, InputStream<Element*> > >(k,"pb_lazy",true);
	check_heap_merger <LazyMerger<ImplBinMergeTree<Element>, InputStream<Element*> > >(k,"impl_lazy",true);
	check_heap_merger <HeapMerger<Element, 20000000, 3500, InputStream<Element*> > > (k,"bin_heap");
}
//#endif // NDEBUG

void merge_test()
{
	{
		stack_allocator<Element> alloc(funnel::merge_tree_types<Element,2,buffer<Element*> >::impl_mveb::size_of(64));
		funnel::merge_tree_types<Element,2,buffer<Element*> >::impl_mveb merge_tree(64,alloc);
//		funnel::merge_tree_types<Element,2,buffer<Element*> >::pb_heap_veb merge_tree(41);
	}
	{
		stack_allocator<Element> alloc(funnel::merge_tree_types<Element,2,buffer<Element*> >::pb_stack_mveb::size_of(41));
		funnel::merge_tree_types<Element,2,buffer<Element*> >::pb_stack_mveb merge_tree(41,alloc);
//		funnel::merge_tree_types<Element,2,buffer<Element*> >::pb_heap_veb merge_tree(41);
	}
	{
		stack_allocator<Element> alloc(funnel::merge_tree_types<Element,3,buffer<Element*> >::pb_stack_mveb::size_of(16));
		funnel::merge_tree_types<Element,3,buffer<Element*> >::pb_stack_mveb merge_tree(16,alloc);
//		funnel::merge_tree_types<Element,3,buffer<Element*> >::pb_heap_veb merge_tree(16);
	}
#ifdef NDEBUG
	for( funnel::order_t k=15; k<=271; k+=5 )
		benchMergers<2>(k);
//		benchMergers<2>(256);
#endif // NDEBUG

/*	checkMerger<merge_tree_types<Element,2,buffer<Element*> >::pb_stack_mveb>(64, "2-pb_stack_mveb");
	checkMerger<merge_tree_types<Element,2,buffer<Element*> >::pb_stack_mveb>(41, "2-pb_stack_mveb");
	checkMerger<merge_tree_types<Element,3,buffer<Element*> >::pb_stack_mveb>(64, "3-pb_stack_mveb");
	checkMerger<merge_tree_types<Element,6,buffer<Element*> >::pb_stack_mveb>(150, "6-pb_stack_mveb");
	checkMerger<merge_tree_types<Element,4,buffer<Element*> >::pb_stack_mveb>(40, "4-pb_stack_mveb");
	checkMerger<merge_tree_types<Element,4,buffer<Element*> >::pb_stack_mveb>(12, "4-pb_stack_mveb");
	checkMerger<merge_tree_types<Element,4,buffer<Element*> >::pb_stack_mveb>(5, "4-pb_stack_mveb");
	checkMerger<merge_tree<Element, buffer<Element*>, mixed_veb_layout<pointer_navigator<pointer_node<8,Element*> >, Element, stack_allocator<Element> >, loser_merger<8,buffer<Element*>,int> > >(128, "8-veb_looser");
	checkMerger<merge_tree<Element, buffer<Element*>, mixed_veb_layout<pointer_navigator<pointer_node<7,Element*> >, Element, stack_allocator<Element> >, loser_merger<7,buffer<Element*>,int> > >(128, "7-veb_looser");
	checkMerger<merge_tree<Element, buffer<Element*>, mixed_veb_layout<pointer_navigator<pointer_node<6,Element*> >, Element, stack_allocator<Element> >, loser_merger<6,buffer<Element*>,int> > >(128, "6-veb_looser");
	checkMerger<merge_tree<Element, buffer<Element*>, mixed_veb_layout<pointer_navigator<pointer_node<5,Element*> >, Element, stack_allocator<Element> >, loser_merger<5,buffer<Element*>,int> > >(128, "5-veb_looser");
	checkMerger<merge_tree<Element, buffer<Element*>, mixed_veb_layout<pointer_navigator<pointer_node<2,Element*> >, Element, stack_allocator<Element> >, loser_merger<2,buffer<Element*>,int> > >(128, "2-veb_looser");
	checkMerger<merge_tree<Element, buffer<Element*>, mixed_veb_layout<pointer_navigator<pointer_node<3,Element*> >, Element, stack_allocator<Element> >, loser_merger<3,buffer<Element*>,int> > >(64, "3-veb_looser");
	checkMerger<merge_tree<Element, buffer<Element*>, mixed_veb_layout<pointer_navigator<pointer_node<2,Element*> >, Element, stack_allocator<Element> >, for_merger<2,buffer<Element*>,int> > >(128, "2-mveb_for");
	checkMerger<merge_tree<Element, buffer<Element*>, mixed_veb_layout<pointer_navigator<pointer_node<5,Element*> >, Element, stack_allocator<Element> >, for_merger<5,buffer<Element*>,int> > >(128, "5-mveb_for");
	checkMerger<merge_tree<Element, buffer<Element*>, mixed_veb_layout<pointer_navigator<pointer_node<2,Element*> >, Element, stack_allocator<Element> >, for_val_merger<2,buffer<Element*>,int> > >(128, "2-mveb_for_val");
	checkMerger<merge_tree<Element, buffer<Element*>, mixed_veb_layout<pointer_navigator<pointer_node<5,Element*> >, Element, stack_allocator<Element> >, for_val_merger<5,buffer<Element*>,int> > >(128, "5-mveb_for_val");
	checkMerger<merge_tree<Element, buffer<Element*>, mixed_veb_layout<pointer_navigator<pointer_node<2,Element*> >, Element, stack_allocator<Element> >, two_merger<2,buffer<Element*>,int> > >(128, "2-mveb_two");
//	checkMerger<merge_tree<Element, buffer<Element*>, mixed_veb_layout<pointer_navigator<pointer_node<5,Element*> >, Element, stack_allocator<Element> >, comp_merger<5,Element*,int> > >(128, "5-mveb_comp");
//	checkMerger<merge_tree<Element, buffer<Element*>, mixed_veb_layout<pointer_navigator<pointer_node<2,Element*> >, Element, stack_allocator<Element> >, comp_merger<2,Element*,int> > >(128, "2-mveb_comp");
	checkMerger<LazyMerger<PBBinMergeTree<Element>,InputStream<Element*>,stack_allocator<Element> > >(128, "LazyMerger");
*/	const size_t SIZE = 22665187;
	const size_t SUBSIZE = 80089;
//	const size_t SIZE = 2920775;
//	const size_t SUBSIZE = 20425;

	Element* input = new Element[SIZE];
	Element* output = new Element[SIZE];

	double T1 = 0, T2 = 0;
	for( int i=0; i<3; ++i )
//	for( ;; )
	{
		{
			stack_allocator<Element> alloc(funnel::merge_tree_types<Element,3,buffer<Element*> >::pb_stack_mveb::size_of((SIZE+SUBSIZE-1)/SUBSIZE));
			funnel::merge_tree_types<Element,3,buffer<Element*> >::pb_stack_mveb merge_tree((SIZE+SUBSIZE-1)/SUBSIZE,alloc);
			Element chksum = RandArray::InitArray<Element*,Element>(input,input+SIZE);
			Element *p = input;
			for( ; p+SUBSIZE<=input+SIZE; p+=SUBSIZE )
			{
				std::sort(p, p+SUBSIZE);
				merge_tree.add_stream(buffer<Element*>(p,p+SUBSIZE));
			}
			std::cerr << "Running pb_veb...";
			StartTimer();
			Element *e = merge_tree(output, output+(p-input));
			T1 += StopTimer();
			std::cerr << std::endl << "Checking pb_veb... ";
			check_sort<Element>(output, output+(p-input), chksum);
			std::cerr << std::endl;
			assert( e == output+(p-input) ); (void)e;
		}
		{
			LazyMerger<PBBinMergeTree<Element>,InputStream<Element*> > merge_tree(512);
//			LazyMerger<PBBinMergeTree<Element>,InputStream<Element*> > merge_tree(256);
			Element chksum = RandArray::InitArray<Element*,Element>(input,input+SIZE);
			Element *p=input;
			for( size_t i=0; p+SUBSIZE<=input+SIZE; p+=SUBSIZE, ++i )
			{
				std::sort(p, p+SUBSIZE);
				merge_tree.add_stream(InputStream<Element*>(p,p+SUBSIZE));
			}
			std::cerr << "Running LazyMerger...";
			StartTimer();
			Element *e = merge_tree(output, output+(p-input));
			T2 += StopTimer();
			std::cerr << std::endl << "Checking LazyMerger... ";
			check_sort<Element>(output, output+(p-input), chksum);
			std::cerr << std::endl;
			assert( e == output+(p-input) ); (void)e;
		}
	}
	std::cout << "Here!!! :>" << std::endl;
	std::cout << "NM: " << T1/3 << std::endl << "OM: " << T2/3 << std::endl;
	delete[] input;
	delete[] output;
}



#include <complex>
///<summary>Does a lot of fun stuff</summary>
void fun()
{
	merge_test();

	int elts[16] = {503, 87, 512, 61, 908, 170, 897, 275, 653, 426, 154, 509, 612, 677, 765, 703};

/*	merge_tree_types<int,4>::impl_veb t(8);

	typedef pointer_node<3,int*,std::less<int> > Node;
	merge_tree<int,veb_layout<pointer_navigator<Node>, int> > tree(33);
	Node *root = tree.root();
	int *b = elts;
	root.merger(b, elts+16);
*/
//	shell_sort(elts, elts+16);
	__dump(elts, elts+16);
	insertion_sort(elts, elts+4);
	insertion_sort(elts+4, elts+8);
	insertion_sort(elts+8, elts+12);
	insertion_sort(elts+12, elts+16);
	__dump(elts, elts+16);
	funnel::for_merger<4,buffer<int*>,int> m;
	buffer<int*> b1(elts, elts+4); m.add_stream(&b1,0);
	buffer<int*> b2(elts+4, elts+8); m.add_stream(&b2,0);
	buffer<int*> b3(elts+8, elts+12); m.add_stream(&b3,0);
	buffer<int*> b4(elts+12, elts+16); m.add_stream(&b4,0);
	int out[16], *o=out;
    m(o,out+16, 0);
	__dump(out,o);

/*	BasicMergers<InputStream<int*> >::BasicMerger4 m4;
	BasicMerga<InputStream<int*>, 4> m4;
	m4.set_input_stream(0, InputStream<int*>(elts, elts+4));
	m4.set_input_stream(1, InputStream<int*>(elts+4, elts+8));
	m4.set_input_stream(2, InputStream<int*>(elts+8, elts+12));
	m4.set_input_stream(3, InputStream<int*>(elts+12, elts+16));
	int out[16];
	__dump(out, m4(out, out+16, Tager));
*/
	std::cout << std::endl << "Size\tinsertion\tselection\theap\t\tshell\t\tbatcher\'s\tintro" << std::endl;
	Element Sb[4096];
	Element* Se = Sb+4096;
	RandArray::InitArray<Element*,Element>(Sb, Se);
	Element Rb[4096];
	Element* Re = Rb+4096;
	double time[6][1024];
	double mytime, maxtime = 0.0, mintime = 0.0;
	for( int size=3; size<1024; size++ )
	{
		int alg = 0;
		for( int i=0; i<5; i++ )
		{
			std::copy(Sb, Se, Rb);
			StartTimer();
			for( Element *begin=Rb, *end=Rb+size; end<=Re; begin=end, end=begin+size )
				insertion_sort(begin, end);
			mytime = StopTimer();
			if( i )
			{
				time[alg][size] += mytime;
				if( mytime < mintime )
					mintime = mytime;
				if( mytime > maxtime )
					maxtime = mytime;
			}
			else
				time[alg][size] = mytime, mintime = mytime, maxtime = mytime;
		}
		for( Element *begin=Rb, *end=Rb+size; end<=Re; begin=end, end=begin+size )
			check_sort<Element>(begin, end);
		time[alg][size] = (time[alg][size]-mintime-maxtime)/3;
		alg++;

		for( int i=0; i<5; i++ )
		{
			std::copy(Sb, Se, Rb);
			StartTimer();
			for( Element *begin=Rb, *end=Rb+size; end<=Re; begin=end, end=begin+size )
				selection_sort(begin, end);
			mytime = StopTimer();
			if( i )
			{
				time[alg][size] += mytime;
				if( mytime < mintime )
					mintime = mytime;
				if( mytime > maxtime )
					maxtime = mytime;
			}
			else
				time[alg][size] = mytime, mintime = mytime, maxtime = mytime;
		}
		for( Element *begin=Rb, *end=Rb+size; end<=Re; begin=end, end=begin+size )
			check_sort<Element>(begin, end);
		time[alg][size] = (time[alg][size]-mintime-maxtime)/3;
		alg++;

		for( int i=0; i<5; i++ )
		{
			std::copy(Sb, Se, Rb);
			StartTimer();
			for( Element *begin=Rb, *end=Rb+size; end<=Re; begin=end, end=begin+size )
			{
				std::make_heap(begin, end);
				std::sort_heap(begin, end);
			}
			mytime = StopTimer();
			if( i )
			{
				time[alg][size] += mytime;
				if( mytime < mintime )
					mintime = mytime;
				if( mytime > maxtime )
					maxtime = mytime;
			}
			else
				time[alg][size] = mytime, mintime = mytime, maxtime = mytime;
		}
		for( Element *begin=Rb, *end=Rb+size; end<=Re; begin=end, end=begin+size )
			check_sort<Element>(begin, end);
		time[alg][size] = (time[alg][size]-mintime-maxtime)/3;
		alg++;
	
		for( int i=0; i<5; i++ )
		{
			std::copy(Sb, Se, Rb);
			StartTimer();
			for( Element *begin=Rb, *end=Rb+size; end<=Re; begin=end, end=begin+size )
				shell_sort(begin, end);
			mytime = StopTimer();
			if( i )
			{
				time[alg][size] += mytime;
				if( mytime < mintime )
					mintime = mytime;
				if( mytime > maxtime )
					maxtime = mytime;
			}
			else
				time[alg][size] = mytime, mintime = mytime, maxtime = mytime;
		}
		for( Element *begin=Rb, *end=Rb+size; end<=Re; begin=end, end=begin+size )
			check_sort<Element>(begin, end);
		time[alg][size] = (time[alg][size]-mintime-maxtime)/3;
		alg++;

		for( int i=0; i<5; i++ )
		{
			std::copy(Sb, Se, Rb);
			StartTimer();
			for( Element *begin=Rb, *end=Rb+size; end<=Re; begin=end, end=begin+size )
				batcher_sort(begin, end);
			mytime = StopTimer();
			if( i )
			{
				time[alg][size] += mytime;
				if( mytime < mintime )
					mintime = mytime;
				if( mytime > maxtime )
					maxtime = mytime;
			}
			else
				time[alg][size] = mytime, mintime = mytime, maxtime = mytime;
		}
		for( Element *begin=Rb, *end=Rb+size; end<=Re; begin=end, end=begin+size )
			check_sort<Element>(begin, end);
		time[alg][size] = (time[alg][size]-mintime-maxtime)/3;
		alg++;

		for( int i=0; i<5; i++ )
		{
			std::copy(Sb, Se, Rb);
			StartTimer();
			for( Element *begin=Rb, *end=Rb+size; end<=Re; begin=end, end=begin+size )
				std::sort(begin, end);
			mytime = StopTimer();
			if( i )
			{
				time[alg][size] += mytime;
				if( mytime < mintime )
					mintime = mytime;
				if( mytime > maxtime )
					maxtime = mytime;
			}
			else
				time[alg][size] = mytime, mintime = mytime, maxtime = mytime;
		}
		for( Element *begin=Rb, *end=Rb+size; end<=Re; begin=end, end=begin+size )
			check_sort<Element>(begin, end);
		time[alg][size] = (time[alg][size]-mintime-maxtime)/3;
		alg++;
	}
	for( int size=4; size<1024; size++ )
	{
		int runs = 4096/size;
		std::cout << size << '\t' << 1000000*time[0][size]/runs << '\t' << 1000000*time[1][size]/runs << '\t' << 1000000*time[2][size]/runs << '\t' << 1000000*time[3][size]/runs << '\t' << 1000000*time[4][size]/runs << '\t' << 1000000*time[5][size]/runs << std::endl;
	}

/*	size_t i=1<<10;
//	for( size_t i=4; i<13000; i<<=1 )
	{
		std::cout << i << std::endl;
		LazyMerger<ImplBinMergeTree<long>, InputStream<long*> > merger(i);
	}*/
//		merger.invoke((int*)NULL, (int*)NULL);

//		PBBinMergeTree<int, void*> tree(i);
//		PBBinMergeTree<int, void*>::MergerLoc root = tree.root(NULL, NULL);

/*	DeleteFile("\\Documents and Settings\\kv\\Local Settings\\Temp\\Test.img");
	MapArray<int> test1(100000, "\\Documents and Settings\\kv\\Local Settings\\Temp\\Test.img");
	MapArray<int> test2("\\Documents and Settings\\kv\\Local Settings\\Temp\\Test.img", 4000, 200);
	MapArray<int> test3("\\Documents and Settings\\kv\\Local Settings\\Temp\\Test.img", 80000, 4100);
	test1[4100] = 8;
	test1[84000] = 8;
	assert( test2[100] == 8 );
	assert( test3[4000] == 8 );
	MultiMapArray<int> test1(1000000, "\\Documents and Settings\\kv\\Local Settings\\Temp\\Test.img");
	MultiMapArray<int> test2 = test1+100000;
	MultiMapArray<int> test3 = test2+100000;
	test1[350000] = 42;
	assert( test2[250000] == 42 );
	assert( test3[150000] == 42 );
	for( size_t k=2; k<255; k++ )
//		std::cerr << "sizeof(" << k << "-merger) = " << FMerger<sort_t, FBuffer<sort_t> >::size(k) << std::endl;
		std::cerr << k << '\t' << FMerger<sort_t, FBuffer<sort_t> >::size(k) << std::endl;
	for( size_t k=3; k<300; k++ )
		std::cerr << k << '\t' << PBLFMerger<sort_t, sort_t*>::size(_clog(k)) << std::endl;
	for( size_t i=10; i<1636; i++ )
	{
		FBuffer<int>* subs = new FBuffer<int>[i];
		FFMerger<int, FBuffer<int> > merger(i, subs);
		std::cout << (char*)StackMemMan::limit-(char*)StackMemMan::next << std::endl;
		delete[] subs;
	}
	for( size_t i=10; i<1636; i++ )
	{
		std::vector<std::pair<int*,int*> > v(i);
		PBLFFMerger<int, std::vector<std::pair<int*,int*> >::iterator> merger(v.begin(), v.end());
		std::cout << (char*)StackMemMan::limit-(char*)StackMemMan::next << std::endl;
	}
	int in1[7] = { 2, 4, 5, 6, 10, 11, 16 };
	int in2[9] = { 1, 2, 3, 7, 8, 12, 15, 19, 22 };
	int out[16];
	int* in1b = in1,* in1e = in1+7,* in2b = in2,* in2e = in2+9;
	LazyMerge(in1b, in1e, in2b, in2e, out, out+16, LazyMergeNop(), LazyMergeNop());
*//*
	const size_t TS = 37;
	int mk[TS] = { 1, 19, 3, 21, 5, 23, 7, 25, 9, 27, 11, 29, 13, 31, 15, 33, 17, 35, 
					2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 30, 32, 34, 36 };
	int des[TS];

	for( size_t i=0; i<TS/2; i+=2 )
		mk[i] = i+1;
	for( size_t i=1; i<TS/2; i+=2 )
		mk[i] = TS/2+i;
	for( size_t i=TS/2; i<TS; i++ )
		mk[i] = 2*(i-TS/2+1);
//	int mk[4200];
//	for( int ii=0; ii<TS; ii++ )
//		mk[ii] = ii+1;
//	MedianKiller(mk, mk+TS);
//	std::sort(mk, mk+TS);
	mergesort(mk, mk+TS, des);
	*/

	size_t SIZE = 10000000;
	DynArray<int> A(SIZE);
	for( int r=0; r<4; r++ )
	{
		RandArray::InitArray<int*,int>(A.begin(), A.end(SIZE));
		int m = -2100000000;
		std::cout << "Scanning..." << std::endl;
		{
			const size_t B = 8;
			const size_t prefadv = 16;
			StartTimer();
			int* it;
			// Go to next page limit
			for( it = A.begin(); (const_cast<int*>(it)-(int*)0) & (B-1); it++ )
			{
//				if( std::imag(std::sqrt(std::complex<double>(1,0)+std::exp(std::complex<double>(2.4,-1.1)*(double)*it))/1.2) > m )
				if( *it > m )
					m = *it;
//				*it = rand()^(rand()<<8)^(rand()<<16)^(rand()<<24);
			}
			// Enum with prefetch
			for( ; it+prefadv*B<A.end(SIZE); )
			{
//				MemPrefetch(it+prefadv);
				CachePrefetch(it+prefadv*B);
				// Enum the next page
				for( size_t o=0; o<B /*&& it<A.end(SIZE)*/; o++, it++ )
				{
					if( *it > m )
						m = *it;
//					*it = rand()^(rand()<<8)^(rand()<<16)^(rand()<<24);
				}
			}
			// Do the rest
			for( ; it<A.end(SIZE); it++ )
			{
				if( *it > m )
					m = *it;
//				*it = rand()^(rand()<<8)^(rand()<<16)^(rand()<<24);
			}
			std::cout << StopTimer() << "; max: " << m << std::endl;
		}

		StartTimer();
		for( int* it = A.begin(); it<A.end(SIZE); it++ )
		{
			if( *it > m )
				m = *it;
//			*it = rand()^(rand()<<8)^(rand()<<16)^(rand()<<24);
		}
		std::cout << StopTimer() << "; max: " << m << std::endl;
	}
}
